#include <bitset>
#include <cassert>
#include <iostream>

#include "fhe_context.h"

using namespace fhe_deck;

long
compose(const std::vector<long>& bits)
{
    long out = 0;
    for (int i = 0; i < bits.size(); ++i)
        out = (out << 1) | bits[bits.size() - i - 1];
    return out;
}

std::vector<long>
test_add4_boolean(FHEContext& ctx, uint8_t a, uint8_t b)
{
    Ciphertext ct_a0 = ctx.encrypt_public((a >> 0) & 1);
    Ciphertext ct_a1 = ctx.encrypt_public((a >> 1) & 1);
    Ciphertext ct_a2 = ctx.encrypt_public((a >> 2) & 1);
    Ciphertext ct_a3 = ctx.encrypt_public((a >> 3) & 1);
    Ciphertext ct_b0 = ctx.encrypt_public((b >> 0) & 1);
    Ciphertext ct_b1 = ctx.encrypt_public((b >> 1) & 1);
    Ciphertext ct_b2 = ctx.encrypt_public((b >> 2) & 1);
    Ciphertext ct_b3 = ctx.encrypt_public((b >> 3) & 1);

    auto fand2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly and2 = ctx.genrate_lut(fand2);

    auto fnand2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly nand2 = ctx.genrate_lut(fnand2);

    auto for2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly or2 = ctx.genrate_lut(for2);

    auto fnor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly nor2 = ctx.genrate_lut(fnor2);

    auto fxor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly xor2 = ctx.genrate_lut(fxor2);

    auto fxnor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly xnor2 = ctx.genrate_lut(fxnor2);

    std::cerr << "\rLUT0   ";
    Ciphertext gin0 = 2 * ct_a0 + ct_b0;
    Ciphertext gout0 = ctx.eval_lut(&gin0, and2);
    std::cerr << "\rLUT1   ";
    Ciphertext gin1 = 2 * ct_a0 + ct_b0;
    Ciphertext gout1 = ctx.eval_lut(&gin1, xor2);
    std::cerr << "\rLUT2   ";
    Ciphertext gin2 = 2 * ct_a3 + ct_b3;
    Ciphertext gout2 = ctx.eval_lut(&gin2, xor2);
    std::cerr << "\rLUT3   ";
    Ciphertext gin3 = 2 * ct_a2 + ct_b2;
    Ciphertext gout3 = ctx.eval_lut(&gin3, nand2);
    std::cerr << "\rLUT4   ";
    Ciphertext gin4 = 2 * ct_a1 + ct_b1;
    Ciphertext gout4 = ctx.eval_lut(&gin4, nand2);
    std::cerr << "\rLUT5   ";
    Ciphertext gin5 = 2 * ct_a3 + ct_b3;
    Ciphertext gout5 = ctx.eval_lut(&gin5, nand2);
    std::cerr << "\rLUT6   ";
    Ciphertext gin6 = 2 * ct_a2 + ct_b2;
    Ciphertext gout6 = ctx.eval_lut(&gin6, xor2);
    std::cerr << "\rLUT7   ";
    Ciphertext gin7 = 2 * ct_a1 + ct_b1;
    Ciphertext gout7 = ctx.eval_lut(&gin7, xor2);
    std::cerr << "\rLUT8   ";
    Ciphertext gin8 = 2 * gout0 + gout7;
    Ciphertext gout8 = ctx.eval_lut(&gin8, xor2);
    std::cerr << "\rLUT9   ";
    Ciphertext gin9 = 2 * gout0 + gout7;
    Ciphertext gout9 = ctx.eval_lut(&gin9, nand2);
    std::cerr << "\rLUT10   ";
    Ciphertext gin10 = 2 * gout4 + gout9;
    Ciphertext gout10 = ctx.eval_lut(&gin10, nand2);
    std::cerr << "\rLUT11   ";
    Ciphertext gin11 = 2 * gout10 + gout6;
    Ciphertext gout11 = ctx.eval_lut(&gin11, xor2);
    std::cerr << "\rLUT12   ";
    Ciphertext gin12 = 2 * gout10 + gout6;
    Ciphertext gout12 = ctx.eval_lut(&gin12, nand2);
    std::cerr << "\rLUT13   ";
    Ciphertext gin13 = 2 * gout3 + gout12;
    Ciphertext gout13 = ctx.eval_lut(&gin13, nand2);
    std::cerr << "\rLUT14   ";
    Ciphertext gin14 = 2 * gout13 + gout2;
    Ciphertext gout14 = ctx.eval_lut(&gin14, nand2);
    std::cerr << "\rLUT15   ";
    Ciphertext gin15 = 2 * gout5 + gout14;
    Ciphertext gout15 = ctx.eval_lut(&gin15, nand2);
    std::cerr << "\rLUT16   ";
    Ciphertext gin16 = 2 * gout13 + gout2;
    Ciphertext gout16 = ctx.eval_lut(&gin16, xor2);

    std::cerr << "\r          \r";
    std::vector<long> out;
    out.push_back(ctx.decrypt(&gout1)); /* out0 */
    out.push_back(ctx.decrypt(&gout8)); /* out1 */
    out.push_back(ctx.decrypt(&gout11)); /* out2 */
    out.push_back(ctx.decrypt(&gout16)); /* out3 */
    out.push_back(ctx.decrypt(&gout15)); /* out4 */
    return out;
}

std::vector<long>
test_calculator_boolean(FHEContext& ctx, uint8_t op, uint16_t x, uint16_t y)
{
    Ciphertext ct_op0 = ctx.encrypt_public((op >> 0) & 1);
    Ciphertext ct_op1 = ctx.encrypt_public((op >> 1) & 1);
    Ciphertext ct_op2 = ctx.encrypt_public((op >> 2) & 1);
    Ciphertext ct_op3 = ctx.encrypt_public((op >> 3) & 1);
    Ciphertext ct_op4 = ctx.encrypt_public((op >> 4) & 1);
    Ciphertext ct_op5 = ctx.encrypt_public((op >> 5) & 1);
    Ciphertext ct_op6 = ctx.encrypt_public((op >> 6) & 1);
    Ciphertext ct_op7 = ctx.encrypt_public((op >> 7) & 1);
    Ciphertext ct_x0 = ctx.encrypt_public((x >> 0) & 1);
    Ciphertext ct_x1 = ctx.encrypt_public((x >> 1) & 1);
    Ciphertext ct_x2 = ctx.encrypt_public((x >> 2) & 1);
    Ciphertext ct_x3 = ctx.encrypt_public((x >> 3) & 1);
    Ciphertext ct_x4 = ctx.encrypt_public((x >> 4) & 1);
    Ciphertext ct_x5 = ctx.encrypt_public((x >> 5) & 1);
    Ciphertext ct_x6 = ctx.encrypt_public((x >> 6) & 1);
    Ciphertext ct_x7 = ctx.encrypt_public((x >> 7) & 1);
    Ciphertext ct_x8 = ctx.encrypt_public((x >> 8) & 1);
    Ciphertext ct_x9 = ctx.encrypt_public((x >> 9) & 1);
    Ciphertext ct_x10 = ctx.encrypt_public((x >> 10) & 1);
    Ciphertext ct_x11 = ctx.encrypt_public((x >> 11) & 1);
    Ciphertext ct_x12 = ctx.encrypt_public((x >> 12) & 1);
    Ciphertext ct_x13 = ctx.encrypt_public((x >> 13) & 1);
    Ciphertext ct_x14 = ctx.encrypt_public((x >> 14) & 1);
    Ciphertext ct_x15 = ctx.encrypt_public((x >> 15) & 1);
    Ciphertext ct_y0 = ctx.encrypt_public((y >> 0) & 1);
    Ciphertext ct_y1 = ctx.encrypt_public((y >> 1) & 1);
    Ciphertext ct_y2 = ctx.encrypt_public((y >> 2) & 1);
    Ciphertext ct_y3 = ctx.encrypt_public((y >> 3) & 1);
    Ciphertext ct_y4 = ctx.encrypt_public((y >> 4) & 1);
    Ciphertext ct_y5 = ctx.encrypt_public((y >> 5) & 1);
    Ciphertext ct_y6 = ctx.encrypt_public((y >> 6) & 1);
    Ciphertext ct_y7 = ctx.encrypt_public((y >> 7) & 1);
    Ciphertext ct_y8 = ctx.encrypt_public((y >> 8) & 1);
    Ciphertext ct_y9 = ctx.encrypt_public((y >> 9) & 1);
    Ciphertext ct_y10 = ctx.encrypt_public((y >> 10) & 1);
    Ciphertext ct_y11 = ctx.encrypt_public((y >> 11) & 1);
    Ciphertext ct_y12 = ctx.encrypt_public((y >> 12) & 1);
    Ciphertext ct_y13 = ctx.encrypt_public((y >> 13) & 1);
    Ciphertext ct_y14 = ctx.encrypt_public((y >> 14) & 1);
    Ciphertext ct_y15 = ctx.encrypt_public((y >> 15) & 1);

    auto fand2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly and2 = ctx.genrate_lut(fand2);

    auto fnand2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly nand2 = ctx.genrate_lut(fnand2);

    auto for2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly or2 = ctx.genrate_lut(for2);

    auto fnor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly nor2 = ctx.genrate_lut(fnor2);

    auto fxor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly xor2 = ctx.genrate_lut(fxor2);

    auto fxnor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly xnor2 = ctx.genrate_lut(fxnor2);

    std::cerr << "\rLUT0   ";
    Ciphertext gin0 = 2 * ct_y13 + ct_x1;
    Ciphertext gout0 = ctx.eval_lut(&gin0, and2);
    std::cerr << "\rLUT1   ";
    Ciphertext gin1 = 2 * ct_y0 + ct_x7;
    Ciphertext gout1 = ctx.eval_lut(&gin1, nand2);
    std::cerr << "\rLUT2   ";
    Ciphertext gin2 = 2 * ct_y0 + ct_x10;
    Ciphertext gout2 = ctx.eval_lut(&gin2, nand2);
    std::cerr << "\rLUT3   ";
    Ciphertext gin3 = 2 * ct_y1 + ct_x13;
    Ciphertext gout3 = ctx.eval_lut(&gin3, and2);
    std::cerr << "\rLUT4   ";
    Ciphertext gin4 = 2 * ct_y4 + ct_x6;
    Ciphertext gout4 = ctx.eval_lut(&gin4, and2);
    std::cerr << "\rLUT5   ";
    Ciphertext gin5 = 2 * ct_y3 + ct_x12;
    Ciphertext gout5 = ctx.eval_lut(&gin5, and2);
    std::cerr << "\rLUT6   ";
    Ciphertext gin6 = 2 * ct_y3 + ct_x0;
    Ciphertext gout6 = ctx.eval_lut(&gin6, nand2);
    std::cerr << "\rLUT7   ";
    Ciphertext gin7 = 2 * ct_y3 + ct_x8;
    Ciphertext gout7 = ctx.eval_lut(&gin7, and2);
    std::cerr << "\rLUT8   ";
    Ciphertext gin8 = 2 * ct_y9 + ct_x6;
    Ciphertext gout8 = ctx.eval_lut(&gin8, and2);
    std::cerr << "\rLUT9   ";
    Ciphertext gin9 = 2 * ct_y0 + ct_x8;
    Ciphertext gout9 = ctx.eval_lut(&gin9, nand2);
    std::cerr << "\rLUT10   ";
    Ciphertext gin10 = 2 * ct_y6 + ct_x9;
    Ciphertext gout10 = ctx.eval_lut(&gin10, and2);
    std::cerr << "\rLUT11   ";
    Ciphertext gin11 = 2 * ct_y7 + ct_x6;
    Ciphertext gout11 = ctx.eval_lut(&gin11, and2);
    std::cerr << "\rLUT12   ";
    Ciphertext gin12 = 2 * ct_y2 + ct_x11;
    Ciphertext gout12 = ctx.eval_lut(&gin12, and2);
    std::cerr << "\rLUT13   ";
    Ciphertext gin13 = 2 * ct_y3 + ct_x10;
    Ciphertext gout13 = ctx.eval_lut(&gin13, and2);
    std::cerr << "\rLUT14   ";
    Ciphertext gin14 = 2 * gout13 + gout12;
    Ciphertext gout14 = ctx.eval_lut(&gin14, nand2);
    std::cerr << "\rLUT15   ";
    Ciphertext gin15 = 2 * ct_y1 + ct_x8;
    Ciphertext gout15 = ctx.eval_lut(&gin15, and2);
    std::cerr << "\rLUT16   ";
    Ciphertext gin16 = 2 * ct_y14 + ct_x1;
    Ciphertext gout16 = ctx.eval_lut(&gin16, and2);
    std::cerr << "\rLUT17   ";
    Ciphertext gin17 = 2 * ct_y3 + ct_x5;
    Ciphertext gout17 = ctx.eval_lut(&gin17, and2);
    std::cerr << "\rINV18   ";
    Ciphertext gout18 = 1 - ct_x10;

    std::cerr << "\rLUT19   ";
    Ciphertext gin19 = 2 * ct_y0 + ct_x13;
    Ciphertext gout19 = ctx.eval_lut(&gin19, and2);
    std::cerr << "\rLUT20   ";
    Ciphertext gin20 = 2 * ct_y10 + ct_x5;
    Ciphertext gout20 = ctx.eval_lut(&gin20, and2);
    std::cerr << "\rLUT21   ";
    Ciphertext gin21 = 2 * ct_y2 + ct_x7;
    Ciphertext gout21 = ctx.eval_lut(&gin21, and2);
    std::cerr << "\rINV22   ";
    Ciphertext gout22 = 1 - ct_y15;

    std::cerr << "\rINV23   ";
    Ciphertext gout23 = 1 - ct_y2;

    std::cerr << "\rLUT24   ";
    Ciphertext gin24 = 2 * ct_y2 + ct_x1;
    Ciphertext gout24 = ctx.eval_lut(&gin24, and2);
    std::cerr << "\rLUT25   ";
    Ciphertext gin25 = 2 * ct_y2 + ct_x4;
    Ciphertext gout25 = ctx.eval_lut(&gin25, and2);
    std::cerr << "\rLUT26   ";
    Ciphertext gin26 = 2 * ct_y11 + ct_x1;
    Ciphertext gout26 = ctx.eval_lut(&gin26, and2);
    std::cerr << "\rLUT27   ";
    Ciphertext gin27 = 2 * ct_y0 + ct_x9;
    Ciphertext gout27 = ctx.eval_lut(&gin27, nand2);
    std::cerr << "\rLUT28   ";
    Ciphertext gin28 = 2 * ct_y7 + ct_x3;
    Ciphertext gout28 = ctx.eval_lut(&gin28, and2);
    std::cerr << "\rLUT29   ";
    Ciphertext gin29 = 2 * ct_y12 + ct_x2;
    Ciphertext gout29 = ctx.eval_lut(&gin29, nand2);
    std::cerr << "\rINV30   ";
    Ciphertext gout30 = 1 - ct_y9;

    std::cerr << "\rLUT31   ";
    Ciphertext gin31 = 2 * ct_op6 + ct_op7;
    Ciphertext gout31 = ctx.eval_lut(&gin31, nor2);
    std::cerr << "\rLUT32   ";
    Ciphertext gin32 = 2 * gout24 + gout6;
    Ciphertext gout32 = ctx.eval_lut(&gin32, xnor2);
    std::cerr << "\rLUT33   ";
    Ciphertext gin33 = 2 * ct_y3 + ct_x6;
    Ciphertext gout33 = ctx.eval_lut(&gin33, and2);
    std::cerr << "\rLUT34   ";
    Ciphertext gin34 = 2 * ct_y6 + ct_x0;
    Ciphertext gout34 = ctx.eval_lut(&gin34, and2);
    std::cerr << "\rLUT35   ";
    Ciphertext gin35 = 2 * ct_y6 + ct_x2;
    Ciphertext gout35 = ctx.eval_lut(&gin35, and2);
    std::cerr << "\rINV36   ";
    Ciphertext gout36 = 1 - ct_x5;

    std::cerr << "\rLUT37   ";
    Ciphertext gin37 = 2 * ct_y0 + ct_x5;
    Ciphertext gout37 = ctx.eval_lut(&gin37, nand2);
    std::cerr << "\rLUT38   ";
    Ciphertext gin38 = 2 * ct_y6 + ct_x6;
    Ciphertext gout38 = ctx.eval_lut(&gin38, and2);
    std::cerr << "\rLUT39   ";
    Ciphertext gin39 = 2 * ct_y0 + ct_x11;
    Ciphertext gout39 = ctx.eval_lut(&gin39, nand2);
    std::cerr << "\rLUT40   ";
    Ciphertext gin40 = 2 * ct_y8 + ct_x7;
    Ciphertext gout40 = ctx.eval_lut(&gin40, and2);
    std::cerr << "\rLUT41   ";
    Ciphertext gin41 = 2 * ct_y13 + ct_x0;
    Ciphertext gout41 = ctx.eval_lut(&gin41, and2);
    std::cerr << "\rLUT42   ";
    Ciphertext gin42 = 2 * ct_y7 + ct_x4;
    Ciphertext gout42 = ctx.eval_lut(&gin42, and2);
    std::cerr << "\rLUT43   ";
    Ciphertext gin43 = 2 * ct_y15 + ct_x15;
    Ciphertext gout43 = ctx.eval_lut(&gin43, xor2);
    std::cerr << "\rINV44   ";
    Ciphertext gout44 = 1 - ct_y1;

    std::cerr << "\rLUT45   ";
    Ciphertext gin45 = 2 * ct_op2 + ct_op3;
    Ciphertext gout45 = ctx.eval_lut(&gin45, and2);
    std::cerr << "\rINV46   ";
    Ciphertext gout46 = 1 - ct_x14;

    std::cerr << "\rLUT47   ";
    Ciphertext gin47 = 2 * ct_y1 + ct_x2;
    Ciphertext gout47 = ctx.eval_lut(&gin47, and2);
    std::cerr << "\rINV48   ";
    Ciphertext gout48 = 1 - ct_x2;

    std::cerr << "\rLUT49   ";
    Ciphertext gin49 = 2 * ct_y4 + ct_x5;
    Ciphertext gout49 = ctx.eval_lut(&gin49, and2);
    std::cerr << "\rINV50   ";
    Ciphertext gout50 = 1 - ct_y3;

    std::cerr << "\rLUT51   ";
    Ciphertext gin51 = 2 * ct_y6 + ct_x5;
    Ciphertext gout51 = ctx.eval_lut(&gin51, and2);
    std::cerr << "\rLUT52   ";
    Ciphertext gin52 = 2 * ct_y10 + ct_x2;
    Ciphertext gout52 = ctx.eval_lut(&gin52, and2);
    std::cerr << "\rINV53   ";
    Ciphertext gout53 = 1 - ct_x11;

    std::cerr << "\rLUT54   ";
    Ciphertext gin54 = 2 * ct_y0 + ct_x1;
    Ciphertext gout54 = ctx.eval_lut(&gin54, and2);
    std::cerr << "\rINV55   ";
    Ciphertext gout55 = 1 - ct_y8;

    std::cerr << "\rLUT56   ";
    Ciphertext gin56 = 2 * ct_y12 + ct_x1;
    Ciphertext gout56 = ctx.eval_lut(&gin56, nand2);
    std::cerr << "\rLUT57   ";
    Ciphertext gin57 = 2 * ct_y11 + ct_x4;
    Ciphertext gout57 = ctx.eval_lut(&gin57, and2);
    std::cerr << "\rLUT58   ";
    Ciphertext gin58 = 2 * gout42 + gout51;
    Ciphertext gout58 = ctx.eval_lut(&gin58, nand2);
    std::cerr << "\rLUT59   ";
    Ciphertext gin59 = 2 * ct_y5 + ct_x4;
    Ciphertext gout59 = ctx.eval_lut(&gin59, and2);
    std::cerr << "\rINV60   ";
    Ciphertext gout60 = 1 - ct_y7;

    std::cerr << "\rINV61   ";
    Ciphertext gout61 = 1 - ct_x1;

    std::cerr << "\rLUT62   ";
    Ciphertext gin62 = 2 * ct_y7 + ct_x5;
    Ciphertext gout62 = ctx.eval_lut(&gin62, and2);
    std::cerr << "\rLUT63   ";
    Ciphertext gin63 = 2 * ct_y12 + ct_x3;
    Ciphertext gout63 = ctx.eval_lut(&gin63, and2);
    std::cerr << "\rLUT64   ";
    Ciphertext gin64 = 2 * ct_y3 + ct_x7;
    Ciphertext gout64 = ctx.eval_lut(&gin64, and2);
    std::cerr << "\rLUT65   ";
    Ciphertext gin65 = 2 * ct_y12 + ct_x0;
    Ciphertext gout65 = ctx.eval_lut(&gin65, and2);
    std::cerr << "\rINV66   ";
    Ciphertext gout66 = 1 - ct_x9;

    std::cerr << "\rLUT67   ";
    Ciphertext gin67 = 2 * ct_y1 + ct_x0;
    Ciphertext gout67 = ctx.eval_lut(&gin67, and2);
    std::cerr << "\rLUT68   ";
    Ciphertext gin68 = 2 * ct_y10 + ct_x1;
    Ciphertext gout68 = ctx.eval_lut(&gin68, and2);
    std::cerr << "\rLUT69   ";
    Ciphertext gin69 = 2 * ct_y4 + ct_x0;
    Ciphertext gout69 = ctx.eval_lut(&gin69, and2);
    std::cerr << "\rLUT70   ";
    Ciphertext gin70 = 2 * ct_y3 + ct_x4;
    Ciphertext gout70 = ctx.eval_lut(&gin70, and2);
    std::cerr << "\rLUT71   ";
    Ciphertext gin71 = 2 * ct_y5 + ct_x0;
    Ciphertext gout71 = ctx.eval_lut(&gin71, and2);
    std::cerr << "\rLUT72   ";
    Ciphertext gin72 = 2 * ct_y9 + ct_x4;
    Ciphertext gout72 = ctx.eval_lut(&gin72, and2);
    std::cerr << "\rLUT73   ";
    Ciphertext gin73 = 2 * gout33 + gout21;
    Ciphertext gout73 = ctx.eval_lut(&gin73, nand2);
    std::cerr << "\rINV74   ";
    Ciphertext gout74 = 1 - ct_x6;

    std::cerr << "\rLUT75   ";
    Ciphertext gin75 = 2 * ct_y6 + ct_x3;
    Ciphertext gout75 = ctx.eval_lut(&gin75, and2);
    std::cerr << "\rLUT76   ";
    Ciphertext gin76 = 2 * gout22 + gout5;
    Ciphertext gout76 = ctx.eval_lut(&gin76, xnor2);
    std::cerr << "\rLUT77   ";
    Ciphertext gin77 = 2 * ct_y4 + ct_x4;
    Ciphertext gout77 = ctx.eval_lut(&gin77, and2);
    std::cerr << "\rLUT78   ";
    Ciphertext gin78 = 2 * ct_y7 + ct_x1;
    Ciphertext gout78 = ctx.eval_lut(&gin78, and2);
    std::cerr << "\rLUT79   ";
    Ciphertext gin79 = 2 * ct_y6 + ct_x4;
    Ciphertext gout79 = ctx.eval_lut(&gin79, and2);
    std::cerr << "\rLUT80   ";
    Ciphertext gin80 = 2 * (1 - ct_op4) + ct_op5;
    Ciphertext gout80 = ctx.eval_lut(&gin80, and2);
    std::cerr << "\rLUT81   ";
    Ciphertext gin81 = 2 * ct_y1 + ct_x9;
    Ciphertext gout81 = ctx.eval_lut(&gin81, and2);
    std::cerr << "\rLUT82   ";
    Ciphertext gin82 = 2 * ct_y8 + ct_x4;
    Ciphertext gout82 = ctx.eval_lut(&gin82, and2);
    std::cerr << "\rLUT83   ";
    Ciphertext gin83 = 2 * ct_y1 + ct_x14;
    Ciphertext gout83 = ctx.eval_lut(&gin83, and2);
    std::cerr << "\rLUT84   ";
    Ciphertext gin84 = 2 * gout28 + gout79;
    Ciphertext gout84 = ctx.eval_lut(&gin84, nand2);
    std::cerr << "\rLUT85   ";
    Ciphertext gin85 = 2 * gout12 + gout5;
    Ciphertext gout85 = ctx.eval_lut(&gin85, nand2);
    std::cerr << "\rINV86   ";
    Ciphertext gout86 = 1 - ct_y4;

    std::cerr << "\rLUT87   ";
    Ciphertext gin87 = 2 * (1 - ct_op1) + ct_op0;
    Ciphertext gout87 = ctx.eval_lut(&gin87, and2);
    std::cerr << "\rLUT88   ";
    Ciphertext gin88 = 2 * ct_y5 + ct_x5;
    Ciphertext gout88 = ctx.eval_lut(&gin88, and2);
    std::cerr << "\rLUT89   ";
    Ciphertext gin89 = 2 * ct_y6 + ct_x7;
    Ciphertext gout89 = ctx.eval_lut(&gin89, and2);
    std::cerr << "\rLUT90   ";
    Ciphertext gin90 = 2 * gout28 + gout79;
    Ciphertext gout90 = ctx.eval_lut(&gin90, xor2);
    std::cerr << "\rLUT91   ";
    Ciphertext gin91 = 2 * ct_y10 + ct_x3;
    Ciphertext gout91 = ctx.eval_lut(&gin91, and2);
    std::cerr << "\rLUT92   ";
    Ciphertext gin92 = 2 * gout40 + gout8;
    Ciphertext gout92 = ctx.eval_lut(&gin92, xnor2);
    std::cerr << "\rLUT93   ";
    Ciphertext gin93 = 2 * ct_y4 + ct_x2;
    Ciphertext gout93 = ctx.eval_lut(&gin93, and2);
    std::cerr << "\rLUT94   ";
    Ciphertext gin94 = 2 * ct_y8 + ct_x1;
    Ciphertext gout94 = ctx.eval_lut(&gin94, and2);
    std::cerr << "\rLUT95   ";
    Ciphertext gin95 = 2 * ct_y5 + ct_x7;
    Ciphertext gout95 = ctx.eval_lut(&gin95, and2);
    std::cerr << "\rLUT96   ";
    Ciphertext gin96 = 2 * ct_y11 + ct_x3;
    Ciphertext gout96 = ctx.eval_lut(&gin96, and2);
    std::cerr << "\rLUT97   ";
    Ciphertext gin97 = 2 * ct_y2 + ct_x13;
    Ciphertext gout97 = ctx.eval_lut(&gin97, and2);
    std::cerr << "\rLUT98   ";
    Ciphertext gin98 = 2 * ct_y5 + ct_x10;
    Ciphertext gout98 = ctx.eval_lut(&gin98, and2);
    std::cerr << "\rLUT99   ";
    Ciphertext gin99 = 2 * ct_y1 + ct_x12;
    Ciphertext gout99 = ctx.eval_lut(&gin99, and2);
    std::cerr << "\rLUT100   ";
    Ciphertext gin100 = 2 * ct_y5 + ct_x2;
    Ciphertext gout100 = ctx.eval_lut(&gin100, and2);
    std::cerr << "\rINV101   ";
    Ciphertext gout101 = 1 - ct_x3;

    std::cerr << "\rLUT102   ";
    Ciphertext gin102 = 2 * gout62 + gout38;
    Ciphertext gout102 = ctx.eval_lut(&gin102, nand2);
    std::cerr << "\rLUT103   ";
    Ciphertext gin103 = 2 * gout42 + gout51;
    Ciphertext gout103 = ctx.eval_lut(&gin103, xor2);
    std::cerr << "\rLUT104   ";
    Ciphertext gin104 = 2 * gout78 + gout35;
    Ciphertext gout104 = ctx.eval_lut(&gin104, xor2);
    std::cerr << "\rLUT105   ";
    Ciphertext gin105 = 2 * ct_y7 + ct_x7;
    Ciphertext gout105 = ctx.eval_lut(&gin105, nand2);
    std::cerr << "\rLUT106   ";
    Ciphertext gin106 = 2 * ct_y0 + ct_x3;
    Ciphertext gout106 = ctx.eval_lut(&gin106, and2);
    std::cerr << "\rLUT107   ";
    Ciphertext gin107 = 2 * ct_y4 + ct_x3;
    Ciphertext gout107 = ctx.eval_lut(&gin107, and2);
    std::cerr << "\rLUT108   ";
    Ciphertext gin108 = 2 * ct_y7 + ct_x2;
    Ciphertext gout108 = ctx.eval_lut(&gin108, and2);
    std::cerr << "\rINV109   ";
    Ciphertext gout109 = 1 - ct_op0;

    std::cerr << "\rLUT110   ";
    Ciphertext gin110 = 2 * ct_y9 + ct_x5;
    Ciphertext gout110 = ctx.eval_lut(&gin110, and2);
    std::cerr << "\rLUT111   ";
    Ciphertext gin111 = 2 * ct_y0 + ct_x12;
    Ciphertext gout111 = ctx.eval_lut(&gin111, and2);
    std::cerr << "\rLUT112   ";
    Ciphertext gin112 = 2 * ct_y3 + ct_x11;
    Ciphertext gout112 = ctx.eval_lut(&gin112, nand2);
    std::cerr << "\rLUT113   ";
    Ciphertext gin113 = 2 * gout57 + gout63;
    Ciphertext gout113 = ctx.eval_lut(&gin113, xnor2);
    std::cerr << "\rLUT114   ";
    Ciphertext gin114 = 2 * ct_y2 + ct_x0;
    Ciphertext gout114 = ctx.eval_lut(&gin114, and2);
    std::cerr << "\rLUT115   ";
    Ciphertext gin115 = 2 * ct_y1 + ct_x4;
    Ciphertext gout115 = ctx.eval_lut(&gin115, and2);
    std::cerr << "\rLUT116   ";
    Ciphertext gin116 = 2 * ct_y0 + ct_x0;
    Ciphertext gout116 = ctx.eval_lut(&gin116, xnor2);
    std::cerr << "\rLUT117   ";
    Ciphertext gin117 = 2 * ct_y2 + ct_x2;
    Ciphertext gout117 = ctx.eval_lut(&gin117, and2);
    std::cerr << "\rLUT118   ";
    Ciphertext gin118 = 2 * ct_y1 + ct_x11;
    Ciphertext gout118 = ctx.eval_lut(&gin118, and2);
    std::cerr << "\rLUT119   ";
    Ciphertext gin119 = 2 * ct_y1 + ct_x6;
    Ciphertext gout119 = ctx.eval_lut(&gin119, and2);
    std::cerr << "\rLUT120   ";
    Ciphertext gin120 = 2 * ct_y7 + ct_x8;
    Ciphertext gout120 = ctx.eval_lut(&gin120, and2);
    std::cerr << "\rLUT121   ";
    Ciphertext gin121 = 2 * ct_y13 + ct_x2;
    Ciphertext gout121 = ctx.eval_lut(&gin121, and2);
    std::cerr << "\rLUT122   ";
    Ciphertext gin122 = 2 * (1 - ct_x0) + ct_y15;
    Ciphertext gout122 = ctx.eval_lut(&gin122, and2);
    std::cerr << "\rLUT123   ";
    Ciphertext gin123 = 2 * ct_y14 + ct_x0;
    Ciphertext gout123 = ctx.eval_lut(&gin123, and2);
    std::cerr << "\rLUT124   ";
    Ciphertext gin124 = 2 * ct_y4 + ct_x7;
    Ciphertext gout124 = ctx.eval_lut(&gin124, and2);
    std::cerr << "\rINV125   ";
    Ciphertext gout125 = 1 - ct_y5;

    std::cerr << "\rLUT126   ";
    Ciphertext gin126 = 2 * ct_y5 + ct_x3;
    Ciphertext gout126 = ctx.eval_lut(&gin126, and2);
    std::cerr << "\rLUT127   ";
    Ciphertext gin127 = 2 * gout97 + gout76;
    Ciphertext gout127 = ctx.eval_lut(&gin127, xnor2);
    std::cerr << "\rLUT128   ";
    Ciphertext gin128 = 2 * ct_y2 + ct_x3;
    Ciphertext gout128 = ctx.eval_lut(&gin128, and2);
    std::cerr << "\rLUT129   ";
    Ciphertext gin129 = 2 * ct_y3 + ct_x2;
    Ciphertext gout129 = ctx.eval_lut(&gin129, and2);
    std::cerr << "\rLUT130   ";
    Ciphertext gin130 = 2 * ct_y9 + ct_x0;
    Ciphertext gout130 = ctx.eval_lut(&gin130, and2);
    std::cerr << "\rINV131   ";
    Ciphertext gout131 = 1 - ct_y10;

    std::cerr << "\rLUT132   ";
    Ciphertext gin132 = 2 * ct_y3 + ct_x1;
    Ciphertext gout132 = ctx.eval_lut(&gin132, nand2);
    std::cerr << "\rLUT133   ";
    Ciphertext gin133 = 2 * ct_y8 + ct_x5;
    Ciphertext gout133 = ctx.eval_lut(&gin133, and2);
    std::cerr << "\rLUT134   ";
    Ciphertext gin134 = 2 * ct_y8 + ct_x0;
    Ciphertext gout134 = ctx.eval_lut(&gin134, and2);
    std::cerr << "\rLUT135   ";
    Ciphertext gin135 = 2 * ct_y0 + ct_x2;
    Ciphertext gout135 = ctx.eval_lut(&gin135, and2);
    std::cerr << "\rLUT136   ";
    Ciphertext gin136 = 2 * ct_y8 + ct_x2;
    Ciphertext gout136 = ctx.eval_lut(&gin136, and2);
    std::cerr << "\rLUT137   ";
    Ciphertext gin137 = 2 * ct_y1 + ct_x7;
    Ciphertext gout137 = ctx.eval_lut(&gin137, and2);
    std::cerr << "\rLUT138   ";
    Ciphertext gin138 = 2 * ct_y0 + ct_x0;
    Ciphertext gout138 = ctx.eval_lut(&gin138, and2);
    std::cerr << "\rLUT139   ";
    Ciphertext gin139 = 2 * ct_y5 + ct_x6;
    Ciphertext gout139 = ctx.eval_lut(&gin139, and2);
    std::cerr << "\rLUT140   ";
    Ciphertext gin140 = 2 * ct_y11 + ct_x2;
    Ciphertext gout140 = ctx.eval_lut(&gin140, and2);
    std::cerr << "\rINV141   ";
    Ciphertext gout141 = 1 - ct_x7;

    std::cerr << "\rLUT142   ";
    Ciphertext gin142 = 2 * ct_y8 + ct_x6;
    Ciphertext gout142 = ctx.eval_lut(&gin142, and2);
    std::cerr << "\rLUT143   ";
    Ciphertext gin143 = 2 * ct_y2 + ct_x9;
    Ciphertext gout143 = ctx.eval_lut(&gin143, and2);
    std::cerr << "\rLUT144   ";
    Ciphertext gin144 = 2 * (1 - ct_op2) + ct_op1;
    Ciphertext gout144 = ctx.eval_lut(&gin144, and2);
    std::cerr << "\rLUT145   ";
    Ciphertext gin145 = 2 * ct_y4 + ct_x1;
    Ciphertext gout145 = ctx.eval_lut(&gin145, and2);
    std::cerr << "\rLUT146   ";
    Ciphertext gin146 = 2 * ct_y4 + ct_x11;
    Ciphertext gout146 = ctx.eval_lut(&gin146, and2);
    std::cerr << "\rLUT147   ";
    Ciphertext gin147 = 2 * gout132 + gout117;
    Ciphertext gout147 = ctx.eval_lut(&gin147, xnor2);
    std::cerr << "\rINV148   ";
    Ciphertext gout148 = 1 - ct_x4;

    std::cerr << "\rLUT149   ";
    Ciphertext gin149 = 2 * ct_y0 + ct_x6;
    Ciphertext gout149 = ctx.eval_lut(&gin149, nand2);
    std::cerr << "\rLUT150   ";
    Ciphertext gin150 = 2 * ct_y11 + ct_x0;
    Ciphertext gout150 = ctx.eval_lut(&gin150, and2);
    std::cerr << "\rINV151   ";
    Ciphertext gout151 = 1 - gout106;

    std::cerr << "\rLUT152   ";
    Ciphertext gin152 = 2 * ct_y1 + ct_x10;
    Ciphertext gout152 = ctx.eval_lut(&gin152, and2);
    std::cerr << "\rLUT153   ";
    Ciphertext gin153 = 2 * ct_y9 + ct_x2;
    Ciphertext gout153 = ctx.eval_lut(&gin153, and2);
    std::cerr << "\rLUT154   ";
    Ciphertext gin154 = 2 * ct_y0 + ct_x4;
    Ciphertext gout154 = ctx.eval_lut(&gin154, nand2);
    std::cerr << "\rLUT155   ";
    Ciphertext gin155 = 2 * gout54 + gout49;
    Ciphertext gout155 = ctx.eval_lut(&gin155, nand2);
    std::cerr << "\rLUT156   ";
    Ciphertext gin156 = 2 * gout49 + gout27;
    Ciphertext gout156 = ctx.eval_lut(&gin156, xnor2);
    std::cerr << "\rINV157   ";
    Ciphertext gout157 = 1 - ct_x13;

    std::cerr << "\rLUT158   ";
    Ciphertext gin158 = 2 * ct_y9 + ct_x3;
    Ciphertext gout158 = ctx.eval_lut(&gin158, and2);
    std::cerr << "\rINV159   ";
    Ciphertext gout159 = 1 - ct_x12;

    std::cerr << "\rLUT160   ";
    Ciphertext gin160 = 2 * gout13 + gout12;
    Ciphertext gout160 = ctx.eval_lut(&gin160, xor2);
    std::cerr << "\rLUT161   ";
    Ciphertext gin161 = 2 * ct_y5 + ct_x9;
    Ciphertext gout161 = ctx.eval_lut(&gin161, and2);
    std::cerr << "\rLUT162   ";
    Ciphertext gin162 = 2 * ct_y6 + ct_x8;
    Ciphertext gout162 = ctx.eval_lut(&gin162, and2);
    std::cerr << "\rLUT163   ";
    Ciphertext gin163 = 2 * gout67 + gout54;
    Ciphertext gout163 = ctx.eval_lut(&gin163, xor2);
    std::cerr << "\rLUT164   ";
    Ciphertext gin164 = 2 * ct_y4 + ct_x10;
    Ciphertext gout164 = ctx.eval_lut(&gin164, and2);
    std::cerr << "\rLUT165   ";
    Ciphertext gin165 = 2 * ct_y2 + ct_x12;
    Ciphertext gout165 = ctx.eval_lut(&gin165, and2);
    std::cerr << "\rLUT166   ";
    Ciphertext gin166 = 2 * ct_y5 + ct_x1;
    Ciphertext gout166 = ctx.eval_lut(&gin166, and2);
    std::cerr << "\rLUT167   ";
    Ciphertext gin167 = 2 * gout89 + gout120;
    Ciphertext gout167 = ctx.eval_lut(&gin167, nand2);
    std::cerr << "\rLUT168   ";
    Ciphertext gin168 = 2 * ct_y1 + ct_x5;
    Ciphertext gout168 = ctx.eval_lut(&gin168, and2);
    std::cerr << "\rLUT169   ";
    Ciphertext gin169 = 2 * ct_y1 + ct_x3;
    Ciphertext gout169 = ctx.eval_lut(&gin169, and2);
    std::cerr << "\rLUT170   ";
    Ciphertext gin170 = 2 * ct_y10 + ct_x0;
    Ciphertext gout170 = ctx.eval_lut(&gin170, and2);
    std::cerr << "\rINV171   ";
    Ciphertext gout171 = 1 - ct_y6;

    std::cerr << "\rLUT172   ";
    Ciphertext gin172 = 2 * ct_y4 + ct_x8;
    Ciphertext gout172 = ctx.eval_lut(&gin172, and2);
    std::cerr << "\rLUT173   ";
    Ciphertext gin173 = 2 * ct_y2 + ct_x10;
    Ciphertext gout173 = ctx.eval_lut(&gin173, and2);
    std::cerr << "\rINV174   ";
    Ciphertext gout174 = 1 - gout65;

    std::cerr << "\rLUT175   ";
    Ciphertext gin175 = 2 * ct_y1 + ct_y0;
    Ciphertext gout175 = ctx.eval_lut(&gin175, nor2);
    std::cerr << "\rINV176   ";
    Ciphertext gout176 = 1 - ct_y14;

    std::cerr << "\rLUT177   ";
    Ciphertext gin177 = 2 * ct_y2 + ct_x5;
    Ciphertext gout177 = ctx.eval_lut(&gin177, and2);
    std::cerr << "\rLUT178   ";
    Ciphertext gin178 = 2 * ct_y0 + ct_x14;
    Ciphertext gout178 = ctx.eval_lut(&gin178, and2);
    std::cerr << "\rLUT179   ";
    Ciphertext gin179 = 2 * ct_y8 + ct_x3;
    Ciphertext gout179 = ctx.eval_lut(&gin179, and2);
    std::cerr << "\rLUT180   ";
    Ciphertext gin180 = 2 * gout80 + gout31;
    Ciphertext gout180 = ctx.eval_lut(&gin180, and2);
    std::cerr << "\rLUT181   ";
    Ciphertext gin181 = 2 * ct_y7 + ct_x0;
    Ciphertext gout181 = ctx.eval_lut(&gin181, and2);
    std::cerr << "\rLUT182   ";
    Ciphertext gin182 = 2 * ct_y10 + ct_x4;
    Ciphertext gout182 = ctx.eval_lut(&gin182, nand2);
    std::cerr << "\rINV183   ";
    Ciphertext gout183 = 1 - ct_x8;

    std::cerr << "\rLUT184   ";
    Ciphertext gin184 = 2 * gout23 + gout175;
    Ciphertext gout184 = ctx.eval_lut(&gin184, and2);
    std::cerr << "\rLUT185   ";
    Ciphertext gin185 = 2 * gout7 + gout143;
    Ciphertext gout185 = ctx.eval_lut(&gin185, nand2);
    std::cerr << "\rLUT186   ";
    Ciphertext gin186 = 2 * gout182 + gout110;
    Ciphertext gout186 = ctx.eval_lut(&gin186, xnor2);
    std::cerr << "\rLUT187   ";
    Ciphertext gin187 = 2 * ct_y9 + ct_x1;
    Ciphertext gout187 = ctx.eval_lut(&gin187, nand2);
    std::cerr << "\rLUT188   ";
    Ciphertext gin188 = 2 * ct_y2 + ct_x6;
    Ciphertext gout188 = ctx.eval_lut(&gin188, and2);
    std::cerr << "\rLUT189   ";
    Ciphertext gin189 = 2 * ct_y4 + ct_x9;
    Ciphertext gout189 = ctx.eval_lut(&gin189, and2);
    std::cerr << "\rLUT190   ";
    Ciphertext gin190 = 2 * gout4 + gout2;
    Ciphertext gout190 = ctx.eval_lut(&gin190, xnor2);
    std::cerr << "\rLUT191   ";
    Ciphertext gin191 = 2 * ct_y3 + ct_x3;
    Ciphertext gout191 = ctx.eval_lut(&gin191, and2);
    std::cerr << "\rLUT192   ";
    Ciphertext gin192 = 2 * ct_y3 + ct_x9;
    Ciphertext gout192 = ctx.eval_lut(&gin192, and2);
    std::cerr << "\rLUT193   ";
    Ciphertext gin193 = 2 * ct_y6 + ct_x1;
    Ciphertext gout193 = ctx.eval_lut(&gin193, nand2);
    std::cerr << "\rLUT194   ";
    Ciphertext gin194 = 2 * ct_y0 + ct_x15;
    Ciphertext gout194 = ctx.eval_lut(&gin194, and2);
    std::cerr << "\rLUT195   ";
    Ciphertext gin195 = 2 * ct_y1 + ct_x1;
    Ciphertext gout195 = ctx.eval_lut(&gin195, nand2);
    std::cerr << "\rLUT196   ";
    Ciphertext gin196 = 2 * gout77 + gout9;
    Ciphertext gout196 = ctx.eval_lut(&gin196, xnor2);
    std::cerr << "\rLUT197   ";
    Ciphertext gin197 = 2 * gout47 + gout32;
    Ciphertext gout197 = ctx.eval_lut(&gin197, xor2);
    std::cerr << "\rLUT198   ";
    Ciphertext gin198 = 2 * gout130 + gout68;
    Ciphertext gout198 = ctx.eval_lut(&gin198, and2);
    std::cerr << "\rLUT199   ";
    Ciphertext gin199 = 2 * gout47 + gout32;
    Ciphertext gout199 = ctx.eval_lut(&gin199, nand2);
    std::cerr << "\rLUT200   ";
    Ciphertext gin200 = 2 * gout136 + gout90;
    Ciphertext gout200 = ctx.eval_lut(&gin200, nand2);
    std::cerr << "\rLUT201   ";
    Ciphertext gin201 = 2 * gout99 + gout160;
    Ciphertext gout201 = ctx.eval_lut(&gin201, xor2);
    std::cerr << "\rLUT202   ";
    Ciphertext gin202 = 2 * gout78 + gout35;
    Ciphertext gout202 = ctx.eval_lut(&gin202, nand2);
    std::cerr << "\rLUT203   ";
    Ciphertext gin203 = 2 * ct_y5 + ct_x8;
    Ciphertext gout203 = ctx.eval_lut(&gin203, and2);
    std::cerr << "\rLUT204   ";
    Ciphertext gin204 = 2 * ct_y2 + ct_x8;
    Ciphertext gout204 = ctx.eval_lut(&gin204, and2);
    std::cerr << "\rLUT205   ";
    Ciphertext gin205 = 2 * gout64 + gout204;
    Ciphertext gout205 = ctx.eval_lut(&gin205, nand2);
    std::cerr << "\rLUT206   ";
    Ciphertext gin206 = 2 * gout1 + gout107;
    Ciphertext gout206 = ctx.eval_lut(&gin206, xnor2);
    std::cerr << "\rINV207   ";
    Ciphertext gout207 = 1 - gout175;

    std::cerr << "\rLUT208   ";
    Ciphertext gin208 = 2 * gout11 + gout89;
    Ciphertext gout208 = ctx.eval_lut(&gin208, nand2);
    std::cerr << "\rLUT209   ";
    Ciphertext gin209 = 2 * gout34 + gout78;
    Ciphertext gout209 = ctx.eval_lut(&gin209, and2);
    std::cerr << "\rLUT210   ";
    Ciphertext gin210 = 2 * gout136 + gout90;
    Ciphertext gout210 = ctx.eval_lut(&gin210, xor2);
    std::cerr << "\rINV211   ";
    Ciphertext gout211 = 1 - gout135;

    std::cerr << "\rLUT212   ";
    Ciphertext gin212 = 2 * gout11 + gout89;
    Ciphertext gout212 = ctx.eval_lut(&gin212, xor2);
    std::cerr << "\rLUT213   ";
    Ciphertext gin213 = 2 * gout189 + gout19;
    Ciphertext gout213 = ctx.eval_lut(&gin213, xor2);
    std::cerr << "\rLUT214   ";
    Ciphertext gin214 = 2 * ct_op3 + gout144;
    Ciphertext gout214 = ctx.eval_lut(&gin214, and2);
    std::cerr << "\rLUT215   ";
    Ciphertext gin215 = 2 * gout56 + gout41;
    Ciphertext gout215 = ctx.eval_lut(&gin215, xnor2);
    std::cerr << "\rLUT216   ";
    Ciphertext gin216 = 2 * gout129 + gout128;
    Ciphertext gout216 = ctx.eval_lut(&gin216, xor2);
    std::cerr << "\rLUT217   ";
    Ciphertext gin217 = 2 * gout91 + gout72;
    Ciphertext gout217 = ctx.eval_lut(&gin217, xor2);
    std::cerr << "\rLUT218   ";
    Ciphertext gin218 = 2 * gout33 + gout21;
    Ciphertext gout218 = ctx.eval_lut(&gin218, xor2);
    std::cerr << "\rLUT219   ";
    Ciphertext gin219 = 2 * gout91 + gout72;
    Ciphertext gout219 = ctx.eval_lut(&gin219, nand2);
    std::cerr << "\rLUT220   ";
    Ciphertext gin220 = 2 * gout1 + (1 - gout146);
    Ciphertext gout220 = ctx.eval_lut(&gin220, or2);
    std::cerr << "\rLUT221   ";
    Ciphertext gin221 = 2 * gout0 + gout29;
    Ciphertext gout221 = ctx.eval_lut(&gin221, xnor2);
    std::cerr << "\rLUT222   ";
    Ciphertext gin222 = 2 * gout164 + gout178;
    Ciphertext gout222 = ctx.eval_lut(&gin222, xor2);
    std::cerr << "\rLUT223   ";
    Ciphertext gin223 = 2 * gout203 + gout213;
    Ciphertext gout223 = ctx.eval_lut(&gin223, xor2);
    std::cerr << "\rLUT224   ";
    Ciphertext gin224 = 2 * gout124 + gout39;
    Ciphertext gout224 = ctx.eval_lut(&gin224, xnor2);
    std::cerr << "\rLUT225   ";
    Ciphertext gin225 = 2 * gout203 + gout213;
    Ciphertext gout225 = ctx.eval_lut(&gin225, nand2);
    std::cerr << "\rLUT226   ";
    Ciphertext gin226 = 2 * gout115 + gout216;
    Ciphertext gout226 = ctx.eval_lut(&gin226, xor2);
    std::cerr << "\rINV227   ";
    Ciphertext gout227 = 1 - gout138;

    std::cerr << "\rLUT228   ";
    Ciphertext gin228 = 2 * gout149 + (1 - gout164);
    Ciphertext gout228 = ctx.eval_lut(&gin228, or2);
    std::cerr << "\rLUT229   ";
    Ciphertext gin229 = 2 * gout7 + gout143;
    Ciphertext gout229 = ctx.eval_lut(&gin229, xor2);
    std::cerr << "\rLUT230   ";
    Ciphertext gin230 = 2 * gout17 + gout188;
    Ciphertext gout230 = ctx.eval_lut(&gin230, xor2);
    std::cerr << "\rLUT231   ";
    Ciphertext gin231 = 2 * gout83 + gout194;
    Ciphertext gout231 = ctx.eval_lut(&gin231, xnor2);
    std::cerr << "\rLUT232   ";
    Ciphertext gin232 = 2 * gout154 + (1 - gout172);
    Ciphertext gout232 = ctx.eval_lut(&gin232, or2);
    std::cerr << "\rLUT233   ";
    Ciphertext gin233 = 2 * gout134 + gout104;
    Ciphertext gout233 = ctx.eval_lut(&gin233, nand2);
    std::cerr << "\rLUT234   ";
    Ciphertext gin234 = 2 * gout59 + gout156;
    Ciphertext gout234 = ctx.eval_lut(&gin234, nand2);
    std::cerr << "\rLUT235   ";
    Ciphertext gin235 = 2 * gout65 + gout0;
    Ciphertext gout235 = ctx.eval_lut(&gin235, and2);
    std::cerr << "\rLUT236   ";
    Ciphertext gin236 = 2 * gout134 + gout104;
    Ciphertext gout236 = ctx.eval_lut(&gin236, xor2);
    std::cerr << "\rLUT237   ";
    Ciphertext gin237 = 2 * gout67 + gout24;
    Ciphertext gout237 = ctx.eval_lut(&gin237, and2);
    std::cerr << "\rLUT238   ";
    Ciphertext gin238 = 2 * gout59 + gout156;
    Ciphertext gout238 = ctx.eval_lut(&gin238, xor2);
    std::cerr << "\rLUT239   ";
    Ciphertext gin239 = 2 * gout133 + gout212;
    Ciphertext gout239 = ctx.eval_lut(&gin239, nand2);
    std::cerr << "\rLUT240   ";
    Ciphertext gin240 = 2 * gout98 + gout10;
    Ciphertext gout240 = ctx.eval_lut(&gin240, xnor2);
    std::cerr << "\rLUT241   ";
    Ciphertext gin241 = 2 * gout92 + gout113;
    Ciphertext gout241 = ctx.eval_lut(&gin241, xnor2);
    std::cerr << "\rLUT242   ";
    Ciphertext gin242 = 2 * gout68 + gout153;
    Ciphertext gout242 = ctx.eval_lut(&gin242, xor2);
    std::cerr << "\rLUT243   ";
    Ciphertext gin243 = 2 * gout149 + gout93;
    Ciphertext gout243 = ctx.eval_lut(&gin243, xnor2);
    std::cerr << "\rLUT244   ";
    Ciphertext gin244 = 2 * gout138 + (1 - gout195);
    Ciphertext gout244 = ctx.eval_lut(&gin244, and2);
    std::cerr << "\rINV245   ";
    Ciphertext gout245 = 1 - gout198;

    std::cerr << "\rLUT246   ";
    Ciphertext gin246 = 2 * gout70 + gout177;
    Ciphertext gout246 = ctx.eval_lut(&gin246, nand2);
    std::cerr << "\rLUT247   ";
    Ciphertext gin247 = 2 * gout64 + gout204;
    Ciphertext gout247 = ctx.eval_lut(&gin247, xor2);
    std::cerr << "\rLUT248   ";
    Ciphertext gin248 = 2 * gout189 + gout19;
    Ciphertext gout248 = ctx.eval_lut(&gin248, nand2);
    std::cerr << "\rLUT249   ";
    Ciphertext gin249 = 2 * gout133 + gout212;
    Ciphertext gout249 = ctx.eval_lut(&gin249, xor2);
    std::cerr << "\rLUT250   ";
    Ciphertext gin250 = 2 * gout50 + gout184;
    Ciphertext gout250 = ctx.eval_lut(&gin250, and2);
    std::cerr << "\rLUT251   ";
    Ciphertext gin251 = 2 * gout62 + gout38;
    Ciphertext gout251 = ctx.eval_lut(&gin251, xor2);
    std::cerr << "\rLUT252   ";
    Ciphertext gin252 = 2 * gout72 + gout20;
    Ciphertext gout252 = ctx.eval_lut(&gin252, nand2);
    std::cerr << "\rLUT253   ";
    Ciphertext gin253 = 2 * gout129 + gout128;
    Ciphertext gout253 = ctx.eval_lut(&gin253, nand2);
    std::cerr << "\rLUT254   ";
    Ciphertext gin254 = 2 * gout192 + gout173;
    Ciphertext gout254 = ctx.eval_lut(&gin254, xor2);
    std::cerr << "\rLUT255   ";
    Ciphertext gin255 = 2 * gout99 + gout160;
    Ciphertext gout255 = ctx.eval_lut(&gin255, nand2);
    std::cerr << "\rLUT256   ";
    Ciphertext gin256 = 2 * gout179 + gout103;
    Ciphertext gout256 = ctx.eval_lut(&gin256, nand2);
    std::cerr << "\rLUT257   ";
    Ciphertext gin257 = 2 * gout192 + gout173;
    Ciphertext gout257 = ctx.eval_lut(&gin257, nand2);
    std::cerr << "\rLUT258   ";
    Ciphertext gin258 = 2 * gout140 + gout217;
    Ciphertext gout258 = ctx.eval_lut(&gin258, xor2);
    std::cerr << "\rLUT259   ";
    Ciphertext gin259 = 2 * gout70 + gout177;
    Ciphertext gout259 = ctx.eval_lut(&gin259, xor2);
    std::cerr << "\rLUT260   ";
    Ciphertext gin260 = 2 * gout140 + gout217;
    Ciphertext gout260 = ctx.eval_lut(&gin260, nand2);
    std::cerr << "\rLUT261   ";
    Ciphertext gin261 = 2 * gout15 + gout218;
    Ciphertext gout261 = ctx.eval_lut(&gin261, xor2);
    std::cerr << "\rLUT262   ";
    Ciphertext gin262 = 2 * gout56 + (1 - gout121);
    Ciphertext gout262 = ctx.eval_lut(&gin262, or2);
    std::cerr << "\rLUT263   ";
    Ciphertext gin263 = 2 * gout139 + gout224;
    Ciphertext gout263 = ctx.eval_lut(&gin263, nand2);
    std::cerr << "\rLUT264   ";
    Ciphertext gin264 = 2 * gout100 + gout206;
    Ciphertext gout264 = ctx.eval_lut(&gin264, xor2);
    std::cerr << "\rLUT265   ";
    Ciphertext gin265 = 2 * gout179 + gout103;
    Ciphertext gout265 = ctx.eval_lut(&gin265, xor2);
    std::cerr << "\rLUT266   ";
    Ciphertext gin266 = 2 * gout88 + gout190;
    Ciphertext gout266 = ctx.eval_lut(&gin266, nand2);
    std::cerr << "\rLUT267   ";
    Ciphertext gin267 = 2 * gout150 + gout242;
    Ciphertext gout267 = ctx.eval_lut(&gin267, xor2);
    std::cerr << "\rLUT268   ";
    Ciphertext gin268 = 2 * gout228 + gout266;
    Ciphertext gout268 = ctx.eval_lut(&gin268, nand2);
    std::cerr << "\rLUT269   ";
    Ciphertext gin269 = 2 * gout152 + gout229;
    Ciphertext gout269 = ctx.eval_lut(&gin269, xor2);
    std::cerr << "\rLUT270   ";
    Ciphertext gin270 = 2 * gout137 + gout230;
    Ciphertext gout270 = ctx.eval_lut(&gin270, xor2);
    std::cerr << "\rLUT271   ";
    Ciphertext gin271 = 2 * gout268 + gout265;
    Ciphertext gout271 = ctx.eval_lut(&gin271, nand2);
    std::cerr << "\rLUT272   ";
    Ciphertext gin272 = 2 * gout126 + gout196;
    Ciphertext gout272 = ctx.eval_lut(&gin272, nand2);
    std::cerr << "\rLUT273   ";
    Ciphertext gin273 = 2 * gout17 + gout188;
    Ciphertext gout273 = ctx.eval_lut(&gin273, nand2);
    std::cerr << "\rLUT274   ";
    Ciphertext gin274 = 2 * gout135 + gout4;
    Ciphertext gout274 = ctx.eval_lut(&gin274, nand2);
    std::cerr << "\rLUT275   ";
    Ciphertext gin275 = 2 * gout106 + gout124;
    Ciphertext gout275 = ctx.eval_lut(&gin275, nand2);
    std::cerr << "\rLUT276   ";
    Ciphertext gin276 = 2 * gout161 + gout222;
    Ciphertext gout276 = ctx.eval_lut(&gin276, xor2);
    std::cerr << "\rLUT277   ";
    Ciphertext gin277 = 2 * gout195 + gout114;
    Ciphertext gout277 = ctx.eval_lut(&gin277, xnor2);
    std::cerr << "\rLUT278   ";
    Ciphertext gin278 = 2 * gout20 + gout241;
    Ciphertext gout278 = ctx.eval_lut(&gin278, xnor2);
    std::cerr << "\rLUT279   ";
    Ciphertext gin279 = 2 * gout122 + gout16;
    Ciphertext gout279 = ctx.eval_lut(&gin279, xnor2);
    std::cerr << "\rLUT280   ";
    Ciphertext gin280 = 2 * gout115 + gout216;
    Ciphertext gout280 = ctx.eval_lut(&gin280, nand2);
    std::cerr << "\rLUT281   ";
    Ciphertext gin281 = 2 * gout161 + gout222;
    Ciphertext gout281 = ctx.eval_lut(&gin281, nand2);
    std::cerr << "\rLUT282   ";
    Ciphertext gin282 = 2 * gout166 + gout243;
    Ciphertext gout282 = ctx.eval_lut(&gin282, xor2);
    std::cerr << "\rLUT283   ";
    Ciphertext gin283 = 2 * gout105 + gout162;
    Ciphertext gout283 = ctx.eval_lut(&gin283, xnor2);
    std::cerr << "\rLUT284   ";
    Ciphertext gin284 = 2 * (1 - gout114) + gout132;
    Ciphertext gout284 = ctx.eval_lut(&gin284, or2);
    std::cerr << "\rLUT285   ";
    Ciphertext gin285 = 2 * gout68 + gout153;
    Ciphertext gout285 = ctx.eval_lut(&gin285, nand2);
    std::cerr << "\rLUT286   ";
    Ciphertext gin286 = 2 * gout87 + gout45;
    Ciphertext gout286 = ctx.eval_lut(&gin286, and2);
    std::cerr << "\rLUT287   ";
    Ciphertext gin287 = 2 * gout138 + gout77;
    Ciphertext gout287 = ctx.eval_lut(&gin287, and2);
    std::cerr << "\rLUT288   ";
    Ciphertext gin288 = 2 * gout118 + gout254;
    Ciphertext gout288 = ctx.eval_lut(&gin288, xor2);
    std::cerr << "\rLUT289   ";
    Ciphertext gin289 = 2 * gout37 + (1 - gout189);
    Ciphertext gout289 = ctx.eval_lut(&gin289, or2);
    std::cerr << "\rLUT290   ";
    Ciphertext gin290 = 2 * gout108 + gout75;
    Ciphertext gout290 = ctx.eval_lut(&gin290, xor2);
    std::cerr << "\rLUT291   ";
    Ciphertext gin291 = 2 * gout118 + gout254;
    Ciphertext gout291 = ctx.eval_lut(&gin291, nand2);
    std::cerr << "\rLUT292   ";
    Ciphertext gin292 = 2 * gout219 + gout260;
    Ciphertext gout292 = ctx.eval_lut(&gin292, nand2);
    std::cerr << "\rLUT293   ";
    Ciphertext gin293 = 2 * gout169 + gout147;
    Ciphertext gout293 = ctx.eval_lut(&gin293, nand2);
    std::cerr << "\rLUT294   ";
    Ciphertext gin294 = 2 * gout180 + gout214;
    Ciphertext gout294 = ctx.eval_lut(&gin294, and2);
    std::cerr << "\rLUT295   ";
    Ciphertext gin295 = 2 * gout123 + gout221;
    Ciphertext gout295 = ctx.eval_lut(&gin295, nand2);
    std::cerr << "\rLUT296   ";
    Ciphertext gin296 = 2 * gout119 + gout259;
    Ciphertext gout296 = ctx.eval_lut(&gin296, xor2);
    std::cerr << "\rLUT297   ";
    Ciphertext gin297 = 2 * gout172 + gout111;
    Ciphertext gout297 = ctx.eval_lut(&gin297, xor2);
    std::cerr << "\rLUT298   ";
    Ciphertext gin298 = 2 * gout37 + gout145;
    Ciphertext gout298 = ctx.eval_lut(&gin298, xnor2);
    std::cerr << "\rLUT299   ";
    Ciphertext gin299 = 2 * gout108 + gout75;
    Ciphertext gout299 = ctx.eval_lut(&gin299, nand2);
    std::cerr << "\rLUT300   ";
    Ciphertext gin300 = 2 * gout169 + gout147;
    Ciphertext gout300 = ctx.eval_lut(&gin300, xor2);
    std::cerr << "\rLUT301   ";
    Ciphertext gin301 = 2 * gout211 + gout277;
    Ciphertext gout301 = ctx.eval_lut(&gin301, xnor2);
    std::cerr << "\rLUT302   ";
    Ciphertext gin302 = 2 * gout191 + gout25;
    Ciphertext gout302 = ctx.eval_lut(&gin302, nand2);
    std::cerr << "\rLUT303   ";
    Ciphertext gin303 = 2 * gout24 + gout129;
    Ciphertext gout303 = ctx.eval_lut(&gin303, nand2);
    std::cerr << "\rLUT304   ";
    Ciphertext gin304 = 2 * gout164 + gout178;
    Ciphertext gout304 = ctx.eval_lut(&gin304, nand2);
    std::cerr << "\rLUT305   ";
    Ciphertext gin305 = 2 * gout150 + gout242;
    Ciphertext gout305 = ctx.eval_lut(&gin305, nand2);
    std::cerr << "\rLUT306   ";
    Ciphertext gin306 = 2 * gout137 + gout230;
    Ciphertext gout306 = ctx.eval_lut(&gin306, nand2);
    std::cerr << "\rLUT307   ";
    Ciphertext gin307 = 2 * gout191 + gout25;
    Ciphertext gout307 = ctx.eval_lut(&gin307, xor2);
    std::cerr << "\rLUT308   ";
    Ciphertext gin308 = 2 * gout52 + gout158;
    Ciphertext gout308 = ctx.eval_lut(&gin308, nand2);
    std::cerr << "\rLUT309   ";
    Ciphertext gin309 = 2 * gout96 + gout186;
    Ciphertext gout309 = ctx.eval_lut(&gin309, xor2);
    std::cerr << "\rLUT310   ";
    Ciphertext gin310 = 2 * gout15 + gout218;
    Ciphertext gout310 = ctx.eval_lut(&gin310, nand2);
    std::cerr << "\rLUT311   ";
    Ciphertext gin311 = 2 * gout135 + gout277;
    Ciphertext gout311 = ctx.eval_lut(&gin311, and2);
    std::cerr << "\rLUT312   ";
    Ciphertext gin312 = 2 * gout96 + gout186;
    Ciphertext gout312 = ctx.eval_lut(&gin312, nand2);
    std::cerr << "\rLUT313   ";
    Ciphertext gin313 = 2 * gout268 + gout265;
    Ciphertext gout313 = ctx.eval_lut(&gin313, xor2);
    std::cerr << "\rLUT314   ";
    Ciphertext gin314 = 2 * gout112 + gout165;
    Ciphertext gout314 = ctx.eval_lut(&gin314, xnor2);
    std::cerr << "\rLUT315   ";
    Ciphertext gin315 = 2 * gout52 + gout158;
    Ciphertext gout315 = ctx.eval_lut(&gin315, xor2);
    std::cerr << "\rLUT316   ";
    Ciphertext gin316 = 2 * gout193 + gout181;
    Ciphertext gout316 = ctx.eval_lut(&gin316, xnor2);
    std::cerr << "\rLUT317   ";
    Ciphertext gin317 = 2 * gout120 + gout240;
    Ciphertext gout317 = ctx.eval_lut(&gin317, xnor2);
    std::cerr << "\rLUT318   ";
    Ciphertext gin318 = 2 * gout71 + gout298;
    Ciphertext gout318 = ctx.eval_lut(&gin318, xor2);
    std::cerr << "\rLUT319   ";
    Ciphertext gin319 = 2 * gout208 + gout239;
    Ciphertext gout319 = ctx.eval_lut(&gin319, nand2);
    std::cerr << "\rLUT320   ";
    Ciphertext gin320 = 2 * gout166 + gout243;
    Ciphertext gout320 = ctx.eval_lut(&gin320, nand2);
    std::cerr << "\rLUT321   ";
    Ciphertext gin321 = 2 * gout289 + gout234;
    Ciphertext gout321 = ctx.eval_lut(&gin321, nand2);
    std::cerr << "\rLUT322   ";
    Ciphertext gin322 = 2 * gout154 + gout69;
    Ciphertext gout322 = ctx.eval_lut(&gin322, xnor2);
    std::cerr << "\rLUT323   ";
    Ciphertext gin323 = 2 * gout94 + gout290;
    Ciphertext gout323 = ctx.eval_lut(&gin323, xor2);
    std::cerr << "\rLUT324   ";
    Ciphertext gin324 = 2 * gout82 + gout251;
    Ciphertext gout324 = ctx.eval_lut(&gin324, nand2);
    std::cerr << "\rLUT325   ";
    Ciphertext gin325 = 2 * gout3 + gout314;
    Ciphertext gout325 = ctx.eval_lut(&gin325, nand2);
    std::cerr << "\rLUT326   ";
    Ciphertext gin326 = 2 * gout232 + gout272;
    Ciphertext gout326 = ctx.eval_lut(&gin326, nand2);
    std::cerr << "\rLUT327   ";
    Ciphertext gin327 = 2 * gout168 + gout307;
    Ciphertext gout327 = ctx.eval_lut(&gin327, nand2);
    std::cerr << "\rLUT328   ";
    Ciphertext gin328 = 2 * gout3 + gout314;
    Ciphertext gout328 = ctx.eval_lut(&gin328, xor2);
    std::cerr << "\rLUT329   ";
    Ciphertext gin329 = 2 * gout172 + gout111;
    Ciphertext gout329 = ctx.eval_lut(&gin329, nand2);
    std::cerr << "\rLUT330   ";
    Ciphertext gin330 = 2 * gout187 + gout170;
    Ciphertext gout330 = ctx.eval_lut(&gin330, xnor2);
    std::cerr << "\rLUT331   ";
    Ciphertext gin331 = 2 * gout58 + gout256;
    Ciphertext gout331 = ctx.eval_lut(&gin331, nand2);
    std::cerr << "\rLUT332   ";
    Ciphertext gin332 = 2 * gout244 + gout301;
    Ciphertext gout332 = ctx.eval_lut(&gin332, xnor2);
    std::cerr << "\rLUT333   ";
    Ciphertext gin333 = 2 * gout253 + gout280;
    Ciphertext gout333 = ctx.eval_lut(&gin333, nand2);
    std::cerr << "\rLUT334   ";
    Ciphertext gin334 = 2 * gout168 + gout307;
    Ciphertext gout334 = ctx.eval_lut(&gin334, xor2);
    std::cerr << "\rLUT335   ";
    Ciphertext gin335 = 2 * gout26 + gout315;
    Ciphertext gout335 = ctx.eval_lut(&gin335, nand2);
    std::cerr << "\rLUT336   ";
    Ciphertext gin336 = 2 * gout292 + gout309;
    Ciphertext gout336 = ctx.eval_lut(&gin336, nand2);
    std::cerr << "\rLUT337   ";
    Ciphertext gin337 = 2 * gout88 + gout190;
    Ciphertext gout337 = ctx.eval_lut(&gin337, xor2);
    std::cerr << "\rLUT338   ";
    Ciphertext gin338 = 2 * gout100 + gout206;
    Ciphertext gout338 = ctx.eval_lut(&gin338, nand2);
    std::cerr << "\rLUT339   ";
    Ciphertext gin339 = 2 * gout285 + gout305;
    Ciphertext gout339 = ctx.eval_lut(&gin339, nand2);
    std::cerr << "\rLUT340   ";
    Ciphertext gin340 = 2 * gout126 + gout196;
    Ciphertext gout340 = ctx.eval_lut(&gin340, xor2);
    std::cerr << "\rLUT341   ";
    Ciphertext gin341 = 2 * gout262 + gout295;
    Ciphertext gout341 = ctx.eval_lut(&gin341, nand2);
    std::cerr << "\rLUT342   ";
    Ciphertext gin342 = 2 * gout341 + gout279;
    Ciphertext gout342 = ctx.eval_lut(&gin342, xnor2);
    std::cerr << "\rLUT343   ";
    Ciphertext gin343 = 2 * gout321 + gout210;
    Ciphertext gout343 = ctx.eval_lut(&gin343, nand2);
    std::cerr << "\rLUT344   ";
    Ciphertext gin344 = 2 * gout237 + gout197;
    Ciphertext gout344 = ctx.eval_lut(&gin344, xor2);
    std::cerr << "\rLUT345   ";
    Ciphertext gin345 = 2 * gout95 + gout297;
    Ciphertext gout345 = ctx.eval_lut(&gin345, xor2);
    std::cerr << "\rLUT346   ";
    Ciphertext gin346 = 2 * gout84 + gout200;
    Ciphertext gout346 = ctx.eval_lut(&gin346, nand2);
    std::cerr << "\rLUT347   ";
    Ciphertext gin347 = 2 * gout321 + gout210;
    Ciphertext gout347 = ctx.eval_lut(&gin347, xor2);
    std::cerr << "\rLUT348   ";
    Ciphertext gin348 = 2 * gout71 + gout298;
    Ciphertext gout348 = ctx.eval_lut(&gin348, nand2);
    std::cerr << "\rINV349   ";
    Ciphertext gout349 = 1 - gout287;

    std::cerr << "\rLUT350   ";
    Ciphertext gin350 = 2 * gout275 + gout338;
    Ciphertext gout350 = ctx.eval_lut(&gin350, nand2);
    std::cerr << "\rLUT351   ";
    Ciphertext gin351 = 2 * gout81 + gout247;
    Ciphertext gout351 = ctx.eval_lut(&gin351, xor2);
    std::cerr << "\rLUT352   ";
    Ciphertext gin352 = 2 * gout350 + gout236;
    Ciphertext gout352 = ctx.eval_lut(&gin352, nand2);
    std::cerr << "\rLUT353   ";
    Ciphertext gin353 = 2 * gout94 + gout290;
    Ciphertext gout353 = ctx.eval_lut(&gin353, nand2);
    std::cerr << "\rLUT354   ";
    Ciphertext gin354 = 2 * gout123 + gout221;
    Ciphertext gout354 = ctx.eval_lut(&gin354, xor2);
    std::cerr << "\rLUT355   ";
    Ciphertext gin355 = 2 * gout333 + gout334;
    Ciphertext gout355 = ctx.eval_lut(&gin355, xnor2);
    std::cerr << "\rLUT356   ";
    Ciphertext gin356 = 2 * gout139 + gout224;
    Ciphertext gout356 = ctx.eval_lut(&gin356, xor2);
    std::cerr << "\rLUT357   ";
    Ciphertext gin357 = 2 * gout244 + gout301;
    Ciphertext gout357 = ctx.eval_lut(&gin357, and2);
    std::cerr << "\rLUT358   ";
    Ciphertext gin358 = 2 * gout257 + gout291;
    Ciphertext gout358 = ctx.eval_lut(&gin358, nand2);
    std::cerr << "\rLUT359   ";
    Ciphertext gin359 = 2 * gout14 + gout255;
    Ciphertext gout359 = ctx.eval_lut(&gin359, nand2);
    std::cerr << "\rINV360   ";
    Ciphertext gout360 = 1 - gout311;

    std::cerr << "\rLUT361   ";
    Ciphertext gin361 = 2 * gout304 + gout281;
    Ciphertext gout361 = ctx.eval_lut(&gin361, and2);
    std::cerr << "\rLUT362   ";
    Ciphertext gin362 = 2 * gout26 + gout315;
    Ciphertext gout362 = ctx.eval_lut(&gin362, xor2);
    std::cerr << "\rLUT363   ";
    Ciphertext gin363 = 2 * gout248 + gout225;
    Ciphertext gout363 = ctx.eval_lut(&gin363, nand2);
    std::cerr << "\rLUT364   ";
    Ciphertext gin364 = 2 * gout231 + gout127;
    Ciphertext gout364 = ctx.eval_lut(&gin364, xnor2);
    std::cerr << "\rLUT365   ";
    Ciphertext gin365 = 2 * gout142 + gout283;
    Ciphertext gout365 = ctx.eval_lut(&gin365, nand2);
    std::cerr << "\rLUT366   ";
    Ciphertext gin366 = 2 * gout273 + gout306;
    Ciphertext gout366 = ctx.eval_lut(&gin366, nand2);
    std::cerr << "\rLUT367   ";
    Ciphertext gin367 = 2 * gout292 + gout309;
    Ciphertext gout367 = ctx.eval_lut(&gin367, xnor2);
    std::cerr << "\rLUT368   ";
    Ciphertext gin368 = 2 * gout202 + gout233;
    Ciphertext gout368 = ctx.eval_lut(&gin368, nand2);
    std::cerr << "\rLUT369   ";
    Ciphertext gin369 = 2 * gout152 + gout229;
    Ciphertext gout369 = ctx.eval_lut(&gin369, nand2);
    std::cerr << "\rLUT370   ";
    Ciphertext gin370 = 2 * gout361 + gout317;
    Ciphertext gout370 = ctx.eval_lut(&gin370, xnor2);
    std::cerr << "\rLUT371   ";
    Ciphertext gin371 = 2 * gout142 + gout283;
    Ciphertext gout371 = ctx.eval_lut(&gin371, xor2);
    std::cerr << "\rLUT372   ";
    Ciphertext gin372 = 2 * gout358 + gout201;
    Ciphertext gout372 = ctx.eval_lut(&gin372, and2);
    std::cerr << "\rLUT373   ";
    Ciphertext gin373 = 2 * gout95 + gout297;
    Ciphertext gout373 = ctx.eval_lut(&gin373, nand2);
    std::cerr << "\rLUT374   ";
    Ciphertext gin374 = 2 * gout284 + gout199;
    Ciphertext gout374 = ctx.eval_lut(&gin374, nand2);
    std::cerr << "\rLUT375   ";
    Ciphertext gin375 = 2 * gout180 + gout286;
    Ciphertext gout375 = ctx.eval_lut(&gin375, and2);
    std::cerr << "\rLUT376   ";
    Ciphertext gin376 = 2 * gout359 + gout328;
    Ciphertext gout376 = ctx.eval_lut(&gin376, nand2);
    std::cerr << "\rLUT377   ";
    Ciphertext gin377 = 2 * (1 - gout367) + gout354;
    Ciphertext gout377 = ctx.eval_lut(&gin377, and2);
    std::cerr << "\rLUT378   ";
    Ciphertext gin378 = 2 * gout151 + gout344;
    Ciphertext gout378 = ctx.eval_lut(&gin378, xnor2);
    std::cerr << "\rLUT379   ";
    Ciphertext gin379 = 2 * gout308 + gout335;
    Ciphertext gout379 = ctx.eval_lut(&gin379, nand2);
    std::cerr << "\rLUT380   ";
    Ciphertext gin380 = 2 * gout358 + gout201;
    Ciphertext gout380 = ctx.eval_lut(&gin380, xnor2);
    std::cerr << "\rLUT381   ";
    Ciphertext gin381 = 2 * gout102 + gout324;
    Ciphertext gout381 = ctx.eval_lut(&gin381, nand2);
    std::cerr << "\rLUT382   ";
    Ciphertext gin382 = 2 * gout359 + gout328;
    Ciphertext gout382 = ctx.eval_lut(&gin382, xnor2);
    std::cerr << "\rLUT383   ";
    Ciphertext gin383 = 2 * gout73 + gout310;
    Ciphertext gout383 = ctx.eval_lut(&gin383, nand2);
    std::cerr << "\rLUT384   ";
    Ciphertext gin384 = 2 * gout220 + gout263;
    Ciphertext gout384 = ctx.eval_lut(&gin384, nand2);
    std::cerr << "\rLUT385   ";
    Ciphertext gin385 = 2 * gout379 + gout258;
    Ciphertext gout385 = ctx.eval_lut(&gin385, xnor2);
    std::cerr << "\rLUT386   ";
    Ciphertext gin386 = 2 * gout237 + gout197;
    Ciphertext gout386 = ctx.eval_lut(&gin386, nand2);
    std::cerr << "\rLUT387   ";
    Ciphertext gin387 = 2 * gout109 + gout294;
    Ciphertext gout387 = ctx.eval_lut(&gin387, and2);
    std::cerr << "\rLUT388   ";
    Ciphertext gin388 = 2 * gout85 + gout325;
    Ciphertext gout388 = ctx.eval_lut(&gin388, and2);
    std::cerr << "\rLUT389   ";
    Ciphertext gin389 = 2 * gout198 + gout267;
    Ciphertext gout389 = ctx.eval_lut(&gin389, and2);
    std::cerr << "\rLUT390   ";
    Ciphertext gin390 = 2 * gout252 + gout312;
    Ciphertext gout390 = ctx.eval_lut(&gin390, and2);
    std::cerr << "\rLUT391   ";
    Ciphertext gin391 = 2 * gout245 + gout267;
    Ciphertext gout391 = ctx.eval_lut(&gin391, xnor2);
    std::cerr << "\rLUT392   ";
    Ciphertext gin392 = 2 * gout274 + gout320;
    Ciphertext gout392 = ctx.eval_lut(&gin392, nand2);
    std::cerr << "\rLUT393   ";
    Ciphertext gin393 = 2 * gout86 + gout250;
    Ciphertext gout393 = ctx.eval_lut(&gin393, and2);
    std::cerr << "\rLUT394   ";
    Ciphertext gin394 = 2 * gout155 + gout348;
    Ciphertext gout394 = ctx.eval_lut(&gin394, nand2);
    std::cerr << "\rLUT395   ";
    Ciphertext gin395 = 2 * (1 - gout382) + gout276;
    Ciphertext gout395 = ctx.eval_lut(&gin395, and2);
    std::cerr << "\rLUT396   ";
    Ciphertext gin396 = 2 * gout167 + gout365;
    Ciphertext gout396 = ctx.eval_lut(&gin396, and2);
    std::cerr << "\rLUT397   ";
    Ciphertext gin397 = 2 * gout81 + gout247;
    Ciphertext gout397 = ctx.eval_lut(&gin397, nand2);
    std::cerr << "\rLUT398   ";
    Ciphertext gin398 = 2 * gout82 + gout251;
    Ciphertext gout398 = ctx.eval_lut(&gin398, xor2);
    std::cerr << "\rLUT399   ";
    Ciphertext gin399 = 2 * gout346 + gout313;
    Ciphertext gout399 = ctx.eval_lut(&gin399, xor2);
    std::cerr << "\rLUT400   ";
    Ciphertext gin400 = 2 * gout346 + gout313;
    Ciphertext gout400 = ctx.eval_lut(&gin400, nand2);
    std::cerr << "\rLUT401   ";
    Ciphertext gin401 = 2 * gout119 + gout259;
    Ciphertext gout401 = ctx.eval_lut(&gin401, nand2);
    std::cerr << "\rLUT402   ";
    Ciphertext gin402 = 2 * gout333 + gout334;
    Ciphertext gout402 = ctx.eval_lut(&gin402, and2);
    std::cerr << "\rLUT403   ";
    Ciphertext gin403 = 2 * gout339 + gout362;
    Ciphertext gout403 = ctx.eval_lut(&gin403, xor2);
    std::cerr << "\rLUT404   ";
    Ciphertext gin404 = 2 * (1 - gout385) + gout215;
    Ciphertext gout404 = ctx.eval_lut(&gin404, and2);
    std::cerr << "\rLUT405   ";
    Ciphertext gin405 = 2 * gout355 + gout282;
    Ciphertext gout405 = ctx.eval_lut(&gin405, xnor2);
    std::cerr << "\rLUT406   ";
    Ciphertext gin406 = 2 * gout375 + gout207;
    Ciphertext gout406 = ctx.eval_lut(&gin406, and2);
    std::cerr << "\rLUT407   ";
    Ciphertext gin407 = 2 * gout363 + gout371;
    Ciphertext gout407 = ctx.eval_lut(&gin407, nand2);
    std::cerr << "\rLUT408   ";
    Ciphertext gin408 = 2 * gout387 + gout116;
    Ciphertext gout408 = ctx.eval_lut(&gin408, or2);
    std::cerr << "\rLUT409   ";
    Ciphertext gin409 = 2 * (1 - gout184) + gout375;
    Ciphertext gout409 = ctx.eval_lut(&gin409, and2);
    std::cerr << "\rLUT410   ";
    Ciphertext gin410 = 2 * gout387 + gout163;
    Ciphertext gout410 = ctx.eval_lut(&gin410, nand2);
    std::cerr << "\rLUT411   ";
    Ciphertext gin411 = 2 * gout302 + gout327;
    Ciphertext gout411 = ctx.eval_lut(&gin411, nand2);
    std::cerr << "\rLUT412   ";
    Ciphertext gin412 = 2 * gout383 + gout351;
    Ciphertext gout412 = ctx.eval_lut(&gin412, xnor2);
    std::cerr << "\rLUT413   ";
    Ciphertext gin413 = 2 * gout350 + gout236;
    Ciphertext gout413 = ctx.eval_lut(&gin413, xor2);
    std::cerr << "\rLUT414   ";
    Ciphertext gin414 = 2 * gout336 + (1 - gout377);
    Ciphertext gout414 = ctx.eval_lut(&gin414, and2);
    std::cerr << "\rLUT415   ";
    Ciphertext gin415 = 2 * (1 - gout355) + gout282;
    Ciphertext gout415 = ctx.eval_lut(&gin415, and2);
    std::cerr << "\rLUT416   ";
    Ciphertext gin416 = 2 * gout294 + gout375;
    Ciphertext gout416 = ctx.eval_lut(&gin416, or2);
    std::cerr << "\rLUT417   ";
    Ciphertext gin417 = 2 * gout363 + gout371;
    Ciphertext gout417 = ctx.eval_lut(&gin417, xor2);
    std::cerr << "\rLUT418   ";
    Ciphertext gin418 = 2 * gout326 + gout323;
    Ciphertext gout418 = ctx.eval_lut(&gin418, xor2);
    std::cerr << "\rLUT419   ";
    Ciphertext gin419 = 2 * gout360 + gout378;
    Ciphertext gout419 = ctx.eval_lut(&gin419, xnor2);
    std::cerr << "\rLUT420   ";
    Ciphertext gin420 = 2 * gout326 + gout323;
    Ciphertext gout420 = ctx.eval_lut(&gin420, nand2);
    std::cerr << "\rLUT421   ";
    Ciphertext gin421 = 2 * gout121 + gout390;
    Ciphertext gout421 = ctx.eval_lut(&gin421, xnor2);
    std::cerr << "\rLUT422   ";
    Ciphertext gin422 = 2 * ct_y12 + gout375;
    Ciphertext gout422 = ctx.eval_lut(&gin422, nand2);
    std::cerr << "\rLUT423   ";
    Ciphertext gin423 = 2 * gout299 + gout353;
    Ciphertext gout423 = ctx.eval_lut(&gin423, nand2);
    std::cerr << "\rLUT424   ";
    Ciphertext gin424 = 2 * gout125 + gout393;
    Ciphertext gout424 = ctx.eval_lut(&gin424, and2);
    std::cerr << "\rLUT425   ";
    Ciphertext gin425 = 2 * gout367 + gout354;
    Ciphertext gout425 = ctx.eval_lut(&gin425, xnor2);
    std::cerr << "\rLUT426   ";
    Ciphertext gin426 = 2 * gout380 + gout223;
    Ciphertext gout426 = ctx.eval_lut(&gin426, xnor2);
    std::cerr << "\rLUT427   ";
    Ciphertext gin427 = 2 * gout106 + gout344;
    Ciphertext gout427 = ctx.eval_lut(&gin427, nand2);
    std::cerr << "\rINV428   ";
    Ciphertext gout428 = 1 - gout387;

    std::cerr << "\rLUT429   ";
    Ciphertext gin429 = 2 * gout146 + gout388;
    Ciphertext gout429 = ctx.eval_lut(&gin429, xnor2);
    std::cerr << "\rLUT430   ";
    Ciphertext gin430 = 2 * (1 - gout380) + gout223;
    Ciphertext gout430 = ctx.eval_lut(&gin430, and2);
    std::cerr << "\rLUT431   ";
    Ciphertext gin431 = 2 * gout303 + gout293;
    Ciphertext gout431 = ctx.eval_lut(&gin431, nand2);
    std::cerr << "\rLUT432   ";
    Ciphertext gin432 = 2 * gout394 + gout34;
    Ciphertext gout432 = ctx.eval_lut(&gin432, and2);
    std::cerr << "\rLUT433   ";
    Ciphertext gin433 = 2 * gout23 + gout406;
    Ciphertext gout433 = ctx.eval_lut(&gin433, xnor2);
    std::cerr << "\rLUT434   ";
    Ciphertext gin434 = 2 * gout392 + gout316;
    Ciphertext gout434 = ctx.eval_lut(&gin434, xor2);
    std::cerr << "\rLUT435   ";
    Ciphertext gin435 = 2 * ct_y14 + gout375;
    Ciphertext gout435 = ctx.eval_lut(&gin435, and2);
    std::cerr << "\rLUT436   ";
    Ciphertext gin436 = 2 * gout379 + gout258;
    Ciphertext gout436 = ctx.eval_lut(&gin436, and2);
    std::cerr << "\rLUT437   ";
    Ciphertext gin437 = 2 * (1 - gout393) + gout375;
    Ciphertext gout437 = ctx.eval_lut(&gin437, and2);
    std::cerr << "\rLUT438   ";
    Ciphertext gin438 = 2 * gout392 + gout316;
    Ciphertext gout438 = ctx.eval_lut(&gin438, and2);
    std::cerr << "\rLUT439   ";
    Ciphertext gin439 = 2 * gout329 + gout373;
    Ciphertext gout439 = ctx.eval_lut(&gin439, nand2);
    std::cerr << "\rLUT440   ";
    Ciphertext gin440 = 2 * gout366 + gout261;
    Ciphertext gout440 = ctx.eval_lut(&gin440, and2);
    std::cerr << "\rLUT441   ";
    Ciphertext gin441 = 2 * gout384 + gout398;
    Ciphertext gout441 = ctx.eval_lut(&gin441, xor2);
    std::cerr << "\rLUT442   ";
    Ciphertext gin442 = 2 * gout429 + gout370;
    Ciphertext gout442 = ctx.eval_lut(&gin442, xnor2);
    std::cerr << "\rLUT443   ";
    Ciphertext gin443 = 2 * gout374 + gout300;
    Ciphertext gout443 = ctx.eval_lut(&gin443, xnor2);
    std::cerr << "\rLUT444   ";
    Ciphertext gin444 = 2 * ct_y13 + gout375;
    Ciphertext gout444 = ctx.eval_lut(&gin444, and2);
    std::cerr << "\rLUT445   ";
    Ciphertext gin445 = 2 * gout384 + gout398;
    Ciphertext gout445 = ctx.eval_lut(&gin445, nand2);
    std::cerr << "\rLUT446   ";
    Ciphertext gin446 = 2 * gout205 + gout397;
    Ciphertext gout446 = ctx.eval_lut(&gin446, nand2);
    std::cerr << "\rLUT447   ";
    Ciphertext gin447 = 2 * gout394 + gout34;
    Ciphertext gout447 = ctx.eval_lut(&gin447, xor2);
    std::cerr << "\rLUT448   ";
    Ciphertext gin448 = 2 * gout366 + gout261;
    Ciphertext gout448 = ctx.eval_lut(&gin448, xnor2);
    std::cerr << "\rLUT449   ";
    Ciphertext gin449 = 2 * gout382 + gout276;
    Ciphertext gout449 = ctx.eval_lut(&gin449, xnor2);
    std::cerr << "\rLUT450   ";
    Ciphertext gin450 = 2 * gout385 + gout215;
    Ciphertext gout450 = ctx.eval_lut(&gin450, xnor2);
    std::cerr << "\rLUT451   ";
    Ciphertext gin451 = 2 * gout376 + (1 - gout395);
    Ciphertext gout451 = ctx.eval_lut(&gin451, and2);
    std::cerr << "\rLUT452   ";
    Ciphertext gin452 = 2 * gout209 + gout413;
    Ciphertext gout452 = ctx.eval_lut(&gin452, nand2);
    std::cerr << "\rLUT453   ";
    Ciphertext gin453 = 2 * gout448 + gout238;
    Ciphertext gout453 = ctx.eval_lut(&gin453, xnor2);
    std::cerr << "\rLUT454   ";
    Ciphertext gin454 = 2 * gout319 + gout417;
    Ciphertext gout454 = ctx.eval_lut(&gin454, nand2);
    std::cerr << "\rLUT455   ";
    Ciphertext gin455 = 2 * (1 - gout250) + gout375;
    Ciphertext gout455 = ctx.eval_lut(&gin455, and2);
    std::cerr << "\rLUT456   ";
    Ciphertext gin456 = 2 * gout339 + gout362;
    Ciphertext gout456 = ctx.eval_lut(&gin456, nand2);
    std::cerr << "\rLUT457   ";
    Ciphertext gin457 = 2 * gout174 + gout403;
    Ciphertext gout457 = ctx.eval_lut(&gin457, xnor2);
    std::cerr << "\rLUT458   ";
    Ciphertext gin458 = 2 * ct_y11 + gout375;
    Ciphertext gout458 = ctx.eval_lut(&gin458, nand2);
    std::cerr << "\rLUT459   ";
    Ciphertext gin459 = 2 * gout209 + gout413;
    Ciphertext gout459 = ctx.eval_lut(&gin459, xor2);
    std::cerr << "\rLUT460   ";
    Ciphertext gin460 = 2 * ct_y0 + gout375;
    Ciphertext gout460 = ctx.eval_lut(&gin460, and2);
    std::cerr << "\rLUT461   ";
    Ciphertext gin461 = 2 * ct_x2 + gout433;
    Ciphertext gout461 = ctx.eval_lut(&gin461, nand2);
    std::cerr << "\rLUT462   ";
    Ciphertext gin462 = 2 * gout368 + gout418;
    Ciphertext gout462 = ctx.eval_lut(&gin462, xor2);
    std::cerr << "\rLUT463   ";
    Ciphertext gin463 = 2 * gout278 + gout421;
    Ciphertext gout463 = ctx.eval_lut(&gin463, xnor2);
    std::cerr << "\rLUT464   ";
    Ciphertext gin464 = 2 * gout368 + gout418;
    Ciphertext gout464 = ctx.eval_lut(&gin464, nand2);
    std::cerr << "\rLUT465   ";
    Ciphertext gin465 = 2 * gout407 + gout454;
    Ciphertext gout465 = ctx.eval_lut(&gin465, and2);
    std::cerr << "\rLUT466   ";
    Ciphertext gin466 = 2 * gout374 + gout300;
    Ciphertext gout466 = ctx.eval_lut(&gin466, and2);
    std::cerr << "\rLUT467   ";
    Ciphertext gin467 = 2 * gout387 + gout138;
    Ciphertext gout467 = ctx.eval_lut(&gin467, nand2);
    std::cerr << "\rLUT468   ";
    Ciphertext gin468 = 2 * gout44 + gout460;
    Ciphertext gout468 = ctx.eval_lut(&gin468, xnor2);
    std::cerr << "\rLUT469   ";
    Ciphertext gin469 = 2 * gout185 + gout369;
    Ciphertext gout469 = ctx.eval_lut(&gin469, nand2);
    std::cerr << "\rLUT470   ";
    Ciphertext gin470 = 2 * gout423 + gout347;
    Ciphertext gout470 = ctx.eval_lut(&gin470, nand2);
    std::cerr << "\rLUT471   ";
    Ciphertext gin471 = 2 * gout364 + gout442;
    Ciphertext gout471 = ctx.eval_lut(&gin471, xnor2);
    std::cerr << "\rLUT472   ";
    Ciphertext gin472 = 2 * gout383 + gout351;
    Ciphertext gout472 = ctx.eval_lut(&gin472, and2);
    std::cerr << "\rLUT473   ";
    Ciphertext gin473 = 2 * gout311 + gout378;
    Ciphertext gout473 = ctx.eval_lut(&gin473, and2);
    std::cerr << "\rLUT474   ";
    Ciphertext gin474 = 2 * gout431 + gout226;
    Ciphertext gout474 = ctx.eval_lut(&gin474, and2);
    std::cerr << "\rLUT475   ";
    Ciphertext gin475 = 2 * gout439 + gout249;
    Ciphertext gout475 = ctx.eval_lut(&gin475, nand2);
    std::cerr << "\rLUT476   ";
    Ciphertext gin476 = 2 * gout412 + gout337;
    Ciphertext gout476 = ctx.eval_lut(&gin476, xnor2);
    std::cerr << "\rLUT477   ";
    Ciphertext gin477 = 2 * gout423 + gout347;
    Ciphertext gout477 = ctx.eval_lut(&gin477, xor2);
    std::cerr << "\rLUT478   ";
    Ciphertext gin478 = 2 * (1 - gout412) + gout337;
    Ciphertext gout478 = ctx.eval_lut(&gin478, and2);
    std::cerr << "\rLUT479   ";
    Ciphertext gin479 = 2 * gout411 + gout296;
    Ciphertext gout479 = ctx.eval_lut(&gin479, xnor2);
    std::cerr << "\rLUT480   ";
    Ciphertext gin480 = 2 * gout171 + gout424;
    Ciphertext gout480 = ctx.eval_lut(&gin480, and2);
    std::cerr << "\rLUT481   ";
    Ciphertext gin481 = 2 * gout402 + gout415;
    Ciphertext gout481 = ctx.eval_lut(&gin481, or2);
    std::cerr << "\rLUT482   ";
    Ciphertext gin482 = 2 * gout451 + gout396;
    Ciphertext gout482 = ctx.eval_lut(&gin482, xnor2);
    std::cerr << "\rLUT483   ";
    Ciphertext gin483 = 2 * gout469 + gout288;
    Ciphertext gout483 = ctx.eval_lut(&gin483, and2);
    std::cerr << "\rLUT484   ";
    Ciphertext gin484 = 2 * gout411 + gout296;
    Ciphertext gout484 = ctx.eval_lut(&gin484, and2);
    std::cerr << "\rLUT485   ";
    Ciphertext gin485 = 2 * (1 - gout479) + gout264;
    Ciphertext gout485 = ctx.eval_lut(&gin485, and2);
    std::cerr << "\rLUT486   ";
    Ciphertext gin486 = 2 * gout331 + gout441;
    Ciphertext gout486 = ctx.eval_lut(&gin486, xor2);
    std::cerr << "\rLUT487   ";
    Ciphertext gin487 = 2 * gout420 + gout464;
    Ciphertext gout487 = ctx.eval_lut(&gin487, nand2);
    std::cerr << "\rLUT488   ";
    Ciphertext gin488 = 2 * gout331 + gout441;
    Ciphertext gout488 = ctx.eval_lut(&gin488, nand2);
    std::cerr << "\rLUT489   ";
    Ciphertext gin489 = 2 * gout469 + gout288;
    Ciphertext gout489 = ctx.eval_lut(&gin489, xnor2);
    std::cerr << "\rLUT490   ";
    Ciphertext gin490 = 2 * gout246 + gout401;
    Ciphertext gout490 = ctx.eval_lut(&gin490, nand2);
    std::cerr << "\rLUT491   ";
    Ciphertext gin491 = 2 * gout416 + gout408;
    Ciphertext gout491 = ctx.eval_lut(&gin491, and2);
    std::cerr << "\rLUT492   ";
    Ciphertext gin492 = 2 * gout445 + gout488;
    Ciphertext gout492 = ctx.eval_lut(&gin492, nand2);
    std::cerr << "\rLUT493   ";
    Ciphertext gin493 = 2 * gout86 + gout455;
    Ciphertext gout493 = ctx.eval_lut(&gin493, xnor2);
    std::cerr << "\rLUT494   ";
    Ciphertext gin494 = 2 * gout48 + gout433;
    Ciphertext gout494 = ctx.eval_lut(&gin494, xnor2);
    std::cerr << "\rLUT495   ";
    Ciphertext gin495 = 2 * (1 - gout489) + gout345;
    Ciphertext gout495 = ctx.eval_lut(&gin495, and2);
    std::cerr << "\rLUT496   ";
    Ciphertext gin496 = 2 * gout490 + gout270;
    Ciphertext gout496 = ctx.eval_lut(&gin496, and2);
    std::cerr << "\rLUT497   ";
    Ciphertext gin497 = 2 * gout482 + gout471;
    Ciphertext gout497 = ctx.eval_lut(&gin497, xnor2);
    std::cerr << "\rLUT498   ";
    Ciphertext gin498 = 2 * gout271 + gout400;
    Ciphertext gout498 = ctx.eval_lut(&gin498, nand2);
    std::cerr << "\rLUT499   ";
    Ciphertext gin499 = 2 * gout50 + gout409;
    Ciphertext gout499 = ctx.eval_lut(&gin499, xnor2);
    std::cerr << "\rLUT500   ";
    Ciphertext gin500 = 2 * gout386 + gout427;
    Ciphertext gout500 = ctx.eval_lut(&gin500, nand2);
    std::cerr << "\rLUT501   ";
    Ciphertext gin501 = 2 * gout443 + gout322;
    Ciphertext gout501 = ctx.eval_lut(&gin501, xnor2);
    std::cerr << "\rLUT502   ";
    Ciphertext gin502 = 2 * gout65 + gout403;
    Ciphertext gout502 = ctx.eval_lut(&gin502, nand2);
    std::cerr << "\rLUT503   ";
    Ciphertext gin503 = 2 * gout372 + gout430;
    Ciphertext gout503 = ctx.eval_lut(&gin503, or2);
    std::cerr << "\rLUT504   ";
    Ciphertext gin504 = 2 * (1 - gout443) + gout322;
    Ciphertext gout504 = ctx.eval_lut(&gin504, and2);
    std::cerr << "\rLUT505   ";
    Ciphertext gin505 = 2 * gout357 + gout419;
    Ciphertext gout505 = ctx.eval_lut(&gin505, nand2);
    std::cerr << "\rLUT506   ";
    Ciphertext gin506 = 2 * gout319 + gout417;
    Ciphertext gout506 = ctx.eval_lut(&gin506, xor2);
    std::cerr << "\rLUT507   ";
    Ciphertext gin507 = 2 * gout446 + gout269;
    Ciphertext gout507 = ctx.eval_lut(&gin507, xnor2);
    std::cerr << "\rLUT508   ";
    Ciphertext gin508 = 2 * ct_x1 + gout468;
    Ciphertext gout508 = ctx.eval_lut(&gin508, nand2);
    std::cerr << "\rLUT509   ";
    Ciphertext gin509 = 2 * gout357 + gout419;
    Ciphertext gout509 = ctx.eval_lut(&gin509, xnor2);
    std::cerr << "\rLUT510   ";
    Ciphertext gin510 = 2 * gout507 + gout356;
    Ciphertext gout510 = ctx.eval_lut(&gin510, xnor2);
    std::cerr << "\rINV511   ";
    Ciphertext gout511 = 1 - gout473;

    std::cerr << "\rLUT512   ";
    Ciphertext gin512 = 2 * gout431 + gout226;
    Ciphertext gout512 = ctx.eval_lut(&gin512, xnor2);
    std::cerr << "\rLUT513   ";
    Ciphertext gin513 = 2 * gout416 + gout410;
    Ciphertext gout513 = ctx.eval_lut(&gin513, and2);
    std::cerr << "\rLUT514   ";
    Ciphertext gin514 = 2 * gout503 + gout449;
    Ciphertext gout514 = ctx.eval_lut(&gin514, and2);
    std::cerr << "\rLUT515   ";
    Ciphertext gin515 = 2 * ct_x4 + gout493;
    Ciphertext gout515 = ctx.eval_lut(&gin515, nand2);
    std::cerr << "\rLUT516   ";
    Ciphertext gin516 = 2 * gout467 + gout491;
    Ciphertext gout516 = ctx.eval_lut(&gin516, nand2);
    std::cerr << "\rLUT517   ";
    Ciphertext gin517 = 2 * (1 - gout424) + gout375;
    Ciphertext gout517 = ctx.eval_lut(&gin517, and2);
    std::cerr << "\rLUT518   ";
    Ciphertext gin518 = 2 * gout436 + gout404;
    Ciphertext gout518 = ctx.eval_lut(&gin518, or2);
    std::cerr << "\rLUT519   ";
    Ciphertext gin519 = 2 * gout125 + gout437;
    Ciphertext gout519 = ctx.eval_lut(&gin519, xnor2);
    std::cerr << "\rLUT520   ";
    Ciphertext gin520 = 2 * gout148 + gout493;
    Ciphertext gout520 = ctx.eval_lut(&gin520, xnor2);
    std::cerr << "\rLUT521   ";
    Ciphertext gin521 = 2 * gout352 + gout452;
    Ciphertext gout521 = ctx.eval_lut(&gin521, nand2);
    std::cerr << "\rLUT522   ";
    Ciphertext gin522 = 2 * gout503 + gout449;
    Ciphertext gout522 = ctx.eval_lut(&gin522, xnor2);
    std::cerr << "\rLUT523   ";
    Ciphertext gin523 = 2 * ct_x3 + gout499;
    Ciphertext gout523 = ctx.eval_lut(&gin523, nand2);
    std::cerr << "\rLUT524   ";
    Ciphertext gin524 = 2 * gout479 + gout264;
    Ciphertext gout524 = ctx.eval_lut(&gin524, xnor2);
    std::cerr << "\rLUT525   ";
    Ciphertext gin525 = 2 * gout439 + gout249;
    Ciphertext gout525 = ctx.eval_lut(&gin525, xor2);
    std::cerr << "\rLUT526   ";
    Ciphertext gin526 = 2 * gout489 + gout345;
    Ciphertext gout526 = ctx.eval_lut(&gin526, xnor2);
    std::cerr << "\rLUT527   ";
    Ciphertext gin527 = 2 * gout484 + gout485;
    Ciphertext gout527 = ctx.eval_lut(&gin527, or2);
    std::cerr << "\rLUT528   ";
    Ciphertext gin528 = 2 * gout446 + gout269;
    Ciphertext gout528 = ctx.eval_lut(&gin528, and2);
    std::cerr << "\rLUT529   ";
    Ciphertext gin529 = 2 * gout465 + gout463;
    Ciphertext gout529 = ctx.eval_lut(&gin529, xnor2);
    std::cerr << "\rLUT530   ";
    Ciphertext gin530 = 2 * gout500 + gout501;
    Ciphertext gout530 = ctx.eval_lut(&gin530, and2);
    std::cerr << "\rLUT531   ";
    Ciphertext gin531 = 2 * gout490 + gout270;
    Ciphertext gout531 = ctx.eval_lut(&gin531, xnor2);
    std::cerr << "\rLUT532   ";
    Ciphertext gin532 = 2 * (1 - gout448) + gout238;
    Ciphertext gout532 = ctx.eval_lut(&gin532, and2);
    std::cerr << "\rLUT533   ";
    Ciphertext gin533 = 2 * gout522 + gout506;
    Ciphertext gout533 = ctx.eval_lut(&gin533, xnor2);
    std::cerr << "\rLUT534   ";
    Ciphertext gin534 = 2 * gout492 + gout450;
    Ciphertext gout534 = ctx.eval_lut(&gin534, xor2);
    std::cerr << "\rLUT535   ";
    Ciphertext gin535 = 2 * gout60 + gout480;
    Ciphertext gout535 = ctx.eval_lut(&gin535, and2);
    std::cerr << "\rLUT536   ";
    Ciphertext gin536 = 2 * gout492 + gout450;
    Ciphertext gout536 = ctx.eval_lut(&gin536, nand2);
    std::cerr << "\rLUT537   ";
    Ciphertext gin537 = 2 * gout500 + gout501;
    Ciphertext gout537 = ctx.eval_lut(&gin537, xor2);
    std::cerr << "\rLUT538   ";
    Ciphertext gin538 = 2 * gout456 + gout502;
    Ciphertext gout538 = ctx.eval_lut(&gin538, nand2);
    std::cerr << "\rLUT539   ";
    Ciphertext gin539 = 2 * gout483 + gout495;
    Ciphertext gout539 = ctx.eval_lut(&gin539, or2);
    std::cerr << "\rLUT540   ";
    Ciphertext gin540 = 2 * gout538 + gout534;
    Ciphertext gout540 = ctx.eval_lut(&gin540, nand2);
    std::cerr << "\rLUT541   ";
    Ciphertext gin541 = 2 * gout343 + gout470;
    Ciphertext gout541 = ctx.eval_lut(&gin541, nand2);
    std::cerr << "\rLUT542   ";
    Ciphertext gin542 = 2 * gout61 + gout468;
    Ciphertext gout542 = ctx.eval_lut(&gin542, xnor2);
    std::cerr << "\rLUT543   ";
    Ciphertext gin543 = 2 * (1 - gout512) + gout318;
    Ciphertext gout543 = ctx.eval_lut(&gin543, and2);
    std::cerr << "\rLUT544   ";
    Ciphertext gin544 = 2 * gout36 + gout519;
    Ciphertext gout544 = ctx.eval_lut(&gin544, xnor2);
    std::cerr << "\rLUT545   ";
    Ciphertext gin545 = 2 * gout101 + gout499;
    Ciphertext gout545 = ctx.eval_lut(&gin545, xnor2);
    std::cerr << "\rLUT546   ";
    Ciphertext gin546 = 2 * gout529 + gout414;
    Ciphertext gout546 = ctx.eval_lut(&gin546, xnor2);
    std::cerr << "\rLUT547   ";
    Ciphertext gin547 = 2 * gout487 + gout330;
    Ciphertext gout547 = ctx.eval_lut(&gin547, xor2);
    std::cerr << "\rLUT548   ";
    Ciphertext gin548 = 2 * gout472 + gout478;
    Ciphertext gout548 = ctx.eval_lut(&gin548, or2);
    std::cerr << "\rLUT549   ";
    Ciphertext gin549 = 2 * gout521 + gout130;
    Ciphertext gout549 = ctx.eval_lut(&gin549, and2);
    std::cerr << "\rLUT550   ";
    Ciphertext gin550 = 2 * gout498 + gout457;
    Ciphertext gout550 = ctx.eval_lut(&gin550, nand2);
    std::cerr << "\rLUT551   ";
    Ciphertext gin551 = 2 * gout539 + gout426;
    Ciphertext gout551 = ctx.eval_lut(&gin551, xnor2);
    std::cerr << "\rLUT552   ";
    Ciphertext gin552 = 2 * gout487 + gout330;
    Ciphertext gout552 = ctx.eval_lut(&gin552, and2);
    std::cerr << "\rLUT553   ";
    Ciphertext gin553 = 2 * gout512 + gout318;
    Ciphertext gout553 = ctx.eval_lut(&gin553, xnor2);
    std::cerr << "\rLUT554   ";
    Ciphertext gin554 = 2 * (1 - gout480) + gout375;
    Ciphertext gout554 = ctx.eval_lut(&gin554, and2);
    std::cerr << "\rLUT555   ";
    Ciphertext gin555 = 2 * gout381 + gout525;
    Ciphertext gout555 = ctx.eval_lut(&gin555, xor2);
    std::cerr << "\rLUT556   ";
    Ciphertext gin556 = 2 * gout498 + gout457;
    Ciphertext gout556 = ctx.eval_lut(&gin556, xor2);
    std::cerr << "\rLUT557   ";
    Ciphertext gin557 = 2 * (1 - gout507) + gout356;
    Ciphertext gout557 = ctx.eval_lut(&gin557, and2);
    std::cerr << "\rLUT558   ";
    Ciphertext gin558 = 2 * gout536 + gout540;
    Ciphertext gout558 = ctx.eval_lut(&gin558, nand2);
    std::cerr << "\rLUT559   ";
    Ciphertext gin559 = 2 * gout227 + gout542;
    Ciphertext gout559 = ctx.eval_lut(&gin559, xnor2);
    std::cerr << "\rLUT560   ";
    Ciphertext gin560 = 2 * gout466 + gout504;
    Ciphertext gout560 = ctx.eval_lut(&gin560, or2);
    std::cerr << "\rLUT561   ";
    Ciphertext gin561 = 2 * gout531 + gout340;
    Ciphertext gout561 = ctx.eval_lut(&gin561, xnor2);
    std::cerr << "\rLUT562   ";
    Ciphertext gin562 = 2 * (1 - gout505) + gout537;
    Ciphertext gout562 = ctx.eval_lut(&gin562, and2);
    std::cerr << "\rLUT563   ";
    Ciphertext gin563 = 2 * (1 - gout535) + gout375;
    Ciphertext gout563 = ctx.eval_lut(&gin563, and2);
    std::cerr << "\rLUT564   ";
    Ciphertext gin564 = 2 * gout473 + gout537;
    Ciphertext gout564 = ctx.eval_lut(&gin564, and2);
    std::cerr << "\rLUT565   ";
    Ciphertext gin565 = 2 * gout511 + gout505;
    Ciphertext gout565 = ctx.eval_lut(&gin565, nand2);
    std::cerr << "\rLUT566   ";
    Ciphertext gin566 = 2 * (1 - gout531) + gout340;
    Ciphertext gout566 = ctx.eval_lut(&gin566, and2);
    std::cerr << "\rLUT567   ";
    Ciphertext gin567 = 2 * gout537 + gout565;
    Ciphertext gout567 = ctx.eval_lut(&gin567, xnor2);
    std::cerr << "\rLUT568   ";
    Ciphertext gin568 = 2 * gout474 + gout543;
    Ciphertext gout568 = ctx.eval_lut(&gin568, or2);
    std::cerr << "\rLUT569   ";
    Ciphertext gin569 = 2 * gout560 + gout553;
    Ciphertext gout569 = ctx.eval_lut(&gin569, xor2);
    std::cerr << "\rLUT570   ";
    Ciphertext gin570 = 2 * gout548 + gout510;
    Ciphertext gout570 = ctx.eval_lut(&gin570, xnor2);
    std::cerr << "\rLUT571   ";
    Ciphertext gin571 = 2 * gout560 + gout553;
    Ciphertext gout571 = ctx.eval_lut(&gin571, nand2);
    std::cerr << "\rLUT572   ";
    Ciphertext gin572 = 2 * gout568 + gout405;
    Ciphertext gout572 = ctx.eval_lut(&gin572, and2);
    std::cerr << "\rLUT573   ";
    Ciphertext gin573 = 2 * gout481 + gout524;
    Ciphertext gout573 = ctx.eval_lut(&gin573, xnor2);
    std::cerr << "\rLUT574   ";
    Ciphertext gin574 = 2 * (1 - gout573) + gout434;
    Ciphertext gout574 = ctx.eval_lut(&gin574, and2);
    std::cerr << "\rLUT575   ";
    Ciphertext gin575 = 2 * gout481 + gout524;
    Ciphertext gout575 = ctx.eval_lut(&gin575, and2);
    std::cerr << "\rLUT576   ";
    Ciphertext gin576 = 2 * (1 - gout551) + gout555;
    Ciphertext gout576 = ctx.eval_lut(&gin576, and2);
    std::cerr << "\rLUT577   ";
    Ciphertext gin577 = 2 * gout541 + gout391;
    Ciphertext gout577 = ctx.eval_lut(&gin577, xor2);
    std::cerr << "\rLUT578   ";
    Ciphertext gin578 = 2 * (1 - gout522) + gout506;
    Ciphertext gout578 = ctx.eval_lut(&gin578, and2);
    std::cerr << "\rLUT579   ";
    Ciphertext gin579 = 2 * gout171 + gout517;
    Ciphertext gout579 = ctx.eval_lut(&gin579, xnor2);
    std::cerr << "\rLUT580   ";
    Ciphertext gin580 = 2 * gout235 + gout558;
    Ciphertext gout580 = ctx.eval_lut(&gin580, xor2);
    std::cerr << "\rLUT581   ";
    Ciphertext gin581 = 2 * gout538 + gout534;
    Ciphertext gout581 = ctx.eval_lut(&gin581, xor2);
    std::cerr << "\rLUT582   ";
    Ciphertext gin582 = 2 * gout539 + gout426;
    Ciphertext gout582 = ctx.eval_lut(&gin582, and2);
    std::cerr << "\rLUT583   ";
    Ciphertext gin583 = 2 * ct_x5 + gout519;
    Ciphertext gout583 = ctx.eval_lut(&gin583, nand2);
    std::cerr << "\rLUT584   ";
    Ciphertext gin584 = 2 * gout521 + gout130;
    Ciphertext gout584 = ctx.eval_lut(&gin584, xor2);
    std::cerr << "\rLUT585   ";
    Ciphertext gin585 = 2 * gout389 + gout556;
    Ciphertext gout585 = ctx.eval_lut(&gin585, xor2);
    std::cerr << "\rLUT586   ";
    Ciphertext gin586 = 2 * gout389 + gout556;
    Ciphertext gout586 = ctx.eval_lut(&gin586, nand2);
    std::cerr << "\rLUT587   ";
    Ciphertext gin587 = 2 * gout381 + gout525;
    Ciphertext gout587 = ctx.eval_lut(&gin587, nand2);
    std::cerr << "\rLUT588   ";
    Ciphertext gin588 = 2 * gout55 + gout535;
    Ciphertext gout588 = ctx.eval_lut(&gin588, and2);
    std::cerr << "\rLUT589   ";
    Ciphertext gin589 = 2 * gout528 + gout557;
    Ciphertext gout589 = ctx.eval_lut(&gin589, or2);
    std::cerr << "\rLUT590   ";
    Ciphertext gin590 = 2 * gout496 + gout566;
    Ciphertext gout590 = ctx.eval_lut(&gin590, or2);
    std::cerr << "\rLUT591   ";
    Ciphertext gin591 = 2 * gout527 + gout561;
    Ciphertext gout591 = ctx.eval_lut(&gin591, xnor2);
    std::cerr << "\rLUT592   ";
    Ciphertext gin592 = 2 * gout514 + gout578;
    Ciphertext gout592 = ctx.eval_lut(&gin592, or2);
    std::cerr << "\rLUT593   ";
    Ciphertext gin593 = 2 * gout527 + gout561;
    Ciphertext gout593 = ctx.eval_lut(&gin593, and2);
    std::cerr << "\rLUT594   ";
    Ciphertext gin594 = 2 * gout590 + gout453;
    Ciphertext gout594 = ctx.eval_lut(&gin594, xnor2);
    std::cerr << "\rLUT595   ";
    Ciphertext gin595 = 2 * gout591 + gout459;
    Ciphertext gout595 = ctx.eval_lut(&gin595, xnor2);
    std::cerr << "\rLUT596   ";
    Ciphertext gin596 = 2 * gout568 + gout405;
    Ciphertext gout596 = ctx.eval_lut(&gin596, xnor2);
    std::cerr << "\rLUT597   ";
    Ciphertext gin597 = 2 * gout287 + gout569;
    Ciphertext gout597 = ctx.eval_lut(&gin597, nand2);
    std::cerr << "\rLUT598   ";
    Ciphertext gin598 = 2 * gout55 + gout563;
    Ciphertext gout598 = ctx.eval_lut(&gin598, xnor2);
    std::cerr << "\rLUT599   ";
    Ciphertext gin599 = 2 * gout74 + gout579;
    Ciphertext gout599 = ctx.eval_lut(&gin599, xnor2);
    std::cerr << "\rLUT600   ";
    Ciphertext gin600 = 2 * gout548 + gout510;
    Ciphertext gout600 = ctx.eval_lut(&gin600, and2);
    std::cerr << "\rLUT601   ";
    Ciphertext gin601 = 2 * (1 - gout570) + gout399;
    Ciphertext gout601 = ctx.eval_lut(&gin601, and2);
    std::cerr << "\rLUT602   ";
    Ciphertext gin602 = 2 * gout440 + gout532;
    Ciphertext gout602 = ctx.eval_lut(&gin602, or2);
    std::cerr << "\rLUT603   ";
    Ciphertext gin603 = 2 * gout530 + gout564;
    Ciphertext gout603 = ctx.eval_lut(&gin603, or2);
    std::cerr << "\rLUT604   ";
    Ciphertext gin604 = 2 * gout546 + gout497;
    Ciphertext gout604 = ctx.eval_lut(&gin604, xnor2);
    std::cerr << "\rLUT605   ";
    Ciphertext gin605 = 2 * gout602 + gout476;
    Ciphertext gout605 = ctx.eval_lut(&gin605, and2);
    std::cerr << "\rLUT606   ";
    Ciphertext gin606 = 2 * ct_x6 + gout579;
    Ciphertext gout606 = ctx.eval_lut(&gin606, nand2);
    std::cerr << "\rLUT607   ";
    Ciphertext gin607 = 2 * gout570 + gout399;
    Ciphertext gout607 = ctx.eval_lut(&gin607, xnor2);
    std::cerr << "\rLUT608   ";
    Ciphertext gin608 = 2 * gout541 + gout391;
    Ciphertext gout608 = ctx.eval_lut(&gin608, and2);
    std::cerr << "\rLUT609   ";
    Ciphertext gin609 = 2 * gout551 + gout555;
    Ciphertext gout609 = ctx.eval_lut(&gin609, xnor2);
    std::cerr << "\rLUT610   ";
    Ciphertext gin610 = 2 * gout138 + gout542;
    Ciphertext gout610 = ctx.eval_lut(&gin610, nand2);
    std::cerr << "\rLUT611   ";
    Ciphertext gin611 = 2 * gout602 + gout476;
    Ciphertext gout611 = ctx.eval_lut(&gin611, xnor2);
    std::cerr << "\rLUT612   ";
    Ciphertext gin612 = 2 * gout60 + gout554;
    Ciphertext gout612 = ctx.eval_lut(&gin612, xnor2);
    std::cerr << "\rLUT613   ";
    Ciphertext gin613 = 2 * gout428 + gout559;
    Ciphertext gout613 = ctx.eval_lut(&gin613, nand2);
    std::cerr << "\rLUT614   ";
    Ciphertext gin614 = 2 * (1 - gout611) + gout477;
    Ciphertext gout614 = ctx.eval_lut(&gin614, and2);
    std::cerr << "\rLUT615   ";
    Ciphertext gin615 = 2 * (1 - gout588) + gout375;
    Ciphertext gout615 = ctx.eval_lut(&gin615, and2);
    std::cerr << "\rLUT616   ";
    Ciphertext gin616 = 2 * gout575 + gout574;
    Ciphertext gout616 = ctx.eval_lut(&gin616, or2);
    std::cerr << "\rLUT617   ";
    Ciphertext gin617 = 2 * gout235 + gout558;
    Ciphertext gout617 = ctx.eval_lut(&gin617, and2);
    std::cerr << "\rLUT618   ";
    Ciphertext gin618 = 2 * gout600 + gout601;
    Ciphertext gout618 = ctx.eval_lut(&gin618, or2);
    std::cerr << "\rLUT619   ";
    Ciphertext gin619 = 2 * gout550 + gout586;
    Ciphertext gout619 = ctx.eval_lut(&gin619, nand2);
    std::cerr << "\rLUT620   ";
    Ciphertext gin620 = 2 * gout589 + gout526;
    Ciphertext gout620 = ctx.eval_lut(&gin620, and2);
    std::cerr << "\rLUT621   ";
    Ciphertext gin621 = 2 * ct_x8 + gout598;
    Ciphertext gout621 = ctx.eval_lut(&gin621, nand2);
    std::cerr << "\rLUT622   ";
    Ciphertext gin622 = 2 * gout582 + gout576;
    Ciphertext gout622 = ctx.eval_lut(&gin622, or2);
    std::cerr << "\rLUT623   ";
    Ciphertext gin623 = 2 * gout590 + gout453;
    Ciphertext gout623 = ctx.eval_lut(&gin623, and2);
    std::cerr << "\rLUT624   ";
    Ciphertext gin624 = 2 * gout594 + gout462;
    Ciphertext gout624 = ctx.eval_lut(&gin624, xnor2);
    std::cerr << "\rLUT625   ";
    Ciphertext gin625 = 2 * (1 - gout591) + gout459;
    Ciphertext gout625 = ctx.eval_lut(&gin625, and2);
    std::cerr << "\rLUT626   ";
    Ciphertext gin626 = 2 * gout349 + gout569;
    Ciphertext gout626 = ctx.eval_lut(&gin626, xnor2);
    std::cerr << "\rLUT627   ";
    Ciphertext gin627 = 2 * ct_x7 + gout612;
    Ciphertext gout627 = ctx.eval_lut(&gin627, nand2);
    std::cerr << "\rLUT628   ";
    Ciphertext gin628 = 2 * gout622 + gout533;
    Ciphertext gout628 = ctx.eval_lut(&gin628, xnor2);
    std::cerr << "\rLUT629   ";
    Ciphertext gin629 = 2 * gout593 + gout625;
    Ciphertext gout629 = ctx.eval_lut(&gin629, or2);
    std::cerr << "\rLUT630   ";
    Ciphertext gin630 = 2 * (1 - gout596) + gout447;
    Ciphertext gout630 = ctx.eval_lut(&gin630, and2);
    std::cerr << "\rLUT631   ";
    Ciphertext gin631 = 2 * gout573 + gout434;
    Ciphertext gout631 = ctx.eval_lut(&gin631, xnor2);
    std::cerr << "\rLUT632   ";
    Ciphertext gin632 = 2 * gout508 + gout610;
    Ciphertext gout632 = ctx.eval_lut(&gin632, nand2);
    std::cerr << "\rLUT633   ";
    Ciphertext gin633 = 2 * gout475 + gout587;
    Ciphertext gout633 = ctx.eval_lut(&gin633, nand2);
    std::cerr << "\rLUT634   ";
    Ciphertext gin634 = 2 * gout596 + gout447;
    Ciphertext gout634 = ctx.eval_lut(&gin634, xnor2);
    std::cerr << "\rLUT635   ";
    Ciphertext gin635 = 2 * gout632 + gout494;
    Ciphertext gout635 = ctx.eval_lut(&gin635, nand2);
    std::cerr << "\rLUT636   ";
    Ciphertext gin636 = 2 * gout589 + gout526;
    Ciphertext gout636 = ctx.eval_lut(&gin636, xnor2);
    std::cerr << "\rLUT637   ";
    Ciphertext gin637 = 2 * gout629 + gout624;
    Ciphertext gout637 = ctx.eval_lut(&gin637, xnor2);
    std::cerr << "\rLUT638   ";
    Ciphertext gin638 = 2 * gout30 + gout588;
    Ciphertext gout638 = ctx.eval_lut(&gin638, and2);
    std::cerr << "\rLUT639   ";
    Ciphertext gin639 = 2 * gout611 + gout477;
    Ciphertext gout639 = ctx.eval_lut(&gin639, xnor2);
    std::cerr << "\rLUT640   ";
    Ciphertext gin640 = 2 * gout183 + gout598;
    Ciphertext gout640 = ctx.eval_lut(&gin640, xnor2);
    std::cerr << "\rLUT641   ";
    Ciphertext gin641 = 2 * gout605 + gout614;
    Ciphertext gout641 = ctx.eval_lut(&gin641, or2);
    std::cerr << "\rLUT642   ";
    Ciphertext gin642 = 2 * (1 - gout594) + gout462;
    Ciphertext gout642 = ctx.eval_lut(&gin642, and2);
    std::cerr << "\rLUT643   ";
    Ciphertext gin643 = 2 * gout616 + gout595;
    Ciphertext gout643 = ctx.eval_lut(&gin643, xor2);
    std::cerr << "\rLUT644   ";
    Ciphertext gin644 = 2 * (1 - gout637) + gout584;
    Ciphertext gout644 = ctx.eval_lut(&gin644, and2);
    std::cerr << "\rLUT645   ";
    Ciphertext gin645 = 2 * gout616 + gout595;
    Ciphertext gout645 = ctx.eval_lut(&gin645, nand2);
    std::cerr << "\rLUT646   ";
    Ciphertext gin646 = 2 * gout641 + gout607;
    Ciphertext gout646 = ctx.eval_lut(&gin646, xnor2);
    std::cerr << "\rLUT647   ";
    Ciphertext gin647 = 2 * gout571 + gout597;
    Ciphertext gout647 = ctx.eval_lut(&gin647, nand2);
    std::cerr << "\rLUT648   ";
    Ciphertext gin648 = 2 * gout572 + gout630;
    Ciphertext gout648 = ctx.eval_lut(&gin648, or2);
    std::cerr << "\rLUT649   ";
    Ciphertext gin649 = 2 * gout592 + gout604;
    Ciphertext gout649 = ctx.eval_lut(&gin649, xnor2);
    std::cerr << "\rINV650   ";
    Ciphertext gout650 = 1 - gout634;

    std::cerr << "\rLUT651   ";
    Ciphertext gin651 = 2 * gout623 + gout642;
    Ciphertext gout651 = ctx.eval_lut(&gin651, or2);
    std::cerr << "\rLUT652   ";
    Ciphertext gin652 = 2 * gout632 + gout494;
    Ciphertext gout652 = ctx.eval_lut(&gin652, xnor2);
    std::cerr << "\rLUT653   ";
    Ciphertext gin653 = 2 * gout651 + gout639;
    Ciphertext gout653 = ctx.eval_lut(&gin653, and2);
    std::cerr << "\rLUT654   ";
    Ciphertext gin654 = 2 * gout141 + gout612;
    Ciphertext gout654 = ctx.eval_lut(&gin654, xnor2);
    std::cerr << "\rIMUX655  ";
    Ciphertext gin655A = 2 * gout332 + gout387;
    Ciphertext gout655A = ctx.eval_lut(&gin655A, and2);
    Ciphertext gin655B = 2 * gout652 + (1 - gout387);
    Ciphertext gout655B = ctx.eval_lut(&gin655B, and2);
    Ciphertext gout655 = gout655A + gout655B;
    std::cerr << "\rLUT656   ";
    Ciphertext gin656 = 2 * gout622 + gout533;
    Ciphertext gout656 = ctx.eval_lut(&gin656, and2);
    std::cerr << "\rLUT657   ";
    Ciphertext gin657 = 2 * gout613 + gout513;
    Ciphertext gout657 = ctx.eval_lut(&gin657, nand2);
    std::cerr << "\rLUT658   ";
    Ciphertext gin658 = 2 * gout461 + gout635;
    Ciphertext gout658 = ctx.eval_lut(&gin658, nand2);
    std::cerr << "\rLUT659   ";
    Ciphertext gin659 = 2 * (1 - gout638) + gout375;
    Ciphertext gout659 = ctx.eval_lut(&gin659, and2);
    std::cerr << "\rLUT660   ";
    Ciphertext gin660 = 2 * gout30 + gout615;
    Ciphertext gout660 = ctx.eval_lut(&gin660, xnor2);
    std::cerr << "\rLUT661   ";
    Ciphertext gin661 = 2 * gout648 + gout631;
    Ciphertext gout661 = ctx.eval_lut(&gin661, nand2);
    std::cerr << "\rLUT662   ";
    Ciphertext gin662 = 2 * gout438 + gout643;
    Ciphertext gout662 = ctx.eval_lut(&gin662, xor2);
    std::cerr << "\rLUT663   ";
    Ciphertext gin663 = 2 * gout629 + gout624;
    Ciphertext gout663 = ctx.eval_lut(&gin663, and2);
    std::cerr << "\rLUT664   ";
    Ciphertext gin664 = 2 * gout438 + gout643;
    Ciphertext gout664 = ctx.eval_lut(&gin664, nand2);
    std::cerr << "\rLUT665   ";
    Ciphertext gin665 = 2 * gout648 + gout631;
    Ciphertext gout665 = ctx.eval_lut(&gin665, xor2);
    std::cerr << "\rLUT666   ";
    Ciphertext gin666 = 2 * gout564 + gout626;
    Ciphertext gout666 = ctx.eval_lut(&gin666, and2);
    std::cerr << "\rLUT667   ";
    Ciphertext gin667 = 2 * gout637 + gout584;
    Ciphertext gout667 = ctx.eval_lut(&gin667, xnor2);
    std::cerr << "\rLUT668   ";
    Ciphertext gin668 = 2 * gout636 + gout486;
    Ciphertext gout668 = ctx.eval_lut(&gin668, xnor2);
    std::cerr << "\rLUT669   ";
    Ciphertext gin669 = 2 * gout530 + gout626;
    Ciphertext gout669 = ctx.eval_lut(&gin669, and2);
    std::cerr << "\rLUT670   ";
    Ciphertext gin670 = 2 * (1 - gout636) + gout486;
    Ciphertext gout670 = ctx.eval_lut(&gin670, and2);
    std::cerr << "\rLUT671   ";
    Ciphertext gin671 = 2 * (1 - gout646) + gout577;
    Ciphertext gout671 = ctx.eval_lut(&gin671, and2);
    std::cerr << "\rLUT672   ";
    Ciphertext gin672 = 2 * gout66 + gout660;
    Ciphertext gout672 = ctx.eval_lut(&gin672, xnor2);
    std::cerr << "\rLUT673   ";
    Ciphertext gin673 = 2 * gout658 + gout545;
    Ciphertext gout673 = ctx.eval_lut(&gin673, xnor2);
    std::cerr << "\rLUT674   ";
    Ciphertext gin674 = 2 * gout663 + gout644;
    Ciphertext gout674 = ctx.eval_lut(&gin674, or2);
    std::cerr << "\rLUT675   ";
    Ciphertext gin675 = 2 * gout626 + gout603;
    Ciphertext gout675 = ctx.eval_lut(&gin675, xor2);
    std::cerr << "\rLUT676   ";
    Ciphertext gin676 = 2 * gout658 + gout545;
    Ciphertext gout676 = ctx.eval_lut(&gin676, nand2);
    std::cerr << "\rLUT677   ";
    Ciphertext gin677 = 2 * gout562 + gout675;
    Ciphertext gout677 = ctx.eval_lut(&gin677, xnor2);
    std::cerr << "\rINV678   ";
    Ciphertext gout678 = 1 - gout672;

    std::cerr << "\rLUT679   ";
    Ciphertext gin679 = 2 * ct_x9 + gout660;
    Ciphertext gout679 = ctx.eval_lut(&gin679, nand2);
    std::cerr << "\rLUT680   ";
    Ciphertext gin680 = 2 * gout633 + gout425;
    Ciphertext gout680 = ctx.eval_lut(&gin680, xor2);
    std::cerr << "\rLUT681   ";
    Ciphertext gin681 = 2 * gout562 + gout675;
    Ciphertext gout681 = ctx.eval_lut(&gin681, nand2);
    std::cerr << "\rLUT682   ";
    Ciphertext gin682 = 2 * ct_x9 + gout660;
    Ciphertext gout682 = ctx.eval_lut(&gin682, or2);
    std::cerr << "\rLUT683   ";
    Ciphertext gin683 = 2 * gout633 + gout425;
    Ciphertext gout683 = ctx.eval_lut(&gin683, nand2);
    std::cerr << "\rLUT684   ";
    Ciphertext gin684 = 2 * gout131 + gout659;
    Ciphertext gout684 = ctx.eval_lut(&gin684, xnor2);
    std::cerr << "\rLUT685   ";
    Ciphertext gin685 = 2 * gout651 + gout639;
    Ciphertext gout685 = ctx.eval_lut(&gin685, xnor2);
    std::cerr << "\rLUT686   ";
    Ciphertext gin686 = 2 * gout620 + gout670;
    Ciphertext gout686 = ctx.eval_lut(&gin686, or2);
    std::cerr << "\rLUT687   ";
    Ciphertext gin687 = 2 * gout645 + gout664;
    Ciphertext gout687 = ctx.eval_lut(&gin687, nand2);
    std::cerr << "\rLUT688   ";
    Ciphertext gin688 = 2 * gout685 + gout547;
    Ciphertext gout688 = ctx.eval_lut(&gin688, xnor2);
    std::cerr << "\rLUT689   ";
    Ciphertext gin689 = 2 * gout686 + gout609;
    Ciphertext gout689 = ctx.eval_lut(&gin689, xnor2);
    std::cerr << "\rLUT690   ";
    Ciphertext gin690 = 2 * gout674 + gout688;
    Ciphertext gout690 = ctx.eval_lut(&gin690, nand2);
    std::cerr << "\rLUT691   ";
    Ciphertext gin691 = 2 * gout432 + gout665;
    Ciphertext gout691 = ctx.eval_lut(&gin691, nand2);
    std::cerr << "\rLUT692   ";
    Ciphertext gin692 = 2 * gout416 + gout655;
    Ciphertext gout692 = ctx.eval_lut(&gin692, nand2);
    std::cerr << "\rLUT693   ";
    Ciphertext gin693 = 2 * gout686 + gout609;
    Ciphertext gout693 = ctx.eval_lut(&gin693, and2);
    std::cerr << "\rLUT694   ";
    Ciphertext gin694 = 2 * gout621 + gout679;
    Ciphertext gout694 = ctx.eval_lut(&gin694, nand2);
    std::cerr << "\rLUT695   ";
    Ciphertext gin695 = 2 * gout661 + gout691;
    Ciphertext gout695 = ctx.eval_lut(&gin695, nand2);
    std::cerr << "\rLUT696   ";
    Ciphertext gin696 = 2 * (1 - gout685) + gout547;
    Ciphertext gout696 = ctx.eval_lut(&gin696, and2);
    std::cerr << "\rLUT697   ";
    Ciphertext gin697 = 2 * gout131 + gout638;
    Ciphertext gout697 = ctx.eval_lut(&gin697, nand2);
    std::cerr << "\rLUT698   ";
    Ciphertext gin698 = 2 * gout375 + gout697;
    Ciphertext gout698 = ctx.eval_lut(&gin698, and2);
    std::cerr << "\rLUT699   ";
    Ciphertext gin699 = 2 * gout687 + gout667;
    Ciphertext gout699 = ctx.eval_lut(&gin699, xor2);
    std::cerr << "\rLUT700   ";
    Ciphertext gin700 = 2 * gout647 + gout650;
    Ciphertext gout700 = ctx.eval_lut(&gin700, xnor2);
    std::cerr << "\rLUT701   ";
    Ciphertext gin701 = 2 * gout432 + gout665;
    Ciphertext gout701 = ctx.eval_lut(&gin701, xor2);
    std::cerr << "\rINV702   ";
    Ciphertext gout702 = 1 - gout666;

    std::cerr << "\rLUT703   ";
    Ciphertext gin703 = 2 * gout523 + gout676;
    Ciphertext gout703 = ctx.eval_lut(&gin703, nand2);
    std::cerr << "\rLUT704   ";
    Ciphertext gin704 = 2 * gout653 + gout696;
    Ciphertext gout704 = ctx.eval_lut(&gin704, or2);
    std::cerr << "\rLUT705   ";
    Ciphertext gin705 = 2 * gout641 + gout607;
    Ciphertext gout705 = ctx.eval_lut(&gin705, and2);
    std::cerr << "\rLUT706   ";
    Ciphertext gin706 = 2 * gout618 + gout668;
    Ciphertext gout706 = ctx.eval_lut(&gin706, xnor2);
    std::cerr << "\rLUT707   ";
    Ciphertext gin707 = 2 * gout646 + gout577;
    Ciphertext gout707 = ctx.eval_lut(&gin707, xnor2);
    std::cerr << "\rINV708   ";
    Ciphertext gout708 = 1 - gout669;

    std::cerr << "\rLUT709   ";
    Ciphertext gin709 = 2 * gout618 + gout668;
    Ciphertext gout709 = ctx.eval_lut(&gin709, and2);
    std::cerr << "\rIMUX710  ";
    Ciphertext gin710A = 2 * gout509 + gout387;
    Ciphertext gout710A = ctx.eval_lut(&gin710A, and2);
    Ciphertext gin710B = 2 * gout673 + (1 - gout387);
    Ciphertext gout710B = ctx.eval_lut(&gin710B, and2);
    Ciphertext gout710 = gout710A + gout710B;
    std::cerr << "\rLUT711   ";
    Ciphertext gin711 = 2 * gout706 + gout585;
    Ciphertext gout711 = ctx.eval_lut(&gin711, xnor2);
    std::cerr << "\rLUT712   ";
    Ciphertext gin712 = 2 * gout647 + gout634;
    Ciphertext gout712 = ctx.eval_lut(&gin712, and2);
    std::cerr << "\rLUT713   ";
    Ciphertext gin713 = 2 * gout687 + gout667;
    Ciphertext gout713 = ctx.eval_lut(&gin713, and2);
    std::cerr << "\rLUT714   ";
    Ciphertext gin714 = 2 * (1 - gout706) + gout585;
    Ciphertext gout714 = ctx.eval_lut(&gin714, and2);
    std::cerr << "\rLUT715   ";
    Ciphertext gin715 = 2 * gout695 + gout662;
    Ciphertext gout715 = ctx.eval_lut(&gin715, and2);
    std::cerr << "\rINV716   ";
    Ciphertext gout716 = 1 - gout681;

    std::cerr << "\rLUT717   ";
    Ciphertext gin717 = 2 * (1 - gout698) + gout458;
    Ciphertext gout717 = ctx.eval_lut(&gin717, and2);
    std::cerr << "\rLUT718   ";
    Ciphertext gin718 = 2 * gout518 + gout680;
    Ciphertext gout718 = ctx.eval_lut(&gin718, nand2);
    std::cerr << "\rLUT719   ";
    Ciphertext gin719 = 2 * gout518 + gout680;
    Ciphertext gout719 = ctx.eval_lut(&gin719, xor2);
    std::cerr << "\rLUT720   ";
    Ciphertext gin720 = 2 * ct_x10 + gout684;
    Ciphertext gout720 = ctx.eval_lut(&gin720, nand2);
    std::cerr << "\rLUT721   ";
    Ciphertext gin721 = 2 * gout689 + gout581;
    Ciphertext gout721 = ctx.eval_lut(&gin721, xnor2);
    std::cerr << "\rLUT722   ";
    Ciphertext gin722 = 2 * gout18 + gout684;
    Ciphertext gout722 = ctx.eval_lut(&gin722, xnor2);
    std::cerr << "\rLUT723   ";
    Ciphertext gin723 = 2 * gout695 + gout662;
    Ciphertext gout723 = ctx.eval_lut(&gin723, xor2);
    std::cerr << "\rLUT724   ";
    Ciphertext gin724 = 2 * (1 - gout689) + gout581;
    Ciphertext gout724 = ctx.eval_lut(&gin724, and2);
    std::cerr << "\rLUT725   ";
    Ciphertext gin725 = 2 * gout705 + gout671;
    Ciphertext gout725 = ctx.eval_lut(&gin725, or2);
    std::cerr << "\rLUT726   ";
    Ciphertext gin726 = 2 * gout416 + gout710;
    Ciphertext gout726 = ctx.eval_lut(&gin726, nand2);
    std::cerr << "\rLUT727   ";
    Ciphertext gin727 = 2 * gout715 + gout699;
    Ciphertext gout727 = ctx.eval_lut(&gin727, and2);
    std::cerr << "\rLUT728   ";
    Ciphertext gin728 = 2 * ct_y11 + gout698;
    Ciphertext gout728 = ctx.eval_lut(&gin728, xnor2);
    std::cerr << "\rLUT729   ";
    Ciphertext gin729 = 2 * gout715 + gout699;
    Ciphertext gout729 = ctx.eval_lut(&gin729, xor2);
    std::cerr << "\rINV730   ";
    Ciphertext gout730 = 1 - gout713;

    std::cerr << "\rLUT731   ";
    Ciphertext gin731 = 2 * gout674 + gout688;
    Ciphertext gout731 = ctx.eval_lut(&gin731, xor2);
    std::cerr << "\rLUT732   ";
    Ciphertext gin732 = 2 * gout693 + gout724;
    Ciphertext gout732 = ctx.eval_lut(&gin732, or2);
    std::cerr << "\rLUT733   ";
    Ciphertext gin733 = 2 * gout682 + gout694;
    Ciphertext gout733 = ctx.eval_lut(&gin733, nand2);
    std::cerr << "\rLUT734   ";
    Ciphertext gin734 = 2 * gout725 + gout711;
    Ciphertext gout734 = ctx.eval_lut(&gin734, xor2);
    std::cerr << "\rLUT735   ";
    Ciphertext gin735 = 2 * gout703 + gout520;
    Ciphertext gout735 = ctx.eval_lut(&gin735, xnor2);
    std::cerr << "\rLUT736   ";
    Ciphertext gin736 = 2 * gout708 + gout700;
    Ciphertext gout736 = ctx.eval_lut(&gin736, xnor2);
    std::cerr << "\rLUT737   ";
    Ciphertext gin737 = 2 * gout669 + gout700;
    Ciphertext gout737 = ctx.eval_lut(&gin737, and2);
    std::cerr << "\rINV738   ";
    Ciphertext gout738 = 1 - gout712;

    std::cerr << "\rLUT739   ";
    Ciphertext gin739 = 2 * gout725 + gout711;
    Ciphertext gout739 = ctx.eval_lut(&gin739, nand2);
    std::cerr << "\rLUT740   ";
    Ciphertext gin740 = 2 * gout704 + gout707;
    Ciphertext gout740 = ctx.eval_lut(&gin740, xor2);
    std::cerr << "\rIMUX741  ";
    Ciphertext gin741A = 2 * gout735 + gout428;
    Ciphertext gout741A = ctx.eval_lut(&gin741A, and2);
    Ciphertext gin741B = 2 * gout567 + (1 - gout428);
    Ciphertext gout741B = ctx.eval_lut(&gin741B, and2);
    Ciphertext gout741 = gout741A + gout741B;
    std::cerr << "\rLUT742   ";
    Ciphertext gin742 = 2 * gout703 + gout520;
    Ciphertext gout742 = ctx.eval_lut(&gin742, nand2);
    std::cerr << "\rLUT743   ";
    Ciphertext gin743 = 2 * gout712 + gout701;
    Ciphertext gout743 = ctx.eval_lut(&gin743, and2);
    std::cerr << "\rLUT744   ";
    Ciphertext gin744 = 2 * gout704 + gout707;
    Ciphertext gout744 = ctx.eval_lut(&gin744, nand2);
    std::cerr << "\rLUT745   ";
    Ciphertext gin745 = 2 * gout709 + gout714;
    Ciphertext gout745 = ctx.eval_lut(&gin745, or2);
    std::cerr << "\rLUT746   ";
    Ciphertext gin746 = 2 * ct_y12 + gout717;
    Ciphertext gout746 = ctx.eval_lut(&gin746, xnor2);
    std::cerr << "\rLUT747   ";
    Ciphertext gin747 = 2 * (1 - gout628) + gout719;
    Ciphertext gout747 = ctx.eval_lut(&gin747, and2);
    std::cerr << "\rLUT748   ";
    Ciphertext gin748 = 2 * gout628 + gout719;
    Ciphertext gout748 = ctx.eval_lut(&gin748, xnor2);
    std::cerr << "\rLUT749   ";
    Ciphertext gin749 = 2 * gout683 + gout718;
    Ciphertext gout749 = ctx.eval_lut(&gin749, nand2);
    std::cerr << "\rLUT750   ";
    Ciphertext gin750 = 2 * gout745 + gout721;
    Ciphertext gout750 = ctx.eval_lut(&gin750, xor2);
    std::cerr << "\rLUT751   ";
    Ciphertext gin751 = 2 * gout515 + gout742;
    Ciphertext gout751 = ctx.eval_lut(&gin751, nand2);
    std::cerr << "\rLUT752   ";
    Ciphertext gin752 = 2 * gout717 + gout422;
    Ciphertext gout752 = ctx.eval_lut(&gin752, nand2);
    std::cerr << "\rLUT753   ";
    Ciphertext gin753 = 2 * gout53 + gout728;
    Ciphertext gout753 = ctx.eval_lut(&gin753, nand2);
    std::cerr << "\rLUT754   ";
    Ciphertext gin754 = 2 * gout159 + gout746;
    Ciphertext gout754 = ctx.eval_lut(&gin754, xnor2);
    std::cerr << "\rLUT755   ";
    Ciphertext gin755 = 2 * gout749 + gout342;
    Ciphertext gout755 = ctx.eval_lut(&gin755, xnor2);
    std::cerr << "\rLUT756   ";
    Ciphertext gin756 = 2 * gout732 + gout748;
    Ciphertext gout756 = ctx.eval_lut(&gin756, xnor2);
    std::cerr << "\rLUT757   ";
    Ciphertext gin757 = 2 * gout756 + gout580;
    Ciphertext gout757 = ctx.eval_lut(&gin757, xnor2);
    std::cerr << "\rLUT758   ";
    Ciphertext gin758 = 2 * gout751 + gout544;
    Ciphertext gout758 = ctx.eval_lut(&gin758, xnor2);
    std::cerr << "\rLUT759   ";
    Ciphertext gin759 = 2 * gout732 + gout748;
    Ciphertext gout759 = ctx.eval_lut(&gin759, nand2);
    std::cerr << "\rLUT760   ";
    Ciphertext gin760 = 2 * gout53 + gout728;
    Ciphertext gout760 = ctx.eval_lut(&gin760, or2);
    std::cerr << "\rLUT761   ";
    Ciphertext gin761 = 2 * ct_x11 + gout728;
    Ciphertext gout761 = ctx.eval_lut(&gin761, xnor2);
    std::cerr << "\rLUT762   ";
    Ciphertext gin762 = 2 * gout416 + gout741;
    Ciphertext gout762 = ctx.eval_lut(&gin762, nand2);
    std::cerr << "\rLUT763   ";
    Ciphertext gin763 = 2 * gout666 + gout736;
    Ciphertext gout763 = ctx.eval_lut(&gin763, nand2);
    std::cerr << "\rLUT764   ";
    Ciphertext gin764 = 2 * gout751 + gout544;
    Ciphertext gout764 = ctx.eval_lut(&gin764, nand2);
    std::cerr << "\rLUT765   ";
    Ciphertext gin765 = 2 * gout743 + gout723;
    Ciphertext gout765 = ctx.eval_lut(&gin765, and2);
    std::cerr << "\rLUT766   ";
    Ciphertext gin766 = 2 * gout755 + gout649;
    Ciphertext gout766 = ctx.eval_lut(&gin766, xnor2);
    std::cerr << "\rLUT767   ";
    Ciphertext gin767 = 2 * gout549 + gout731;
    Ciphertext gout767 = ctx.eval_lut(&gin767, xor2);
    std::cerr << "\rLUT768   ";
    Ciphertext gin768 = 2 * gout608 + gout734;
    Ciphertext gout768 = ctx.eval_lut(&gin768, nand2);
    std::cerr << "\rLUT769   ";
    Ciphertext gin769 = 2 * (1 - gout756) + gout580;
    Ciphertext gout769 = ctx.eval_lut(&gin769, and2);
    std::cerr << "\rLUT770   ";
    Ciphertext gin770 = 2 * gout549 + gout731;
    Ciphertext gout770 = ctx.eval_lut(&gin770, nand2);
    std::cerr << "\rLUT771   ";
    Ciphertext gin771 = 2 * gout702 + gout736;
    Ciphertext gout771 = ctx.eval_lut(&gin771, xnor2);
    std::cerr << "\rINV772   ";
    Ciphertext gout772 = 1 - gout737;

    std::cerr << "\rLUT773   ";
    Ciphertext gin773 = 2 * gout765 + gout729;
    Ciphertext gout773 = ctx.eval_lut(&gin773, xnor2);
    std::cerr << "\rLUT774   ";
    Ciphertext gin774 = 2 * gout583 + gout764;
    Ciphertext gout774 = ctx.eval_lut(&gin774, nand2);
    std::cerr << "\rLUT775   ";
    Ciphertext gin775 = 2 * gout608 + gout734;
    Ciphertext gout775 = ctx.eval_lut(&gin775, xor2);
    std::cerr << "\rLUT776   ";
    Ciphertext gin776 = 2 * gout656 + gout747;
    Ciphertext gout776 = ctx.eval_lut(&gin776, or2);
    std::cerr << "\rLUT777   ";
    Ciphertext gin777 = 2 * ct_x12 + gout746;
    Ciphertext gout777 = ctx.eval_lut(&gin777, nand2);
    std::cerr << "\rLUT778   ";
    Ciphertext gin778 = 2 * gout738 + gout701;
    Ciphertext gout778 = ctx.eval_lut(&gin778, xnor2);
    std::cerr << "\rINV779   ";
    Ciphertext gout779 = 1 - gout761;

    std::cerr << "\rLUT780   ";
    Ciphertext gin780 = 2 * gout774 + gout599;
    Ciphertext gout780 = ctx.eval_lut(&gin780, xnor2);
    std::cerr << "\rINV781   ";
    Ciphertext gout781 = 1 - gout743;

    std::cerr << "\rLUT782   ";
    Ciphertext gin782 = 2 * gout552 + gout740;
    Ciphertext gout782 = ctx.eval_lut(&gin782, nand2);
    std::cerr << "\rLUT783   ";
    Ciphertext gin783 = 2 * gout552 + gout740;
    Ciphertext gout783 = ctx.eval_lut(&gin783, xor2);
    std::cerr << "\rLUT784   ";
    Ciphertext gin784 = 2 * gout619 + gout750;
    Ciphertext gout784 = ctx.eval_lut(&gin784, nand2);
    std::cerr << "\rLUT785   ";
    Ciphertext gin785 = 2 * gout765 + gout729;
    Ciphertext gout785 = ctx.eval_lut(&gin785, nand2);
    std::cerr << "\rLUT786   ";
    Ciphertext gin786 = 2 * gout745 + gout721;
    Ciphertext gout786 = ctx.eval_lut(&gin786, nand2);
    std::cerr << "\rLUT787   ";
    Ciphertext gin787 = 2 * gout619 + gout750;
    Ciphertext gout787 = ctx.eval_lut(&gin787, xor2);
    std::cerr << "\rLUT788   ";
    Ciphertext gin788 = 2 * ct_y13 + gout752;
    Ciphertext gout788 = ctx.eval_lut(&gin788, xnor2);
    std::cerr << "\rLUT789   ";
    Ciphertext gin789 = 2 * gout752 + gout444;
    Ciphertext gout789 = ctx.eval_lut(&gin789, or2);
    std::cerr << "\rLUT790   ";
    Ciphertext gin790 = 2 * gout739 + gout768;
    Ciphertext gout790 = ctx.eval_lut(&gin790, nand2);
    std::cerr << "\rLUT791   ";
    Ciphertext gin791 = 2 * gout157 + gout788;
    Ciphertext gout791 = ctx.eval_lut(&gin791, nand2);
    std::cerr << "\rIMUX792  ";
    Ciphertext gin792A = 2 * gout677 + gout387;
    Ciphertext gout792A = ctx.eval_lut(&gin792A, and2);
    Ciphertext gin792B = 2 * gout758 + (1 - gout387);
    Ciphertext gout792B = ctx.eval_lut(&gin792B, and2);
    Ciphertext gout792 = gout792A + gout792B;
    std::cerr << "\rLUT793   ";
    Ciphertext gin793 = 2 * gout730 + gout767;
    Ciphertext gout793 = ctx.eval_lut(&gin793, xnor2);
    std::cerr << "\rLUT794   ";
    Ciphertext gin794 = 2 * gout157 + gout788;
    Ciphertext gout794 = ctx.eval_lut(&gin794, or2);
    std::cerr << "\rLUT795   ";
    Ciphertext gin795 = 2 * gout720 + gout760;
    Ciphertext gout795 = ctx.eval_lut(&gin795, nand2);
    std::cerr << "\rLUT796   ";
    Ciphertext gin796 = 2 * gout744 + gout782;
    Ciphertext gout796 = ctx.eval_lut(&gin796, nand2);
    std::cerr << "\rLUT797   ";
    Ciphertext gin797 = 2 * gout759 + (1 - gout769);
    Ciphertext gout797 = ctx.eval_lut(&gin797, and2);
    std::cerr << "\rLUT798   ";
    Ciphertext gin798 = 2 * gout713 + gout767;
    Ciphertext gout798 = ctx.eval_lut(&gin798, and2);
    std::cerr << "\rLUT799   ";
    Ciphertext gin799 = 2 * gout716 + gout771;
    Ciphertext gout799 = ctx.eval_lut(&gin799, xnor2);
    std::cerr << "\rLUT800   ";
    Ciphertext gin800 = 2 * gout772 + gout778;
    Ciphertext gout800 = ctx.eval_lut(&gin800, xnor2);
    std::cerr << "\rLUT801   ";
    Ciphertext gin801 = 2 * gout416 + gout792;
    Ciphertext gout801 = ctx.eval_lut(&gin801, nand2);
    std::cerr << "\rLUT802   ";
    Ciphertext gin802 = 2 * gout716 + gout771;
    Ciphertext gout802 = ctx.eval_lut(&gin802, nand2);
    std::cerr << "\rLUT803   ";
    Ciphertext gin803 = 2 * gout690 + gout770;
    Ciphertext gout803 = ctx.eval_lut(&gin803, nand2);
    std::cerr << "\rLUT804   ";
    Ciphertext gin804 = 2 * gout790 + gout787;
    Ciphertext gout804 = ctx.eval_lut(&gin804, xor2);
    std::cerr << "\rLUT805   ";
    Ciphertext gin805 = 2 * gout786 + gout784;
    Ciphertext gout805 = ctx.eval_lut(&gin805, nand2);
    std::cerr << "\rLUT806   ";
    Ciphertext gin806 = 2 * gout790 + gout787;
    Ciphertext gout806 = ctx.eval_lut(&gin806, and2);
    std::cerr << "\rLUT807   ";
    Ciphertext gin807 = 2 * ct_x13 + gout788;
    Ciphertext gout807 = ctx.eval_lut(&gin807, xnor2);
    std::cerr << "\rLUT808   ";
    Ciphertext gin808 = 2 * gout737 + gout778;
    Ciphertext gout808 = ctx.eval_lut(&gin808, nand2);
    std::cerr << "\rLUT809   ";
    Ciphertext gin809 = 2 * gout777 + gout794;
    Ciphertext gout809 = ctx.eval_lut(&gin809, nand2);
    std::cerr << "\rLUT810   ";
    Ciphertext gin810 = 2 * gout776 + gout766;
    Ciphertext gout810 = ctx.eval_lut(&gin810, xnor2);
    std::cerr << "\rLUT811   ";
    Ciphertext gin811 = 2 * gout774 + gout599;
    Ciphertext gout811 = ctx.eval_lut(&gin811, nand2);
    std::cerr << "\rLUT812   ";
    Ciphertext gin812 = 2 * gout176 + gout789;
    Ciphertext gout812 = ctx.eval_lut(&gin812, xnor2);
    std::cerr << "\rIMUX813  ";
    Ciphertext gin813A = 2 * gout799 + gout387;
    Ciphertext gout813A = ctx.eval_lut(&gin813A, and2);
    Ciphertext gin813B = 2 * gout780 + (1 - gout387);
    Ciphertext gout813B = ctx.eval_lut(&gin813B, and2);
    Ciphertext gout813 = gout813A + gout813B;
    std::cerr << "\rLUT814   ";
    Ciphertext gin814 = 2 * gout789 + gout435;
    Ciphertext gout814 = ctx.eval_lut(&gin814, or2);
    std::cerr << "\rLUT815   ";
    Ciphertext gin815 = 2 * gout808 + (1 - gout723);
    Ciphertext gout815 = ctx.eval_lut(&gin815, or2);
    std::cerr << "\rLUT816   ";
    Ciphertext gin816 = 2 * gout781 + gout808;
    Ciphertext gout816 = ctx.eval_lut(&gin816, and2);
    std::cerr << "\rLUT817   ";
    Ciphertext gin817 = 2 * gout46 + gout812;
    Ciphertext gout817 = ctx.eval_lut(&gin817, xnor2);
    std::cerr << "\rLUT818   ";
    Ciphertext gin818 = 2 * gout796 + gout775;
    Ciphertext gout818 = ctx.eval_lut(&gin818, and2);
    std::cerr << "\rINV819   ";
    Ciphertext gout819 = 1 - gout807;

    std::cerr << "\rLUT820   ";
    Ciphertext gin820 = 2 * gout727 + gout793;
    Ciphertext gout820 = ctx.eval_lut(&gin820, nand2);
    std::cerr << "\rLUT821   ";
    Ciphertext gin821 = 2 * gout727 + gout793;
    Ciphertext gout821 = ctx.eval_lut(&gin821, xor2);
    std::cerr << "\rLUT822   ";
    Ciphertext gin822 = 2 * gout796 + gout775;
    Ciphertext gout822 = ctx.eval_lut(&gin822, xor2);
    std::cerr << "\rLUT823   ";
    Ciphertext gin823 = 2 * gout805 + gout757;
    Ciphertext gout823 = ctx.eval_lut(&gin823, and2);
    std::cerr << "\rLUT824   ";
    Ciphertext gin824 = 2 * gout416 + gout813;
    Ciphertext gout824 = ctx.eval_lut(&gin824, nand2);
    std::cerr << "\rLUT825   ";
    Ciphertext gin825 = 2 * gout805 + gout757;
    Ciphertext gout825 = ctx.eval_lut(&gin825, xor2);
    std::cerr << "\rINV826   ";
    Ciphertext gout826 = 1 - gout798;

    std::cerr << "\rLUT827   ";
    Ciphertext gin827 = 2 * gout753 + gout795;
    Ciphertext gout827 = ctx.eval_lut(&gin827, nand2);
    std::cerr << "\rLUT828   ";
    Ciphertext gin828 = 2 * gout791 + gout809;
    Ciphertext gout828 = ctx.eval_lut(&gin828, nand2);
    std::cerr << "\rLUT829   ";
    Ciphertext gin829 = 2 * gout803 + gout783;
    Ciphertext gout829 = ctx.eval_lut(&gin829, and2);
    std::cerr << "\rLUT830   ";
    Ciphertext gin830 = 2 * gout606 + gout811;
    Ciphertext gout830 = ctx.eval_lut(&gin830, nand2);
    std::cerr << "\rLUT831   ";
    Ciphertext gin831 = 2 * gout617 + gout823;
    Ciphertext gout831 = ctx.eval_lut(&gin831, xnor2);
    std::cerr << "\rLUT832   ";
    Ciphertext gin832 = 2 * gout763 + gout802;
    Ciphertext gout832 = ctx.eval_lut(&gin832, nand2);
    std::cerr << "\rLUT833   ";
    Ciphertext gin833 = 2 * gout830 + gout654;
    Ciphertext gout833 = ctx.eval_lut(&gin833, nand2);
    std::cerr << "\rLUT834   ";
    Ciphertext gin834 = 2 * gout814 + gout43;
    Ciphertext gout834 = ctx.eval_lut(&gin834, xnor2);
    std::cerr << "\rLUT835   ";
    Ciphertext gin835 = 2 * gout803 + gout783;
    Ciphertext gout835 = ctx.eval_lut(&gin835, xor2);
    std::cerr << "\rLUT836   ";
    Ciphertext gin836 = 2 * gout818 + gout804;
    Ciphertext gout836 = ctx.eval_lut(&gin836, xor2);
    std::cerr << "\rLUT837   ";
    Ciphertext gin837 = 2 * gout829 + gout822;
    Ciphertext gout837 = ctx.eval_lut(&gin837, xor2);
    std::cerr << "\rLUT838   ";
    Ciphertext gin838 = 2 * gout818 + gout804;
    Ciphertext gout838 = ctx.eval_lut(&gin838, nand2);
    std::cerr << "\rLUT839   ";
    Ciphertext gin839 = 2 * gout830 + gout654;
    Ciphertext gout839 = ctx.eval_lut(&gin839, xnor2);
    std::cerr << "\rLUT840   ";
    Ciphertext gin840 = 2 * gout829 + gout822;
    Ciphertext gout840 = ctx.eval_lut(&gin840, nand2);
    std::cerr << "\rLUT841   ";
    Ciphertext gin841 = 2 * gout797 + gout810;
    Ciphertext gout841 = ctx.eval_lut(&gin841, xnor2);
    std::cerr << "\rLUT842   ";
    Ciphertext gin842 = 2 * gout806 + gout825;
    Ciphertext gout842 = ctx.eval_lut(&gin842, nand2);
    std::cerr << "\rLUT843   ";
    Ciphertext gin843 = 2 * gout800 + gout832;
    Ciphertext gout843 = ctx.eval_lut(&gin843, xnor2);
    std::cerr << "\rLUT844   ";
    Ciphertext gin844 = 2 * gout800 + gout832;
    Ciphertext gout844 = ctx.eval_lut(&gin844, and2);
    std::cerr << "\rLUT845   ";
    Ciphertext gin845 = 2 * ct_x14 + gout812;
    Ciphertext gout845 = ctx.eval_lut(&gin845, nand2);
    std::cerr << "\rLUT846   ";
    Ciphertext gin846 = 2 * gout798 + gout835;
    Ciphertext gout846 = ctx.eval_lut(&gin846, nand2);
    std::cerr << "\rLUT847   ";
    Ciphertext gin847 = 2 * gout806 + gout825;
    Ciphertext gout847 = ctx.eval_lut(&gin847, xor2);
    std::cerr << "\rLUT848   ";
    Ciphertext gin848 = 2 * gout826 + gout835;
    Ciphertext gout848 = ctx.eval_lut(&gin848, xnor2);
    std::cerr << "\rLUT849   ";
    Ciphertext gin849 = 2 * gout627 + gout833;
    Ciphertext gout849 = ctx.eval_lut(&gin849, nand2);
    std::cerr << "\rLUT850   ";
    Ciphertext gin850 = 2 * gout723 + gout816;
    Ciphertext gout850 = ctx.eval_lut(&gin850, xnor2);
    std::cerr << "\rLUT851   ";
    Ciphertext gin851 = 2 * gout844 + gout850;
    Ciphertext gout851 = ctx.eval_lut(&gin851, nand2);
    std::cerr << "\rLUT852   ";
    Ciphertext gin852 = 2 * gout841 + gout831;
    Ciphertext gout852 = ctx.eval_lut(&gin852, xnor2);
    std::cerr << "\rLUT853   ";
    Ciphertext gin853 = 2 * gout849 + gout640;
    Ciphertext gout853 = ctx.eval_lut(&gin853, nand2);
    std::cerr << "\rIMUX854  ";
    Ciphertext gin854A = 2 * gout843 + gout387;
    Ciphertext gout854A = ctx.eval_lut(&gin854A, and2);
    Ciphertext gin854B = 2 * gout839 + (1 - gout387);
    Ciphertext gout854B = ctx.eval_lut(&gin854B, and2);
    Ciphertext gout854 = gout854A + gout854B;
    std::cerr << "\rLUT855   ";
    Ciphertext gin855 = 2 * gout844 + gout850;
    Ciphertext gout855 = ctx.eval_lut(&gin855, xnor2);
    std::cerr << "\rLUT856   ";
    Ciphertext gin856 = 2 * gout621 + gout853;
    Ciphertext gout856 = ctx.eval_lut(&gin856, nand2);
    std::cerr << "\rLUT857   ";
    Ciphertext gin857 = 2 * gout416 + gout854;
    Ciphertext gout857 = ctx.eval_lut(&gin857, nand2);
    std::cerr << "\rLUT858   ";
    Ciphertext gin858 = 2 * gout853 + gout678;
    Ciphertext gout858 = ctx.eval_lut(&gin858, or2);
    std::cerr << "\rLUT859   ";
    Ciphertext gin859 = 2 * gout849 + gout640;
    Ciphertext gout859 = ctx.eval_lut(&gin859, xnor2);
    std::cerr << "\rLUT860   ";
    Ciphertext gin860 = 2 * gout851 + gout815;
    Ciphertext gout860 = ctx.eval_lut(&gin860, and2);
    std::cerr << "\rIMUX861  ";
    Ciphertext gin861A = 2 * gout855 + gout387;
    Ciphertext gout861A = ctx.eval_lut(&gin861A, and2);
    Ciphertext gin861B = 2 * gout859 + (1 - gout387);
    Ciphertext gout861B = ctx.eval_lut(&gin861B, and2);
    Ciphertext gout861 = gout861A + gout861B;
    std::cerr << "\rLUT862   ";
    Ciphertext gin862 = 2 * gout416 + gout861;
    Ciphertext gout862 = ctx.eval_lut(&gin862, nand2);
    std::cerr << "\rLUT863   ";
    Ciphertext gin863 = 2 * gout856 + gout672;
    Ciphertext gout863 = ctx.eval_lut(&gin863, nand2);
    std::cerr << "\rLUT864   ";
    Ciphertext gin864 = 2 * gout856 + gout672;
    Ciphertext gout864 = ctx.eval_lut(&gin864, or2);
    std::cerr << "\rLUT865   ";
    Ciphertext gin865 = 2 * gout858 + gout733;
    Ciphertext gout865 = ctx.eval_lut(&gin865, nand2);
    std::cerr << "\rLUT866   ";
    Ciphertext gin866 = 2 * gout860 + gout773;
    Ciphertext gout866 = ctx.eval_lut(&gin866, or2);
    std::cerr << "\rLUT867   ";
    Ciphertext gin867 = 2 * gout785 + gout866;
    Ciphertext gout867 = ctx.eval_lut(&gin867, nand2);
    std::cerr << "\rLUT868   ";
    Ciphertext gin868 = 2 * gout860 + gout773;
    Ciphertext gout868 = ctx.eval_lut(&gin868, nand2);
    std::cerr << "\rLUT869   ";
    Ciphertext gin869 = 2 * gout865 + gout722;
    Ciphertext gout869 = ctx.eval_lut(&gin869, xnor2);
    std::cerr << "\rLUT870   ";
    Ciphertext gin870 = 2 * gout867 + gout821;
    Ciphertext gout870 = ctx.eval_lut(&gin870, nand2);
    std::cerr << "\rLUT871   ";
    Ciphertext gin871 = 2 * gout428 + gout864;
    Ciphertext gout871 = ctx.eval_lut(&gin871, and2);
    std::cerr << "\rLUT872   ";
    Ciphertext gin872 = 2 * gout865 + gout722;
    Ciphertext gout872 = ctx.eval_lut(&gin872, nand2);
    std::cerr << "\rLUT873   ";
    Ciphertext gin873 = 2 * gout872 + gout779;
    Ciphertext gout873 = ctx.eval_lut(&gin873, or2);
    std::cerr << "\rLUT874   ";
    Ciphertext gin874 = 2 * gout720 + gout872;
    Ciphertext gout874 = ctx.eval_lut(&gin874, nand2);
    std::cerr << "\rLUT875   ";
    Ciphertext gin875 = 2 * gout863 + gout871;
    Ciphertext gout875 = ctx.eval_lut(&gin875, nand2);
    std::cerr << "\rLUT876   ";
    Ciphertext gin876 = 2 * gout874 + gout761;
    Ciphertext gout876 = ctx.eval_lut(&gin876, or2);
    std::cerr << "\rLUT877   ";
    Ciphertext gin877 = 2 * gout874 + gout761;
    Ciphertext gout877 = ctx.eval_lut(&gin877, nand2);
    std::cerr << "\rLUT878   ";
    Ciphertext gin878 = 2 * gout867 + gout821;
    Ciphertext gout878 = ctx.eval_lut(&gin878, xnor2);
    std::cerr << "\rLUT879   ";
    Ciphertext gin879 = 2 * gout387 + gout866;
    Ciphertext gout879 = ctx.eval_lut(&gin879, and2);
    std::cerr << "\rLUT880   ";
    Ciphertext gin880 = 2 * gout820 + gout870;
    Ciphertext gout880 = ctx.eval_lut(&gin880, nand2);
    std::cerr << "\rLUT881   ";
    Ciphertext gin881 = 2 * gout868 + gout879;
    Ciphertext gout881 = ctx.eval_lut(&gin881, nand2);
    std::cerr << "\rLUT882   ";
    Ciphertext gin882 = 2 * gout873 + gout827;
    Ciphertext gout882 = ctx.eval_lut(&gin882, nand2);
    std::cerr << "\rLUT883   ";
    Ciphertext gin883 = 2 * gout416 + gout881;
    Ciphertext gout883 = ctx.eval_lut(&gin883, and2);
    std::cerr << "\rLUT884   ";
    Ciphertext gin884 = 2 * gout428 + gout876;
    Ciphertext gout884 = ctx.eval_lut(&gin884, and2);
    std::cerr << "\rLUT885   ";
    Ciphertext gin885 = 2 * gout880 + gout848;
    Ciphertext gout885 = ctx.eval_lut(&gin885, or2);
    std::cerr << "\rLUT886   ";
    Ciphertext gin886 = 2 * gout882 + gout754;
    Ciphertext gout886 = ctx.eval_lut(&gin886, xnor2);
    std::cerr << "\rLUT887   ";
    Ciphertext gin887 = 2 * gout875 + gout883;
    Ciphertext gout887 = ctx.eval_lut(&gin887, nand2);
    std::cerr << "\rIMUX888  ";
    Ciphertext gin888A = 2 * gout878 + gout387;
    Ciphertext gout888A = ctx.eval_lut(&gin888A, and2);
    Ciphertext gin888B = 2 * gout869 + (1 - gout387);
    Ciphertext gout888B = ctx.eval_lut(&gin888B, and2);
    Ciphertext gout888 = gout888A + gout888B;
    std::cerr << "\rLUT889   ";
    Ciphertext gin889 = 2 * gout877 + gout884;
    Ciphertext gout889 = ctx.eval_lut(&gin889, nand2);
    std::cerr << "\rLUT890   ";
    Ciphertext gin890 = 2 * gout880 + gout848;
    Ciphertext gout890 = ctx.eval_lut(&gin890, nand2);
    std::cerr << "\rLUT891   ";
    Ciphertext gin891 = 2 * gout882 + gout754;
    Ciphertext gout891 = ctx.eval_lut(&gin891, nand2);
    std::cerr << "\rLUT892   ";
    Ciphertext gin892 = 2 * gout416 + gout888;
    Ciphertext gout892 = ctx.eval_lut(&gin892, nand2);
    std::cerr << "\rLUT893   ";
    Ciphertext gin893 = 2 * gout891 + gout819;
    Ciphertext gout893 = ctx.eval_lut(&gin893, or2);
    std::cerr << "\rLUT894   ";
    Ciphertext gin894 = 2 * gout387 + gout885;
    Ciphertext gout894 = ctx.eval_lut(&gin894, and2);
    std::cerr << "\rLUT895   ";
    Ciphertext gin895 = 2 * gout890 + gout894;
    Ciphertext gout895 = ctx.eval_lut(&gin895, nand2);
    std::cerr << "\rLUT896   ";
    Ciphertext gin896 = 2 * gout777 + gout891;
    Ciphertext gout896 = ctx.eval_lut(&gin896, nand2);
    std::cerr << "\rLUT897   ";
    Ciphertext gin897 = 2 * gout846 + gout890;
    Ciphertext gout897 = ctx.eval_lut(&gin897, nand2);
    std::cerr << "\rLUT898   ";
    Ciphertext gin898 = 2 * gout896 + gout807;
    Ciphertext gout898 = ctx.eval_lut(&gin898, nand2);
    std::cerr << "\rLUT899   ";
    Ciphertext gin899 = 2 * gout416 + gout895;
    Ciphertext gout899 = ctx.eval_lut(&gin899, and2);
    std::cerr << "\rLUT900   ";
    Ciphertext gin900 = 2 * gout893 + gout828;
    Ciphertext gout900 = ctx.eval_lut(&gin900, nand2);
    std::cerr << "\rLUT901   ";
    Ciphertext gin901 = 2 * gout897 + gout837;
    Ciphertext gout901 = ctx.eval_lut(&gin901, xnor2);
    std::cerr << "\rLUT902   ";
    Ciphertext gin902 = 2 * gout900 + gout817;
    Ciphertext gout902 = ctx.eval_lut(&gin902, nand2);
    std::cerr << "\rLUT903   ";
    Ciphertext gin903 = 2 * gout897 + gout837;
    Ciphertext gout903 = ctx.eval_lut(&gin903, nand2);
    std::cerr << "\rIMUX904  ";
    Ciphertext gin904A = 2 * gout901 + gout387;
    Ciphertext gout904A = ctx.eval_lut(&gin904A, and2);
    Ciphertext gin904B = 2 * gout886 + (1 - gout387);
    Ciphertext gout904B = ctx.eval_lut(&gin904B, and2);
    Ciphertext gout904 = gout904A + gout904B;
    std::cerr << "\rLUT905   ";
    Ciphertext gin905 = 2 * gout840 + gout903;
    Ciphertext gout905 = ctx.eval_lut(&gin905, nand2);
    std::cerr << "\rLUT906   ";
    Ciphertext gin906 = 2 * gout889 + gout899;
    Ciphertext gout906 = ctx.eval_lut(&gin906, nand2);
    std::cerr << "\rLUT907   ";
    Ciphertext gin907 = 2 * gout905 + gout836;
    Ciphertext gout907 = ctx.eval_lut(&gin907, or2);
    std::cerr << "\rLUT908   ";
    Ciphertext gin908 = 2 * gout900 + gout817;
    Ciphertext gout908 = ctx.eval_lut(&gin908, xnor2);
    std::cerr << "\rLUT909   ";
    Ciphertext gin909 = 2 * gout845 + gout902;
    Ciphertext gout909 = ctx.eval_lut(&gin909, and2);
    std::cerr << "\rLUT910   ";
    Ciphertext gin910 = 2 * gout416 + gout904;
    Ciphertext gout910 = ctx.eval_lut(&gin910, nand2);
    std::cerr << "\rLUT911   ";
    Ciphertext gin911 = 2 * gout896 + gout807;
    Ciphertext gout911 = ctx.eval_lut(&gin911, or2);
    std::cerr << "\rLUT912   ";
    Ciphertext gin912 = 2 * gout387 + gout907;
    Ciphertext gout912 = ctx.eval_lut(&gin912, and2);
    std::cerr << "\rLUT913   ";
    Ciphertext gin913 = 2 * gout905 + gout836;
    Ciphertext gout913 = ctx.eval_lut(&gin913, nand2);
    std::cerr << "\rLUT914   ";
    Ciphertext gin914 = 2 * gout909 + gout834;
    Ciphertext gout914 = ctx.eval_lut(&gin914, or2);
    std::cerr << "\rLUT915   ";
    Ciphertext gin915 = 2 * gout909 + gout834;
    Ciphertext gout915 = ctx.eval_lut(&gin915, nand2);
    std::cerr << "\rLUT916   ";
    Ciphertext gin916 = 2 * gout838 + gout913;
    Ciphertext gout916 = ctx.eval_lut(&gin916, nand2);
    std::cerr << "\rLUT917   ";
    Ciphertext gin917 = 2 * gout428 + gout911;
    Ciphertext gout917 = ctx.eval_lut(&gin917, and2);
    std::cerr << "\rLUT918   ";
    Ciphertext gin918 = 2 * gout916 + gout847;
    Ciphertext gout918 = ctx.eval_lut(&gin918, xnor2);
    std::cerr << "\rLUT919   ";
    Ciphertext gin919 = 2 * gout428 + gout914;
    Ciphertext gout919 = ctx.eval_lut(&gin919, and2);
    std::cerr << "\rLUT920   ";
    Ciphertext gin920 = 2 * gout913 + gout912;
    Ciphertext gout920 = ctx.eval_lut(&gin920, nand2);
    std::cerr << "\rLUT921   ";
    Ciphertext gin921 = 2 * gout898 + gout917;
    Ciphertext gout921 = ctx.eval_lut(&gin921, nand2);
    std::cerr << "\rLUT922   ";
    Ciphertext gin922 = 2 * gout416 + gout920;
    Ciphertext gout922 = ctx.eval_lut(&gin922, and2);
    std::cerr << "\rLUT923   ";
    Ciphertext gin923 = 2 * gout916 + gout847;
    Ciphertext gout923 = ctx.eval_lut(&gin923, nand2);
    std::cerr << "\rLUT924   ";
    Ciphertext gin924 = 2 * gout915 + gout919;
    Ciphertext gout924 = ctx.eval_lut(&gin924, nand2);
    std::cerr << "\rLUT925   ";
    Ciphertext gin925 = 2 * gout842 + gout923;
    Ciphertext gout925 = ctx.eval_lut(&gin925, and2);
    std::cerr << "\rLUT926   ";
    Ciphertext gin926 = 2 * gout416 + gout924;
    Ciphertext gout926 = ctx.eval_lut(&gin926, and2);
    std::cerr << "\rIMUX927  ";
    Ciphertext gin927A = 2 * gout918 + gout387;
    Ciphertext gout927A = ctx.eval_lut(&gin927A, and2);
    Ciphertext gin927B = 2 * gout908 + (1 - gout387);
    Ciphertext gout927B = ctx.eval_lut(&gin927B, and2);
    Ciphertext gout927 = gout927A + gout927B;
    std::cerr << "\rLUT928   ";
    Ciphertext gin928 = 2 * gout416 + gout927;
    Ciphertext gout928 = ctx.eval_lut(&gin928, nand2);
    std::cerr << "\rLUT929   ";
    Ciphertext gin929 = 2 * gout925 + gout852;
    Ciphertext gout929 = ctx.eval_lut(&gin929, or2);
    std::cerr << "\rLUT930   ";
    Ciphertext gin930 = 2 * gout387 + gout929;
    Ciphertext gout930 = ctx.eval_lut(&gin930, and2);
    std::cerr << "\rLUT931   ";
    Ciphertext gin931 = 2 * gout925 + gout852;
    Ciphertext gout931 = ctx.eval_lut(&gin931, nand2);
    std::cerr << "\rLUT932   ";
    Ciphertext gin932 = 2 * gout921 + gout922;
    Ciphertext gout932 = ctx.eval_lut(&gin932, nand2);
    std::cerr << "\rLUT933   ";
    Ciphertext gin933 = 2 * gout931 + gout930;
    Ciphertext gout933 = ctx.eval_lut(&gin933, nand2);
    std::cerr << "\rLUT934   ";
    Ciphertext gin934 = 2 * gout933 + gout926;
    Ciphertext gout934 = ctx.eval_lut(&gin934, nand2);

    std::cerr << "\r          \r";
    std::vector<long> out;
    out.push_back(ctx.decrypt(&gout516)); /* out0 */
    out.push_back(ctx.decrypt(&gout657)); /* out1 */
    out.push_back(ctx.decrypt(&gout692)); /* out2 */
    out.push_back(ctx.decrypt(&gout726)); /* out3 */
    out.push_back(ctx.decrypt(&gout762)); /* out4 */
    out.push_back(ctx.decrypt(&gout801)); /* out5 */
    out.push_back(ctx.decrypt(&gout824)); /* out6 */
    out.push_back(ctx.decrypt(&gout857)); /* out7 */
    out.push_back(ctx.decrypt(&gout862)); /* out8 */
    out.push_back(ctx.decrypt(&gout887)); /* out9 */
    out.push_back(ctx.decrypt(&gout892)); /* out10 */
    out.push_back(ctx.decrypt(&gout906)); /* out11 */
    out.push_back(ctx.decrypt(&gout910)); /* out12 */
    out.push_back(ctx.decrypt(&gout932)); /* out13 */
    out.push_back(ctx.decrypt(&gout928)); /* out14 */
    out.push_back(ctx.decrypt(&gout934)); /* out15 */
    return out;
}

std::vector<long>
test_sqrt_boolean(FHEContext& ctx, uint16_t num)
{
    Ciphertext ct_num0 = ctx.encrypt_public((num >> 0) & 1);
    Ciphertext ct_num1 = ctx.encrypt_public((num >> 1) & 1);
    Ciphertext ct_num2 = ctx.encrypt_public((num >> 2) & 1);
    Ciphertext ct_num3 = ctx.encrypt_public((num >> 3) & 1);
    Ciphertext ct_num4 = ctx.encrypt_public((num >> 4) & 1);
    Ciphertext ct_num5 = ctx.encrypt_public((num >> 5) & 1);
    Ciphertext ct_num6 = ctx.encrypt_public((num >> 6) & 1);
    Ciphertext ct_num7 = ctx.encrypt_public((num >> 7) & 1);
    Ciphertext ct_num8 = ctx.encrypt_public((num >> 8) & 1);
    Ciphertext ct_num9 = ctx.encrypt_public((num >> 9) & 1);
    Ciphertext ct_num10 = ctx.encrypt_public((num >> 10) & 1);
    Ciphertext ct_num11 = ctx.encrypt_public((num >> 11) & 1);
    Ciphertext ct_num12 = ctx.encrypt_public((num >> 12) & 1);
    Ciphertext ct_num13 = ctx.encrypt_public((num >> 13) & 1);
    Ciphertext ct_num14 = ctx.encrypt_public((num >> 14) & 1);
    Ciphertext ct_num15 = ctx.encrypt_public((num >> 15) & 1);

    auto fand2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly and2 = ctx.genrate_lut(fand2);

    auto fnand2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly nand2 = ctx.genrate_lut(fnand2);

    auto for2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly or2 = ctx.genrate_lut(for2);

    auto fnor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly nor2 = ctx.genrate_lut(fnor2);

    auto fxor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 0;
            case 1:  return 1;
            case 2:  return 1;
            case 3:  return 0;
            default: assert(0);
        }
    };
    RotationPoly xor2 = ctx.genrate_lut(fxor2);

    auto fxnor2 = [](long i, long t) -> long {
        switch (i) {
            case 0:  return 1;
            case 1:  return 0;
            case 2:  return 0;
            case 3:  return 1;
            default: assert(0);
        }
    };
    RotationPoly xnor2 = ctx.genrate_lut(fxnor2);

    std::cerr << "\rLUT0   ";
    Ciphertext gin0 = 2 * ct_num9 + ct_num8;
    Ciphertext gout0 = ctx.eval_lut(&gin0, nor2);
    std::cerr << "\rINV1   ";
    Ciphertext gout1 = 1 - ct_num0;

    std::cerr << "\rLUT2   ";
    Ciphertext gin2 = 2 * ct_num13 + ct_num12;
    Ciphertext gout2 = ctx.eval_lut(&gin2, or2);
    std::cerr << "\rINV3   ";
    Ciphertext gout3 = 1 - ct_num3;

    std::cerr << "\rINV4   ";
    Ciphertext gout4 = 1 - ct_num6;

    std::cerr << "\rLUT5   ";
    Ciphertext gin5 = 2 * ct_num14 + ct_num15;
    Ciphertext gout5 = ctx.eval_lut(&gin5, nor2);
    std::cerr << "\rINV6   ";
    Ciphertext gout6 = 1 - ct_num5;

    std::cerr << "\rLUT7   ";
    Ciphertext gin7 = 2 * ct_num2 + ct_num3;
    Ciphertext gout7 = ctx.eval_lut(&gin7, nor2);
    std::cerr << "\rLUT8   ";
    Ciphertext gin8 = 2 * ct_num9 + ct_num8;
    Ciphertext gout8 = ctx.eval_lut(&gin8, nand2);
    std::cerr << "\rLUT9   ";
    Ciphertext gin9 = 2 * (1 - ct_num15) + ct_num13;
    Ciphertext gout9 = ctx.eval_lut(&gin9, and2);
    std::cerr << "\rLUT10   ";
    Ciphertext gin10 = 2 * ct_num7 + ct_num6;
    Ciphertext gout10 = ctx.eval_lut(&gin10, or2);
    std::cerr << "\rLUT11   ";
    Ciphertext gin11 = 2 * (1 - ct_num14) + ct_num15;
    Ciphertext gout11 = ctx.eval_lut(&gin11, or2);
    std::cerr << "\rLUT12   ";
    Ciphertext gin12 = 2 * gout5 + gout2;
    Ciphertext gout12 = ctx.eval_lut(&gin12, nand2);
    std::cerr << "\rINV13   ";
    Ciphertext gout13 = 1 - ct_num11;

    std::cerr << "\rINV14   ";
    Ciphertext gout14 = 1 - ct_num12;

    std::cerr << "\rLUT15   ";
    Ciphertext gin15 = 2 * ct_num2 + ct_num1;
    Ciphertext gout15 = ctx.eval_lut(&gin15, nor2);
    std::cerr << "\rINV16   ";
    Ciphertext gout16 = 1 - ct_num15;

    std::cerr << "\rINV17   ";
    Ciphertext gout17 = 1 - ct_num4;

    std::cerr << "\rLUT18   ";
    Ciphertext gin18 = 2 * ct_num12 + gout5;
    Ciphertext gout18 = ctx.eval_lut(&gin18, nand2);
    std::cerr << "\rLUT19   ";
    Ciphertext gin19 = 2 * ct_num11 + ct_num12;
    Ciphertext gout19 = ctx.eval_lut(&gin19, nor2);
    std::cerr << "\rLUT20   ";
    Ciphertext gin20 = 2 * (1 - ct_num3) + ct_num2;
    Ciphertext gout20 = ctx.eval_lut(&gin20, and2);
    std::cerr << "\rLUT21   ";
    Ciphertext gin21 = 2 * gout14 + gout5;
    Ciphertext gout21 = ctx.eval_lut(&gin21, nand2);
    std::cerr << "\rLUT22   ";
    Ciphertext gin22 = 2 * gout17 + gout7;
    Ciphertext gout22 = ctx.eval_lut(&gin22, xnor2);
    std::cerr << "\rLUT23   ";
    Ciphertext gin23 = 2 * ct_num5 + ct_num4;
    Ciphertext gout23 = ctx.eval_lut(&gin23, nor2);
    std::cerr << "\rINV24   ";
    Ciphertext gout24 = 1 - gout0;

    std::cerr << "\rINV25   ";
    Ciphertext gout25 = 1 - ct_num10;

    std::cerr << "\rINV26   ";
    Ciphertext gout26 = 1 - ct_num9;

    std::cerr << "\rLUT27   ";
    Ciphertext gin27 = 2 * gout25 + gout19;
    Ciphertext gout27 = ctx.eval_lut(&gin27, nand2);
    std::cerr << "\rINV28   ";
    Ciphertext gout28 = 1 - gout11;

    std::cerr << "\rLUT29   ";
    Ciphertext gin29 = 2 * ct_num6 + gout23;
    Ciphertext gout29 = ctx.eval_lut(&gin29, xnor2);
    std::cerr << "\rLUT30   ";
    Ciphertext gin30 = 2 * gout5 + gout9;
    Ciphertext gout30 = ctx.eval_lut(&gin30, or2);
    std::cerr << "\rLUT31   ";
    Ciphertext gin31 = 2 * gout1 + gout15;
    Ciphertext gout31 = ctx.eval_lut(&gin31, nand2);
    std::cerr << "\rINV32   ";
    Ciphertext gout32 = 1 - gout12;

    std::cerr << "\rLUT33   ";
    Ciphertext gin33 = 2 * ct_num13 + gout21;
    Ciphertext gout33 = ctx.eval_lut(&gin33, and2);
    std::cerr << "\rLUT34   ";
    Ciphertext gin34 = 2 * gout18 + gout27;
    Ciphertext gout34 = ctx.eval_lut(&gin34, nand2);
    std::cerr << "\rLUT35   ";
    Ciphertext gin35 = 2 * gout11 + gout33;
    Ciphertext gout35 = ctx.eval_lut(&gin35, nand2);
    std::cerr << "\rLUT36   ";
    Ciphertext gin36 = 2 * gout35 + gout34;
    Ciphertext gout36 = ctx.eval_lut(&gin36, nand2);
    std::cerr << "\rLUT37   ";
    Ciphertext gin37 = 2 * gout36 + gout12;
    Ciphertext gout37 = ctx.eval_lut(&gin37, or2);
    std::cerr << "\rLUT38   ";
    Ciphertext gin38 = 2 * gout36 + gout30;
    Ciphertext gout38 = ctx.eval_lut(&gin38, and2);
    std::cerr << "\rLUT39   ";
    Ciphertext gin39 = 2 * gout25 + gout38;
    Ciphertext gout39 = ctx.eval_lut(&gin39, and2);
    std::cerr << "\rLUT40   ";
    Ciphertext gin40 = 2 * gout13 + gout39;
    Ciphertext gout40 = ctx.eval_lut(&gin40, nand2);
    std::cerr << "\rIMUX41  ";
    Ciphertext gin41A = 2 * gout38 + ct_num10;
    Ciphertext gout41A = ctx.eval_lut(&gin41A, and2);
    Ciphertext gin41B = 2 * gout0 + (1 - ct_num10);
    Ciphertext gout41B = ctx.eval_lut(&gin41B, and2);
    Ciphertext gout41 = gout41A + gout41B;
    std::cerr << "\rLUT42   ";
    Ciphertext gin42 = 2 * ct_num11 + gout39;
    Ciphertext gout42 = ctx.eval_lut(&gin42, xnor2);
    std::cerr << "\rINV43   ";
    Ciphertext gout43 = 1 - gout38;

    std::cerr << "\rLUT44   ";
    Ciphertext gin44 = 2 * gout40 + gout37;
    Ciphertext gout44 = ctx.eval_lut(&gin44, and2);
    std::cerr << "\rLUT45   ";
    Ciphertext gin45 = 2 * (1 - gout33) + gout38;
    Ciphertext gout45 = ctx.eval_lut(&gin45, or2);
    std::cerr << "\rLUT46   ";
    Ciphertext gin46 = 2 * gout32 + gout42;
    Ciphertext gout46 = ctx.eval_lut(&gin46, or2);
    std::cerr << "\rLUT47   ";
    Ciphertext gin47 = 2 * gout12 + gout42;
    Ciphertext gout47 = ctx.eval_lut(&gin47, xnor2);
    std::cerr << "\rINV48   ";
    Ciphertext gout48 = 1 - gout42;

    std::cerr << "\rLUT49   ";
    Ciphertext gin49 = 2 * ct_num12 + gout44;
    Ciphertext gout49 = ctx.eval_lut(&gin49, xnor2);
    std::cerr << "\rINV50   ";
    Ciphertext gout50 = 1 - gout41;

    std::cerr << "\rLUT51   ";
    Ciphertext gin51 = 2 * (1 - gout49) + gout45;
    Ciphertext gout51 = ctx.eval_lut(&gin51, and2);
    std::cerr << "\rLUT52   ";
    Ciphertext gin52 = 2 * gout47 + gout41;
    Ciphertext gout52 = ctx.eval_lut(&gin52, xnor2);
    std::cerr << "\rLUT53   ";
    Ciphertext gin53 = 2 * gout28 + gout49;
    Ciphertext gout53 = ctx.eval_lut(&gin53, xnor2);
    std::cerr << "\rLUT54   ";
    Ciphertext gin54 = 2 * gout11 + gout49;
    Ciphertext gout54 = ctx.eval_lut(&gin54, nand2);
    std::cerr << "\rLUT55   ";
    Ciphertext gin55 = 2 * gout47 + gout50;
    Ciphertext gout55 = ctx.eval_lut(&gin55, nand2);
    std::cerr << "\rLUT56   ";
    Ciphertext gin56 = 2 * gout54 + gout45;
    Ciphertext gout56 = ctx.eval_lut(&gin56, and2);
    std::cerr << "\rLUT57   ";
    Ciphertext gin57 = 2 * gout46 + gout55;
    Ciphertext gout57 = ctx.eval_lut(&gin57, nand2);
    std::cerr << "\rLUT58   ";
    Ciphertext gin58 = 2 * gout53 + gout57;
    Ciphertext gout58 = ctx.eval_lut(&gin58, xnor2);
    std::cerr << "\rLUT59   ";
    Ciphertext gin59 = 2 * gout53 + gout57;
    Ciphertext gout59 = ctx.eval_lut(&gin59, nand2);
    std::cerr << "\rLUT60   ";
    Ciphertext gin60 = 2 * gout59 + gout56;
    Ciphertext gout60 = ctx.eval_lut(&gin60, nand2);
    std::cerr << "\rLUT61   ";
    Ciphertext gin61 = 2 * gout16 + gout60;
    Ciphertext gout61 = ctx.eval_lut(&gin61, nand2);
    std::cerr << "\rLUT62   ";
    Ciphertext gin62 = 2 * gout26 + gout61;
    Ciphertext gout62 = ctx.eval_lut(&gin62, nand2);
    std::cerr << "\rIMUX63  ";
    Ciphertext gin63A = 2 * gout38 + gout61;
    Ciphertext gout63A = ctx.eval_lut(&gin63A, and2);
    Ciphertext gin63B = 2 * gout0 + (1 - gout61);
    Ciphertext gout63B = ctx.eval_lut(&gin63B, and2);
    Ciphertext gout63 = gout63A + gout63B;
    std::cerr << "\rINV64   ";
    Ciphertext gout64 = 1 - gout61;

    std::cerr << "\rIMUX65  ";
    Ciphertext gin65A = 2 * gout61 + ct_num8;
    Ciphertext gout65A = ctx.eval_lut(&gin65A, and2);
    Ciphertext gin65B = 2 * gout10 + (1 - ct_num8);
    Ciphertext gout65B = ctx.eval_lut(&gin65B, and2);
    Ciphertext gout65 = gout65A + gout65B;
    std::cerr << "\rIMUX66  ";
    Ciphertext gin66A = 2 * gout48 + gout61;
    Ciphertext gout66A = ctx.eval_lut(&gin66A, and2);
    Ciphertext gin66B = 2 * gout52 + (1 - gout61);
    Ciphertext gout66B = ctx.eval_lut(&gin66B, and2);
    Ciphertext gout66 = gout66A + gout66B;
    std::cerr << "\rIMUX67  ";
    Ciphertext gin67A = 2 * gout58 + gout64;
    Ciphertext gout67A = ctx.eval_lut(&gin67A, and2);
    Ciphertext gin67B = 2 * gout51 + (1 - gout64);
    Ciphertext gout67B = ctx.eval_lut(&gin67B, and2);
    Ciphertext gout67 = gout67A + gout67B;
    std::cerr << "\rLUT68   ";
    Ciphertext gin68 = 2 * gout11 + gout66;
    Ciphertext gout68 = ctx.eval_lut(&gin68, nand2);
    std::cerr << "\rLUT69   ";
    Ciphertext gin69 = 2 * gout28 + gout66;
    Ciphertext gout69 = ctx.eval_lut(&gin69, xnor2);
    std::cerr << "\rLUT70   ";
    Ciphertext gin70 = 2 * gout68 + gout67;
    Ciphertext gout70 = ctx.eval_lut(&gin70, and2);
    std::cerr << "\rLUT71   ";
    Ciphertext gin71 = 2 * gout24 + gout64;
    Ciphertext gout71 = ctx.eval_lut(&gin71, and2);
    std::cerr << "\rLUT72   ";
    Ciphertext gin72 = 2 * gout25 + gout63;
    Ciphertext gout72 = ctx.eval_lut(&gin72, xnor2);
    std::cerr << "\rLUT73   ";
    Ciphertext gin73 = 2 * gout71 + gout8;
    Ciphertext gout73 = ctx.eval_lut(&gin73, nand2);
    std::cerr << "\rLUT74   ";
    Ciphertext gin74 = 2 * gout12 + gout72;
    Ciphertext gout74 = ctx.eval_lut(&gin74, nand2);
    std::cerr << "\rLUT75   ";
    Ciphertext gin75 = 2 * gout32 + gout72;
    Ciphertext gout75 = ctx.eval_lut(&gin75, xnor2);
    std::cerr << "\rLUT76   ";
    Ciphertext gin76 = 2 * gout73 + gout62;
    Ciphertext gout76 = ctx.eval_lut(&gin76, and2);
    std::cerr << "\rLUT77   ";
    Ciphertext gin77 = 2 * gout73 + gout62;
    Ciphertext gout77 = ctx.eval_lut(&gin77, nand2);
    std::cerr << "\rINV78   ";
    Ciphertext gout78 = 1 - gout75;

    std::cerr << "\rLUT79   ";
    Ciphertext gin79 = 2 * gout38 + gout76;
    Ciphertext gout79 = ctx.eval_lut(&gin79, xnor2);
    std::cerr << "\rLUT80   ";
    Ciphertext gin80 = 2 * gout43 + gout76;
    Ciphertext gout80 = ctx.eval_lut(&gin80, nand2);
    std::cerr << "\rLUT81   ";
    Ciphertext gin81 = 2 * gout38 + gout77;
    Ciphertext gout81 = ctx.eval_lut(&gin81, nand2);
    std::cerr << "\rLUT82   ";
    Ciphertext gin82 = 2 * gout81 + gout65;
    Ciphertext gout82 = ctx.eval_lut(&gin82, nand2);
    std::cerr << "\rLUT83   ";
    Ciphertext gin83 = 2 * gout65 + gout79;
    Ciphertext gout83 = ctx.eval_lut(&gin83, xnor2);
    std::cerr << "\rLUT84   ";
    Ciphertext gin84 = 2 * gout80 + gout82;
    Ciphertext gout84 = ctx.eval_lut(&gin84, and2);
    std::cerr << "\rLUT85   ";
    Ciphertext gin85 = 2 * gout75 + gout84;
    Ciphertext gout85 = ctx.eval_lut(&gin85, xnor2);
    std::cerr << "\rLUT86   ";
    Ciphertext gin86 = 2 * gout78 + gout84;
    Ciphertext gout86 = ctx.eval_lut(&gin86, or2);
    std::cerr << "\rLUT87   ";
    Ciphertext gin87 = 2 * gout74 + gout86;
    Ciphertext gout87 = ctx.eval_lut(&gin87, nand2);
    std::cerr << "\rLUT88   ";
    Ciphertext gin88 = 2 * gout69 + gout87;
    Ciphertext gout88 = ctx.eval_lut(&gin88, xor2);
    std::cerr << "\rLUT89   ";
    Ciphertext gin89 = 2 * gout69 + gout87;
    Ciphertext gout89 = ctx.eval_lut(&gin89, nand2);
    std::cerr << "\rLUT90   ";
    Ciphertext gin90 = 2 * gout89 + gout70;
    Ciphertext gout90 = ctx.eval_lut(&gin90, nand2);
    std::cerr << "\rLUT91   ";
    Ciphertext gin91 = 2 * gout61 + gout90;
    Ciphertext gout91 = ctx.eval_lut(&gin91, or2);
    std::cerr << "\rLUT92   ";
    Ciphertext gin92 = 2 * gout16 + gout90;
    Ciphertext gout92 = ctx.eval_lut(&gin92, nand2);
    std::cerr << "\rIMUX93  ";
    Ciphertext gin93A = 2 * gout66 + gout92;
    Ciphertext gout93A = ctx.eval_lut(&gin93A, and2);
    Ciphertext gin93B = 2 * gout88 + (1 - gout92);
    Ciphertext gout93B = ctx.eval_lut(&gin93B, and2);
    Ciphertext gout93 = gout93A + gout93B;
    std::cerr << "\rLUT94   ";
    Ciphertext gin94 = 2 * (1 - gout92) + gout4;
    Ciphertext gout94 = ctx.eval_lut(&gin94, and2);
    std::cerr << "\rIMUX95  ";
    Ciphertext gin95A = 2 * gout77 + gout92;
    Ciphertext gout95A = ctx.eval_lut(&gin95A, and2);
    Ciphertext gin95B = 2 * gout83 + (1 - gout92);
    Ciphertext gout95B = ctx.eval_lut(&gin95B, and2);
    Ciphertext gout95 = gout95A + gout95B;
    std::cerr << "\rLUT96   ";
    Ciphertext gin96 = 2 * gout10 + gout92;
    Ciphertext gout96 = ctx.eval_lut(&gin96, or2);
    std::cerr << "\rLUT97   ";
    Ciphertext gin97 = 2 * ct_num7 + gout94;
    Ciphertext gout97 = ctx.eval_lut(&gin97, xnor2);
    std::cerr << "\rLUT98   ";
    Ciphertext gin98 = 2 * gout32 + gout95;
    Ciphertext gout98 = ctx.eval_lut(&gin98, or2);
    std::cerr << "\rLUT99   ";
    Ciphertext gin99 = 2 * gout4 + gout92;
    Ciphertext gout99 = ctx.eval_lut(&gin99, xnor2);
    std::cerr << "\rLUT100   ";
    Ciphertext gin100 = 2 * gout61 + gout97;
    Ciphertext gout100 = ctx.eval_lut(&gin100, xnor2);
    std::cerr << "\rLUT101   ";
    Ciphertext gin101 = 2 * gout64 + gout97;
    Ciphertext gout101 = ctx.eval_lut(&gin101, or2);
    std::cerr << "\rIMUX102  ";
    Ciphertext gin102A = 2 * gout72 + gout92;
    Ciphertext gout102A = ctx.eval_lut(&gin102A, and2);
    Ciphertext gin102B = 2 * gout85 + (1 - gout92);
    Ciphertext gout102B = ctx.eval_lut(&gin102B, and2);
    Ciphertext gout102 = gout102A + gout102B;
    std::cerr << "\rLUT103   ";
    Ciphertext gin103 = 2 * gout96 + gout91;
    Ciphertext gout103 = ctx.eval_lut(&gin103, and2);
    std::cerr << "\rINV104   ";
    Ciphertext gout104 = 1 - gout97;

    std::cerr << "\rLUT105   ";
    Ciphertext gin105 = 2 * ct_num8 + gout103;
    Ciphertext gout105 = ctx.eval_lut(&gin105, xnor2);
    std::cerr << "\rLUT106   ";
    Ciphertext gin106 = 2 * gout38 + gout105;
    Ciphertext gout106 = ctx.eval_lut(&gin106, xnor2);
    std::cerr << "\rINV107   ";
    Ciphertext gout107 = 1 - gout92;

    std::cerr << "\rLUT108   ";
    Ciphertext gin108 = 2 * gout32 + gout95;
    Ciphertext gout108 = ctx.eval_lut(&gin108, xnor2);
    std::cerr << "\rINV109   ";
    Ciphertext gout109 = 1 - gout102;

    std::cerr << "\rLUT110   ";
    Ciphertext gin110 = 2 * gout28 + gout102;
    Ciphertext gout110 = ctx.eval_lut(&gin110, xnor2);
    std::cerr << "\rLUT111   ";
    Ciphertext gin111 = 2 * gout11 + gout102;
    Ciphertext gout111 = ctx.eval_lut(&gin111, nand2);
    std::cerr << "\rIMUX112  ";
    Ciphertext gin112A = 2 * gout107 + ct_num6;
    Ciphertext gout112A = ctx.eval_lut(&gin112A, and2);
    Ciphertext gin112B = 2 * gout23 + (1 - ct_num6);
    Ciphertext gout112B = ctx.eval_lut(&gin112B, and2);
    Ciphertext gout112 = gout112A + gout112B;
    std::cerr << "\rLUT113   ";
    Ciphertext gin113 = 2 * gout43 + gout105;
    Ciphertext gout113 = ctx.eval_lut(&gin113, nand2);
    std::cerr << "\rINV114   ";
    Ciphertext gout114 = 1 - gout112;

    std::cerr << "\rINV115   ";
    Ciphertext gout115 = 1 - gout105;

    std::cerr << "\rLUT116   ";
    Ciphertext gin116 = 2 * gout100 + gout112;
    Ciphertext gout116 = ctx.eval_lut(&gin116, xnor2);
    std::cerr << "\rLUT117   ";
    Ciphertext gin117 = 2 * (1 - gout93) + gout111;
    Ciphertext gout117 = ctx.eval_lut(&gin117, and2);
    std::cerr << "\rLUT118   ";
    Ciphertext gin118 = 2 * gout100 + gout114;
    Ciphertext gout118 = ctx.eval_lut(&gin118, nand2);
    std::cerr << "\rLUT119   ";
    Ciphertext gin119 = 2 * gout101 + gout118;
    Ciphertext gout119 = ctx.eval_lut(&gin119, nand2);
    std::cerr << "\rLUT120   ";
    Ciphertext gin120 = 2 * gout119 + gout106;
    Ciphertext gout120 = ctx.eval_lut(&gin120, xnor2);
    std::cerr << "\rLUT121   ";
    Ciphertext gin121 = 2 * gout119 + gout106;
    Ciphertext gout121 = ctx.eval_lut(&gin121, nand2);
    std::cerr << "\rLUT122   ";
    Ciphertext gin122 = 2 * gout113 + gout121;
    Ciphertext gout122 = ctx.eval_lut(&gin122, and2);
    std::cerr << "\rLUT123   ";
    Ciphertext gin123 = 2 * gout108 + gout122;
    Ciphertext gout123 = ctx.eval_lut(&gin123, or2);
    std::cerr << "\rLUT124   ";
    Ciphertext gin124 = 2 * gout108 + gout122;
    Ciphertext gout124 = ctx.eval_lut(&gin124, xnor2);
    std::cerr << "\rLUT125   ";
    Ciphertext gin125 = 2 * gout98 + gout123;
    Ciphertext gout125 = ctx.eval_lut(&gin125, nand2);
    std::cerr << "\rLUT126   ";
    Ciphertext gin126 = 2 * gout110 + gout125;
    Ciphertext gout126 = ctx.eval_lut(&gin126, nand2);
    std::cerr << "\rLUT127   ";
    Ciphertext gin127 = 2 * gout110 + gout125;
    Ciphertext gout127 = ctx.eval_lut(&gin127, xnor2);
    std::cerr << "\rLUT128   ";
    Ciphertext gin128 = 2 * gout126 + gout117;
    Ciphertext gout128 = ctx.eval_lut(&gin128, nand2);
    std::cerr << "\rLUT129   ";
    Ciphertext gin129 = 2 * gout16 + gout128;
    Ciphertext gout129 = ctx.eval_lut(&gin129, nand2);
    std::cerr << "\rIMUX130  ";
    Ciphertext gin130A = 2 * gout99 + gout129;
    Ciphertext gout130A = ctx.eval_lut(&gin130A, and2);
    Ciphertext gin130B = 2 * gout29 + (1 - gout129);
    Ciphertext gout130B = ctx.eval_lut(&gin130B, and2);
    Ciphertext gout130 = gout130A + gout130B;
    std::cerr << "\rIMUX131  ";
    Ciphertext gin131A = 2 * gout95 + gout129;
    Ciphertext gout131A = ctx.eval_lut(&gin131A, and2);
    Ciphertext gin131B = 2 * gout124 + (1 - gout129);
    Ciphertext gout131B = ctx.eval_lut(&gin131B, and2);
    Ciphertext gout131 = gout131A + gout131B;
    std::cerr << "\rIMUX132  ";
    Ciphertext gin132A = 2 * gout115 + gout129;
    Ciphertext gout132A = ctx.eval_lut(&gin132A, and2);
    Ciphertext gin132B = 2 * gout120 + (1 - gout129);
    Ciphertext gout132B = ctx.eval_lut(&gin132B, and2);
    Ciphertext gout132 = gout132A + gout132B;
    std::cerr << "\rLUT133   ";
    Ciphertext gin133 = 2 * ct_num4 + gout129;
    Ciphertext gout133 = ctx.eval_lut(&gin133, xnor2);
    std::cerr << "\rLUT134   ";
    Ciphertext gin134 = 2 * gout129 + gout31;
    Ciphertext gout134 = ctx.eval_lut(&gin134, or2);
    std::cerr << "\rIMUX135  ";
    Ciphertext gin135A = 2 * gout104 + gout129;
    Ciphertext gout135A = ctx.eval_lut(&gin135A, and2);
    Ciphertext gin135B = 2 * gout116 + (1 - gout129);
    Ciphertext gout135B = ctx.eval_lut(&gin135B, and2);
    Ciphertext gout135 = gout135A + gout135B;
    std::cerr << "\rLUT136   ";
    Ciphertext gin136 = 2 * gout129 + gout31;
    Ciphertext gout136 = ctx.eval_lut(&gin136, nand2);
    std::cerr << "\rINV137   ";
    Ciphertext gout137 = 1 - gout129;

    std::cerr << "\rLUT138   ";
    Ciphertext gin138 = 2 * (1 - gout129) + gout17;
    Ciphertext gout138 = ctx.eval_lut(&gin138, and2);
    std::cerr << "\rLUT139   ";
    Ciphertext gin139 = 2 * gout11 + gout131;
    Ciphertext gout139 = ctx.eval_lut(&gin139, xnor2);
    std::cerr << "\rLUT140   ";
    Ciphertext gin140 = 2 * gout43 + gout135;
    Ciphertext gout140 = ctx.eval_lut(&gin140, nand2);
    std::cerr << "\rLUT141   ";
    Ciphertext gin141 = 2 * gout61 + gout130;
    Ciphertext gout141 = ctx.eval_lut(&gin141, xnor2);
    std::cerr << "\rIMUX142  ";
    Ciphertext gin142A = 2 * gout109 + gout129;
    Ciphertext gout142A = ctx.eval_lut(&gin142A, and2);
    Ciphertext gin142B = 2 * gout127 + (1 - gout129);
    Ciphertext gout142B = ctx.eval_lut(&gin142B, and2);
    Ciphertext gout142 = gout142A + gout142B;
    std::cerr << "\rLUT143   ";
    Ciphertext gin143 = 2 * gout6 + gout138;
    Ciphertext gout143 = ctx.eval_lut(&gin143, xnor2);
    std::cerr << "\rLUT144   ";
    Ciphertext gin144 = 2 * gout38 + gout135;
    Ciphertext gout144 = ctx.eval_lut(&gin144, xnor2);
    std::cerr << "\rLUT145   ";
    Ciphertext gin145 = 2 * gout92 + gout143;
    Ciphertext gout145 = ctx.eval_lut(&gin145, nand2);
    std::cerr << "\rLUT146   ";
    Ciphertext gin146 = 2 * gout64 + gout130;
    Ciphertext gout146 = ctx.eval_lut(&gin146, or2);
    std::cerr << "\rINV147   ";
    Ciphertext gout147 = 1 - gout141;

    std::cerr << "\rLUT148   ";
    Ciphertext gin148 = 2 * gout131 + gout142;
    Ciphertext gout148 = ctx.eval_lut(&gin148, nand2);
    std::cerr << "\rINV149   ";
    Ciphertext gout149 = 1 - gout144;

    std::cerr << "\rLUT150   ";
    Ciphertext gin150 = 2 * gout12 + gout132;
    Ciphertext gout150 = ctx.eval_lut(&gin150, xnor2);
    std::cerr << "\rLUT151   ";
    Ciphertext gin151 = 2 * gout107 + gout143;
    Ciphertext gout151 = ctx.eval_lut(&gin151, xnor2);
    std::cerr << "\rLUT152   ";
    Ciphertext gin152 = 2 * gout28 + gout131;
    Ciphertext gout152 = ctx.eval_lut(&gin152, or2);
    std::cerr << "\rIMUX153  ";
    Ciphertext gin153A = 2 * gout137 + ct_num4;
    Ciphertext gout153A = ctx.eval_lut(&gin153A, and2);
    Ciphertext gin153B = 2 * gout7 + (1 - ct_num4);
    Ciphertext gout153B = ctx.eval_lut(&gin153B, and2);
    Ciphertext gout153 = gout153A + gout153B;
    std::cerr << "\rLUT154   ";
    Ciphertext gin154 = 2 * gout32 + gout132;
    Ciphertext gout154 = ctx.eval_lut(&gin154, or2);
    std::cerr << "\rLUT155   ";
    Ciphertext gin155 = 2 * gout152 + gout142;
    Ciphertext gout155 = ctx.eval_lut(&gin155, and2);
    std::cerr << "\rINV156   ";
    Ciphertext gout156 = 1 - gout153;

    std::cerr << "\rLUT157   ";
    Ciphertext gin157 = 2 * gout151 + gout153;
    Ciphertext gout157 = ctx.eval_lut(&gin157, xnor2);
    std::cerr << "\rLUT158   ";
    Ciphertext gin158 = 2 * gout151 + gout156;
    Ciphertext gout158 = ctx.eval_lut(&gin158, nand2);
    std::cerr << "\rLUT159   ";
    Ciphertext gin159 = 2 * gout145 + gout158;
    Ciphertext gout159 = ctx.eval_lut(&gin159, and2);
    std::cerr << "\rLUT160   ";
    Ciphertext gin160 = 2 * gout147 + gout159;
    Ciphertext gout160 = ctx.eval_lut(&gin160, or2);
    std::cerr << "\rLUT161   ";
    Ciphertext gin161 = 2 * gout146 + gout160;
    Ciphertext gout161 = ctx.eval_lut(&gin161, and2);
    std::cerr << "\rLUT162   ";
    Ciphertext gin162 = 2 * gout147 + gout159;
    Ciphertext gout162 = ctx.eval_lut(&gin162, xnor2);
    std::cerr << "\rLUT163   ";
    Ciphertext gin163 = 2 * gout144 + gout161;
    Ciphertext gout163 = ctx.eval_lut(&gin163, xnor2);
    std::cerr << "\rLUT164   ";
    Ciphertext gin164 = 2 * gout149 + gout161;
    Ciphertext gout164 = ctx.eval_lut(&gin164, or2);
    std::cerr << "\rLUT165   ";
    Ciphertext gin165 = 2 * gout140 + gout164;
    Ciphertext gout165 = ctx.eval_lut(&gin165, nand2);
    std::cerr << "\rLUT166   ";
    Ciphertext gin166 = 2 * gout165 + gout150;
    Ciphertext gout166 = ctx.eval_lut(&gin166, nand2);
    std::cerr << "\rLUT167   ";
    Ciphertext gin167 = 2 * gout165 + gout150;
    Ciphertext gout167 = ctx.eval_lut(&gin167, xnor2);
    std::cerr << "\rLUT168   ";
    Ciphertext gin168 = 2 * gout154 + gout166;
    Ciphertext gout168 = ctx.eval_lut(&gin168, nand2);
    std::cerr << "\rLUT169   ";
    Ciphertext gin169 = 2 * gout139 + gout168;
    Ciphertext gout169 = ctx.eval_lut(&gin169, xor2);
    std::cerr << "\rLUT170   ";
    Ciphertext gin170 = 2 * gout139 + gout168;
    Ciphertext gout170 = ctx.eval_lut(&gin170, nand2);
    std::cerr << "\rLUT171   ";
    Ciphertext gin171 = 2 * gout170 + gout155;
    Ciphertext gout171 = ctx.eval_lut(&gin171, nand2);
    std::cerr << "\rLUT172   ";
    Ciphertext gin172 = 2 * gout16 + gout171;
    Ciphertext gout172 = ctx.eval_lut(&gin172, nand2);
    std::cerr << "\rIMUX173  ";
    Ciphertext gin173A = 2 * gout148 + gout172;
    Ciphertext gout173A = ctx.eval_lut(&gin173A, and2);
    Ciphertext gin173B = 2 * gout169 + (1 - gout172);
    Ciphertext gout173B = ctx.eval_lut(&gin173B, and2);
    Ciphertext gout173 = gout173A + gout173B;
    std::cerr << "\rINV174   ";
    Ciphertext gout174 = 1 - gout172;

    std::cerr << "\rIMUX175  ";
    Ciphertext gin175A = 2 * gout130 + gout172;
    Ciphertext gout175A = ctx.eval_lut(&gin175A, and2);
    Ciphertext gin175B = 2 * gout162 + (1 - gout172);
    Ciphertext gout175B = ctx.eval_lut(&gin175B, and2);
    Ciphertext gout175 = gout175A + gout175B;
    std::cerr << "\rLUT176   ";
    Ciphertext gin176 = 2 * gout3 + gout172;
    Ciphertext gout176 = ctx.eval_lut(&gin176, xnor2);
    std::cerr << "\rIMUX177  ";
    Ciphertext gin177A = 2 * gout135 + gout172;
    Ciphertext gout177A = ctx.eval_lut(&gin177A, and2);
    Ciphertext gin177B = 2 * gout163 + (1 - gout172);
    Ciphertext gout177B = ctx.eval_lut(&gin177B, and2);
    Ciphertext gout177 = gout177A + gout177B;
    std::cerr << "\rIMUX178  ";
    Ciphertext gin178A = 2 * gout143 + gout172;
    Ciphertext gout178A = ctx.eval_lut(&gin178A, and2);
    Ciphertext gin178B = 2 * gout157 + (1 - gout172);
    Ciphertext gout178B = ctx.eval_lut(&gin178B, and2);
    Ciphertext gout178 = gout178A + gout178B;
    std::cerr << "\rLUT179   ";
    Ciphertext gin179 = 2 * gout38 + gout175;
    Ciphertext gout179 = ctx.eval_lut(&gin179, or2);
    std::cerr << "\rIMUX180  ";
    Ciphertext gin180A = 2 * gout133 + gout172;
    Ciphertext gout180A = ctx.eval_lut(&gin180A, and2);
    Ciphertext gin180B = 2 * gout22 + (1 - gout172);
    Ciphertext gout180B = ctx.eval_lut(&gin180B, and2);
    Ciphertext gout180 = gout180A + gout180B;
    std::cerr << "\rIMUX181  ";
    Ciphertext gin181A = 2 * gout132 + gout172;
    Ciphertext gout181A = ctx.eval_lut(&gin181A, and2);
    Ciphertext gin181B = 2 * gout167 + (1 - gout172);
    Ciphertext gout181B = ctx.eval_lut(&gin181B, and2);
    Ciphertext gout181 = gout181A + gout181B;
    std::cerr << "\rLUT182   ";
    Ciphertext gin182 = 2 * gout28 + gout181;
    Ciphertext gout182 = ctx.eval_lut(&gin182, nand2);
    std::cerr << "\rLUT183   ";
    Ciphertext gin183 = 2 * gout92 + gout180;
    Ciphertext gout183 = ctx.eval_lut(&gin183, or2);
    std::cerr << "\rLUT184   ";
    Ciphertext gin184 = 2 * gout92 + gout180;
    Ciphertext gout184 = ctx.eval_lut(&gin184, nand2);
    std::cerr << "\rLUT185   ";
    Ciphertext gin185 = 2 * gout12 + gout177;
    Ciphertext gout185 = ctx.eval_lut(&gin185, or2);
    std::cerr << "\rLUT186   ";
    Ciphertext gin186 = 2 * gout61 + gout178;
    Ciphertext gout186 = ctx.eval_lut(&gin186, or2);
    std::cerr << "\rLUT187   ";
    Ciphertext gin187 = 2 * gout38 + gout175;
    Ciphertext gout187 = ctx.eval_lut(&gin187, nand2);
    std::cerr << "\rLUT188   ";
    Ciphertext gin188 = 2 * gout136 + gout176;
    Ciphertext gout188 = ctx.eval_lut(&gin188, nand2);
    std::cerr << "\rLUT189   ";
    Ciphertext gin189 = 2 * gout12 + gout177;
    Ciphertext gout189 = ctx.eval_lut(&gin189, nand2);
    std::cerr << "\rLUT190   ";
    Ciphertext gin190 = 2 * gout61 + gout178;
    Ciphertext gout190 = ctx.eval_lut(&gin190, nand2);
    std::cerr << "\rLUT191   ";
    Ciphertext gin191 = 2 * gout28 + gout181;
    Ciphertext gout191 = ctx.eval_lut(&gin191, or2);
    std::cerr << "\rLUT192   ";
    Ciphertext gin192 = 2 * gout190 + gout184;
    Ciphertext gout192 = ctx.eval_lut(&gin192, and2);
    std::cerr << "\rLUT193   ";
    Ciphertext gin193 = 2 * (1 - gout172) + gout20;
    Ciphertext gout193 = ctx.eval_lut(&gin193, and2);
    std::cerr << "\rLUT194   ";
    Ciphertext gin194 = 2 * gout191 + (1 - gout173);
    Ciphertext gout194 = ctx.eval_lut(&gin194, and2);
    std::cerr << "\rLUT195   ";
    Ciphertext gin195 = 2 * (1 - gout193) + gout134;
    Ciphertext gout195 = ctx.eval_lut(&gin195, and2);
    std::cerr << "\rLUT196   ";
    Ciphertext gin196 = 2 * gout183 + gout195;
    Ciphertext gout196 = ctx.eval_lut(&gin196, and2);
    std::cerr << "\rLUT197   ";
    Ciphertext gin197 = 2 * gout182 + gout185;
    Ciphertext gout197 = ctx.eval_lut(&gin197, and2);
    std::cerr << "\rLUT198   ";
    Ciphertext gin198 = 2 * gout187 + gout186;
    Ciphertext gout198 = ctx.eval_lut(&gin198, and2);
    std::cerr << "\rLUT199   ";
    Ciphertext gin199 = 2 * gout189 + gout179;
    Ciphertext gout199 = ctx.eval_lut(&gin199, and2);
    std::cerr << "\rLUT200   ";
    Ciphertext gin200 = 2 * gout188 + gout196;
    Ciphertext gout200 = ctx.eval_lut(&gin200, nand2);
    std::cerr << "\rLUT201   ";
    Ciphertext gin201 = 2 * gout200 + gout192;
    Ciphertext gout201 = ctx.eval_lut(&gin201, nand2);
    std::cerr << "\rLUT202   ";
    Ciphertext gin202 = 2 * gout201 + gout198;
    Ciphertext gout202 = ctx.eval_lut(&gin202, nand2);
    std::cerr << "\rLUT203   ";
    Ciphertext gin203 = 2 * gout202 + gout199;
    Ciphertext gout203 = ctx.eval_lut(&gin203, nand2);
    std::cerr << "\rLUT204   ";
    Ciphertext gin204 = 2 * gout203 + gout197;
    Ciphertext gout204 = ctx.eval_lut(&gin204, nand2);
    std::cerr << "\rLUT205   ";
    Ciphertext gin205 = 2 * gout204 + gout194;
    Ciphertext gout205 = ctx.eval_lut(&gin205, nand2);
    std::cerr << "\rLUT206   ";
    Ciphertext gin206 = 2 * gout16 + gout205;
    Ciphertext gout206 = ctx.eval_lut(&gin206, and2);

    std::cerr << "\r          \r";
    std::vector<long> out;
    out.push_back(ctx.decrypt(&gout206)); /* out0 */
    out.push_back(ctx.decrypt(&gout174)); /* out1 */
    out.push_back(ctx.decrypt(&gout137)); /* out2 */
    out.push_back(ctx.decrypt(&gout107)); /* out3 */
    out.push_back(ctx.decrypt(&gout64)); /* out4 */
    out.push_back(ctx.decrypt(&gout38)); /* out5 */
    out.push_back(ctx.decrypt(&gout32)); /* out6 */
    out.push_back(ctx.decrypt(&gout28)); /* out7 */
    return out;
}

int
main(void)
{
    /* setup */
    FHEContext ctx;
    long out;

    std::cout << "[+] ";
    ctx.generate_context(tfhe_11_NTT);
    ctx.set_default_message_encoding_type(partial_domain);
    ctx.set_default_plaintext_space(4);

    /* add4 tests */
    std::cout << "[+] Testing add4" << std::endl;
    for (int i = 0; i < 16; ++i) {
        for (int j = 0; j < 16; ++j) {
            std::cerr << i << " + " << j;
            out = compose(test_add4_boolean(ctx, i, j));
            assert(out == i + j);
        }
    }

    /* calculator tests */
    std::cout << "[+] Testing calculator" << std::endl;
    std::cout << "[>] 0 0 0" << std::endl;
    std::cout << "exp:    1111111111111111" << std::endl;
    out = compose(test_calculator_boolean(ctx, 0, 0, 0));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 6 + 7" << std::endl;
    std::cout << "exp:    0000000000001101" << std::endl;
    out = compose(test_calculator_boolean(ctx, '+', 6, 7));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 6 - 7" << std::endl;
    std::cout << "exp:    1111111111111111" << std::endl;
    out = compose(test_calculator_boolean(ctx, '-', 6, 7));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 6 * 7" << std::endl;
    std::cout << "exp:    0000000000101010" << std::endl;
    out = compose(test_calculator_boolean(ctx, '*', 6, 7));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 0xba + 0xbe" << std::endl;
    std::cout << "exp:    0000000101111000" << std::endl;
    out = compose(test_calculator_boolean(ctx, '+', 0xba, 0xbe));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 0xba - 0xbe" << std::endl;
    std::cout << "exp:    1111111111111100" << std::endl;
    out = compose(test_calculator_boolean(ctx, '-', 0xba, 0xbe));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 0xba * 0xbe" << std::endl;
    std::cout << "exp:    1000101000001100" << std::endl;
    out = compose(test_calculator_boolean(ctx, '*', 0xba, 0xbe));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 0xffff + 0xffff" << std::endl;
    std::cout << "exp:    1111111111111110" << std::endl;
    out = compose(test_calculator_boolean(ctx, '+', 0xffff, 0xffff));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 0xffff - 0xffff" << std::endl;
    std::cout << "exp:    0000000000000000" << std::endl;
    out = compose(test_calculator_boolean(ctx, '-', 0xffff, 0xffff));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 0xffff * 0xffff" << std::endl;
    std::cout << "exp:    0000000000000001" << std::endl;
    out = compose(test_calculator_boolean(ctx, '*', 0xffff, 0xffff));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    /* sqrt tests */
    std::cout << "[+] Testing sqrt" << std::endl;
    std::cout << "[>] 25" << std::endl;
    std::cout << "exp:    0000000000000101" << std::endl;
    out = compose(test_sqrt_boolean(ctx, 25));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 35" << std::endl;
    std::cout << "exp:    0000000000000101" << std::endl;
    out = compose(test_sqrt_boolean(ctx, 35));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 22500" << std::endl;
    std::cout << "exp:    0000000010010110" << std::endl;
    out = compose(test_sqrt_boolean(ctx, 22500));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    std::cout << "[>] 22800" << std::endl;
    std::cout << "exp:    0000000010010110" << std::endl;
    out = compose(test_sqrt_boolean(ctx, 22800));
    std::cout << "bool:   " << std::bitset<16>(out) << std::endl;

    return 0;
}
